const std = @import("std");
const DEBUG = false;
const MAX_FRAMES = 1024;
const MAX_OFFSET = 1024;

//  h?llo matches hello, hallo and hxllo
//  h*llo matches hllo and heeeello
//  h[ae]llo matches hello and hallo, but not hillo
//  h[^e]llo matches hallo, hbllo, ... but not hello
//  h[a-b]llo matches hallo and hbllo

pub const MatchingError = error{
    IllegalCall,
    MaxFrames,
    Overmatch,
};

pub const ParsingError = error{
    DoubleAsterisk,
    IllegalBrace,
    IllegalCall,
    InvalidRange,
    OutOfMemory,
    PrematureEnd,
};

const Token = union(enum) {
    literal: []const u8,
    question_mark: void,
    asterisk: void,
    set: []const u8,
    antiset: []const u8,
};

pub fn parse_glob(pattern: []const u8, allocator: std.mem.Allocator) ParsingError!Glob {
    var parser = Parser.init(pattern, allocator);
    return try parser.parse();
}

pub const Glob = struct {
    tokens: []const Token,

    pub fn match(self: Glob, text: []const u8) MatchingError!bool {
        if (self.tokens.len == 0 and text.len == 0) return true;
        if (self.tokens.len == 0) return false;
        var D = DebugHelper.init("Glob.match()", 0);
        var partial_result = PartialResult.new(self.tokens, text);
        while (!partial_result.is_done()) {
            partial_result = try partial_result.resolve_tokens(
                text[partial_result.matched..],
                MAX_FRAMES,
            );
            D.show_partial_result("partial_result", partial_result);
        }
        return partial_result.is_passed();
    }
};

const PartialResult = struct {
    hungry_tokens: []const Token,
    matched: usize,
    to_match: usize,
    is_fail: bool,

    const Range = struct { bgn: usize, end: usize };

    fn new(hungry_tokens: []const Token, input: []const u8) PartialResult {
        return .{
            .hungry_tokens = hungry_tokens,
            .matched = 0,
            .to_match = input.len,
            .is_fail = false,
        };
    }

    /// Find closest matching range of *needle* glob token in *haystack* string.
    fn _find_anchor(haystack: []const u8, start_idx: usize, needle: Token) ?PartialResult.Range {
        switch (needle) {
            .question_mark => unreachable,
            .asterisk => unreachable,
            .literal => |s| {
                const bgn = std.mem.indexOfPos(u8, haystack, start_idx, s) orelse return null;
                return .{ .bgn = bgn, .end = bgn + s.len };
            },
            .set => |s| {
                const bgn = std.mem.indexOfAnyPos(u8, haystack, start_idx, s) orelse return null;
                return .{ .bgn = bgn, .end = bgn + 1 };
            },
            .antiset => |s| {
                const bgn = std.mem.indexOfNonePos(u8, haystack, start_idx, s) orelse return null;
                return .{ .bgn = bgn, .end = bgn + 1 };
            },
        }
    }

    fn _str_contains_char(haystack: []const u8, needle: u8) bool {
        for (haystack) |h| if (h == needle) return true;
        return false;
    }

    fn one_satied(self: PartialResult, n_bytes: usize) MatchingError!PartialResult {
        if (self.to_match < n_bytes) return MatchingError.Overmatch;
        return .{
            .hungry_tokens = self.hungry_tokens[1..],
            .matched = self.matched + n_bytes,
            .to_match = self.to_match - n_bytes,
            .is_fail = false,
        };
    }

    fn is_done(self: PartialResult) bool {
        if (self.is_fail) return true;
        return self.hungry_tokens.len == 0;
    }

    fn is_passed(self: PartialResult) bool {
        if (self.is_fail) return false;
        if (self.hungry_tokens.len > 0) return false;
        if (self.to_match > 0) return false;
        return true;
    }

    fn resolve_tokens(self: PartialResult, input: []const u8, frames: usize) MatchingError!PartialResult {
        if (frames == 0) return MatchingError.MaxFrames;
        var D = DebugHelper.init("PartialResult.resolve_tokens()", MAX_FRAMES - frames);
        D.show_int("depth", MAX_FRAMES - frames);
        D.show_str("input", input);
        D.show_partial_result("self", self);
        if (self.is_done()) return MatchingError.IllegalCall;
        const head = self.hungry_tokens[0];
        D.show_token("head", head);
        switch (head) {
            .asterisk => {
                // Conceptually we split the rest of tokens to three parts:
                //  - *head* - is "current" token, guarranteed asterisk,
                //  - *anchor* - the next token following the asterisk,
                //  - *future* - zero or more tokens following the anchor.
                //
                // The job is to find a range of text where the range matches anchor,
                // and the following text matches *future*. (This second check
                // is done by recursive call, guarded by `frames`.)
                //
                // Note that the asterisk is non-greedy, i.e. when we're looking fore
                // the anchor we start at 0 and work towards the end, short-cutting
                // when we find a match.
                D.show_message("------[*]------");
                if (self.hungry_tokens.len < 2) return try self.one_satied(input.len); // we're `*` at the end
                const anchor_token = self.hungry_tokens[1];
                const future_tokens = self.hungry_tokens[2..];
                D.show_token("anchor_token", anchor_token);
                D.show_tokens("future_tokens", future_tokens);
                switch (anchor_token) {
                    .asterisk, .question_mark => unreachable, // by Parser._normalized_tokens()
                    .literal, .set, .antiset => {
                        var offset: usize = 0;
                        while (offset <= MAX_OFFSET) {
                            D.show_int("......offset", offset);
                            const anchor_range = PartialResult._find_anchor(
                                input,
                                offset,
                                anchor_token,
                            ) orelse break;
                            offset = anchor_range.bgn + 1;
                            D.show_range("anchor_range", anchor_range);
                            const anchor_text = input[anchor_range.bgn..anchor_range.end];
                            D.show_str("anchor_text", anchor_text);
                            const post_asterisk_result = try self.one_satied(anchor_range.bgn);
                            const post_anchor_result = try post_asterisk_result.one_satied(anchor_range.end - anchor_range.bgn);
                            D.show_partial_result("post_asterisk_result", post_asterisk_result);
                            D.show_partial_result("post_anchor_result", post_anchor_result);
                            if (post_anchor_result.is_passed()) return post_anchor_result;
                            if (post_anchor_result.is_done()) continue;
                            const future_input = input[anchor_range.end..];
                            D.show_str("future_input", future_input);
                            const post_future_result = try post_anchor_result.resolve_tokens(
                                future_input,
                                frames - 1,
                            );
                            D.show_partial_result("post_future_result", post_future_result);
                            if (post_future_result.is_passed()) return post_future_result;
                        }
                        return self.to_failed();
                    },
                }
            },
            .literal => |s| {
                if (!std.mem.startsWith(u8, input, s)) return self.to_failed();
                return try self.one_satied(s.len);
            },
            .set => |s| {
                if (input.len == 0) return self.to_failed();
                if (!PartialResult._str_contains_char(s, input[0])) return self.to_failed();
                return try self.one_satied(1);
            },
            .antiset => |s| {
                if (input.len == 0) return self.to_failed();
                if (PartialResult._str_contains_char(s, input[0])) return self.to_failed();
                return try self.one_satied(1);
            },
            .question_mark => {
                if (input.len < 1) return self.to_failed();
                return try self.one_satied(1);
            },
        }
    }

    fn to_failed(self: PartialResult) PartialResult {
        return .{
            .hungry_tokens = self.hungry_tokens,
            .matched = self.matched,
            .to_match = self.to_match,
            .is_fail = true,
        };
    }
};

const Parser = struct {
    const State = enum { root, brace, rbrace };
    const Char = struct { char: u8, last: bool };
    const CharBuff = struct {
        buff: std.ArrayList(u8),
        buftype: State,
        fn is_empty(self: CharBuff) bool {
            return self.buff.items.len == 0;
        }
        fn append(self: *CharBuff, char: u8) ParsingError!void {
            self.buff.append(char) catch return ParsingError.OutOfMemory;
        }
        fn append_range(self: *CharBuff, end: u8) ParsingError!void {
            // assuming that range start was already pushed
            if (self.buftype == .root) return ParsingError.IllegalCall;
            if (self.is_empty()) return ParsingError.InvalidRange;
            const start = self.buff.items[self.buff.items.len - 1];
            var c = start;
            while (c != end) {
                if (c > end) c -= 1;
                if (c < end) c += 1;
                self.buff.append(c) catch return ParsingError.OutOfMemory;
            }
        }
        fn flush(self: *CharBuff) ParsingError!?[]const u8 {
            if (self.is_empty()) return null;
            const clone = try self.buff.clone();
            self.buff.clearAndFree();
            return clone.items;
        }
        fn set_type(self: *CharBuff, buftype: State) ParsingError!void {
            if (!self.is_empty()) return ParsingError.IllegalCall;
            self.buftype = buftype;
        }
    };

    buff: []const u8,
    cur: usize,
    state: State,
    allocator: std.mem.Allocator,
    char_buff: CharBuff,
    raw_tokens: std.ArrayList(Token),

    fn init(buff: []const u8, allocator: std.mem.Allocator) Parser {
        return Parser{
            .buff = buff,
            .state = .root,
            .cur = 0,
            .allocator = allocator,
            .char_buff = CharBuff{
                .buftype = .root,
                .buff = std.ArrayList(u8).init(allocator),
            },
            .raw_tokens = std.ArrayList(Token).init(allocator),
        };
    }

    /// Create antiset token from current char buffer, save the token
    /// to self.raw_tokens and flush the char buffer.  If the buffer was empty,
    /// do nothing.
    fn _finalize_antiset(self: *Parser) ParsingError!void {
        const chars = try self.char_buff.flush() orelse return;
        const token = Token{ .antiset = chars };
        self.raw_tokens.append(token) catch return ParsingError.OutOfMemory;
    }

    /// Create literal token from current char buffer, save the token
    /// to self.raw_tokens and flush the char buffer.  If the buffer was empty,
    /// do nothing.
    fn _finalize_literal(self: *Parser) ParsingError!void {
        const chars = try self.char_buff.flush() orelse return;
        const token = Token{ .literal = chars };
        self.raw_tokens.append(token) catch return ParsingError.OutOfMemory;
    }

    /// Create set token from current char buffer, save the token
    /// to self.raw_tokens and flush the char buffer.  If the buffer was empty,
    /// do nothing.
    fn _finalize_set(self: *Parser) ParsingError!void {
        const chars = try self.char_buff.flush() orelse return;
        const token = Token{ .set = chars };
        self.raw_tokens.append(token) catch return ParsingError.OutOfMemory;
    }

    fn _finish_brace(self: *Parser) ParsingError!void {
        switch (self.state) {
            .brace => try self._finalize_set(),
            .rbrace => try self._finalize_antiset(),
            else => return ParsingError.IllegalCall,
        }
        const next_state: State = .root;
        try self._setup_charbuff(.root);
        self.state = next_state;
    }

    /// Finish parsing of current literal token, add asterisk token and
    /// prepare for next literal token.
    fn _inject_asterisk(self: *Parser) ParsingError!void {
        if (self.state != .root) return ParsingError.IllegalCall;
        try self._finalize_literal();
        if (self.raw_tokens.items.len > 0 and self.raw_tokens.getLast() == .asterisk) return ParsingError.DoubleAsterisk;
        self.raw_tokens.append(Token{ .asterisk = {} }) catch return ParsingError.OutOfMemory;
        try self._setup_charbuff(.root);
    }

    /// Finish parsing of current literal token, add question_mark token and
    /// prepare for next literal token.
    fn _inject_question_mark(self: *Parser) ParsingError!void {
        if (self.state != .root) return ParsingError.IllegalCall;
        try self._finalize_literal();
        self.raw_tokens.append(Token{ .question_mark = {} }) catch return ParsingError.OutOfMemory;
        try self._setup_charbuff(.root);
    }

    /// Return next char, or null if there are no more chars.  Don't move the cursor.
    fn _peek_char(self: Parser) ?u8 {
        if (self.cur >= self.buff.len) return null;
        return self.buff[self.cur];
    }

    /// Pop next char (move cursor), return it, or ParsingError.PrematureEnd if there
    /// are no more chars.
    fn _require_char(self: *Parser) ParsingError!u8 {
        if (self.cur >= self.buff.len) return ParsingError.PrematureEnd;
        self.cur += 1;
        return self.buff[self.cur - 1];
    }

    /// Assert that the charbuff is empty and set its type to *buftype*.
    fn _setup_charbuff(self: *Parser, buftype: State) ParsingError!void {
        if (!self.char_buff.is_empty()) return ParsingError.IllegalCall;
        self.char_buff.buftype = buftype;
    }

    /// Switch to parsing of set ("brace" state) or antiset ("rbrace" state).
    /// Assert that the state was .root before.
    fn _start_brace(self: *Parser) ParsingError!void {
        if (self.state != .root) return ParsingError.IllegalCall;
        try self._finalize_literal();
        const next_state: State = if (self._take_exact('^')) .rbrace else .brace;
        try self._setup_charbuff(next_state);
        self.state = next_state;
    }

    /// Return next char and move cursor.  If there are no more chars, return null.
    fn _take_char(self: *Parser) ?u8 {
        if (self.cur >= self.buff.len) return null;
        const c = self.buff[self.cur];
        self.cur += 1;
        return c;
    }

    /// Move cursor if, and only if, the next char is exactly *c*.  Return true
    /// if it was the case and false otherwise.  If there are no more chars,
    /// false is returned as well.
    fn _take_exact(self: *Parser, c: u8) bool {
        if (self.cur >= self.buff.len) return false;
        if (c != self.buff[self.cur]) return false;
        self.cur += 1;
        return true;
    }

    /// Return own raw_tokens with sequences of asterisks and question marks
    /// normalized.  Any sequence of asterisks and question marks is replaced
    /// such that all question marks come first, and all asterisks are collapsed
    /// into single asterisk.
    ///
    /// For example, `foo*?*?bar` becomes `foo??*bar`.
    fn _normalized_tokens(self: Parser) ![]const Token {
        var normalized = std.ArrayList(Token).init(self.allocator);
        var asterisk = false;
        var question_marks: usize = 0;
        for (self.raw_tokens.items) |token| {
            switch (token) {
                .asterisk => asterisk = true,
                // ^^ any non-zero sequence of asterisks is equivalent to single asterisk
                .question_mark => question_marks += 1,
                .literal, .set, .antiset => {
                    try normalized.appendNTimes(Token{ .question_mark = {} }, question_marks);
                    if (asterisk) try normalized.append(Token{ .asterisk = {} });
                    question_marks = 0;
                    asterisk = false;
                    try normalized.append(token);
                },
            }
        }
        try normalized.appendNTimes(Token{ .question_mark = {} }, question_marks);
        if (asterisk) try normalized.append(Token{ .asterisk = {} });
        return normalized.items;
    }

    /// Parse our buffer and return Glob.
    fn parse(self: *Parser) ParsingError!Glob {
        if (self.buff.len == 0) return ParsingError.PrematureEnd;
        while (self._take_char()) |char| {
            switch (self.state) {
                .root => switch (char) {
                    '\\' => try self.char_buff.append(try self._require_char()),
                    '*' => try self._inject_asterisk(),
                    '?' => try self._inject_question_mark(),
                    '[' => try self._start_brace(),
                    ']' => return ParsingError.IllegalBrace,
                    else => try self.char_buff.append(char),
                },
                .brace, .rbrace => switch (char) {
                    '\\' => try self.char_buff.append(try self._require_char()),
                    ']' => try self._finish_brace(),
                    '-' => {
                        if (self._peek_char() == ']') try self.char_buff.append(char);
                        // ^^^ because '-' is taken literally if at the end of a group
                        const rng_end = try self._require_char();
                        try self.char_buff.append_range(rng_end);
                    },
                    else => try self.char_buff.append(char),
                },
            }
        }
        switch (self.state) {
            .root => try self._finalize_literal(),
            else => return ParsingError.PrematureEnd,
        }
        const normalized_tokens = self._normalized_tokens() catch |e| switch (e) {
            error.OutOfMemory => return ParsingError.OutOfMemory,
        };
        return Glob{ .tokens = normalized_tokens };
    }
};

const TestCase = struct {
    rv: ParsingError!Glob,
    buff: []const u8,

    fn init(buff: []const u8) TestCase {
        var parser = Parser.init(buff, std.heap.page_allocator);
        return .{ .rv = parser.parse(), .buff = buff };
    }

    fn expect_token_at(self: TestCase, idx: usize, oracle: Token) !void {
        const v = try self.rv;
        if (v.tokens.len <= idx) return error.NotEnoughItems;
        try std.testing.expectEqualDeep(oracle, v.tokens[idx]);
    }

    fn expect_match(self: TestCase, haystack: []const u8) !void {
        const glob = try self.rv;
        var D = DebugHelper.init("TestCase.expect_match()", 0);
        D.show_message("=======");
        D.show_fmt_message("pattern='{s}' haystack='{s}'", .{ self.buff, haystack });
        D.show_message("=======");
        try std.testing.expectEqual(true, glob.match(haystack));
    }

    fn expect_nonmatch(self: TestCase, haystack: []const u8) !void {
        const glob = try self.rv;
        var D = DebugHelper.init("TestCase.expect_nonmatch()", 0);
        D.show_message("=======");
        D.show_fmt_message("pattern='{s}' haystack='{s}'", .{ self.buff, haystack });
        D.show_message("=======");
        try std.testing.expectEqual(false, glob.match(haystack));
    }

    fn expect_len(self: TestCase, oracle: usize) !void {
        const v = try self.rv;
        try std.testing.expectEqual(oracle, v.tokens.len);
    }

    fn expect_error(self: TestCase, oracle: ParsingError) !void {
        try std.testing.expectEqual(oracle, self.rv);
    }
};

test "single-token patterns" {
    const mkcase = TestCase.init;
    try mkcase("").expect_error(ParsingError.PrematureEnd);
    try mkcase("a").expect_len(1);
    try mkcase("a").expect_token_at(0, Token{ .literal = "a" });
    try mkcase("ab").expect_token_at(0, Token{ .literal = "ab" });
    try mkcase("\\a").expect_token_at(0, Token{ .literal = "a" });
    try mkcase("\\?").expect_token_at(0, Token{ .literal = "?" });
    try mkcase("\\*").expect_token_at(0, Token{ .literal = "*" });
    try mkcase("\\[").expect_token_at(0, Token{ .literal = "[" });
    try mkcase("\\]").expect_token_at(0, Token{ .literal = "]" });
    try mkcase("\\t").expect_token_at(0, Token{ .literal = "t" });
    try mkcase("\\\\").expect_token_at(0, Token{ .literal = "\\" });
    try mkcase("?").expect_len(1);
    try mkcase("?").expect_token_at(0, Token{ .question_mark = {} });
    try mkcase("*").expect_len(1);
    try mkcase("*").expect_token_at(0, Token{ .asterisk = {} });
    try mkcase("[").expect_error(ParsingError.PrematureEnd);
    try mkcase("[]").expect_len(0);
    try mkcase("[[").expect_error(ParsingError.PrematureEnd);
    try mkcase("[[]").expect_len(1);
    try mkcase("[[]").expect_token_at(0, Token{ .set = "[" });
    try mkcase("[a]").expect_token_at(0, Token{ .set = "a" });
    try mkcase("[ab]").expect_token_at(0, Token{ .set = "ab" });
    try mkcase("[a\\-]").expect_token_at(0, Token{ .set = "a-" });
    try mkcase("[a-a]").expect_token_at(0, Token{ .set = "a" });
    try mkcase("[a-b]").expect_token_at(0, Token{ .set = "ab" });
    try mkcase("[a-c]").expect_token_at(0, Token{ .set = "abc" });
    try mkcase("[b-a]").expect_token_at(0, Token{ .set = "ba" });
    try mkcase("[c-a]").expect_token_at(0, Token{ .set = "cba" });
    try mkcase("[^").expect_error(ParsingError.PrematureEnd);
    try mkcase("[^]").expect_len(0);
    try mkcase("[^[").expect_error(ParsingError.PrematureEnd);
    try mkcase("[^[]").expect_len(1);
    try mkcase("[^[]").expect_token_at(0, Token{ .antiset = "[" });
    try mkcase("[^a]").expect_token_at(0, Token{ .antiset = "a" });
    try mkcase("[^ab]").expect_token_at(0, Token{ .antiset = "ab" });
    try mkcase("[^a\\-]").expect_token_at(0, Token{ .antiset = "a-" });
    try mkcase("[^a-a]").expect_token_at(0, Token{ .antiset = "a" });
    try mkcase("[^a-b]").expect_token_at(0, Token{ .antiset = "ab" });
    try mkcase("[^a-c]").expect_token_at(0, Token{ .antiset = "abc" });
    try mkcase("[^b-a]").expect_token_at(0, Token{ .antiset = "ba" });
    try mkcase("[^c-a]").expect_token_at(0, Token{ .antiset = "cba" });
}

test "single-token: literal" {
    const mkcase = TestCase.init;
    try mkcase("a").expect_nonmatch("");
    try mkcase("a").expect_match("a");
    try mkcase("a").expect_nonmatch("ab");
    try mkcase("a").expect_nonmatch("x");
    try mkcase("a").expect_nonmatch("xy");
    try mkcase("ab").expect_nonmatch("");
    try mkcase("ab").expect_nonmatch("a");
    try mkcase("ab").expect_match("ab");
    try mkcase("ab").expect_nonmatch("abc");
    try mkcase("ab").expect_nonmatch("x");
    try mkcase("ab").expect_nonmatch("xy");
    try mkcase("\\?").expect_match("?");
    try mkcase("\\*").expect_match("*");
    try mkcase("\\[").expect_match("[");
}

test "single-token: question_mark" {
    const mkcase = TestCase.init;
    try mkcase("?").expect_nonmatch("");
    try mkcase("?").expect_match("a");
    try mkcase("?").expect_nonmatch("ab");
}

test "single-token: asterisk" {
    const mkcase = TestCase.init;
    try mkcase("*").expect_match("");
    try mkcase("*").expect_match("a");
    try mkcase("*").expect_match("ab");
}

test "single-token: set" {
    const mkcase = TestCase.init;
    try mkcase("[a]").expect_nonmatch("");
    try mkcase("[a]").expect_match("a");
    try mkcase("[a]").expect_nonmatch("ab");
    try mkcase("[a]").expect_nonmatch("x");
    try mkcase("[a]").expect_nonmatch("xa");
    try mkcase("[ab]").expect_nonmatch("");
    try mkcase("[ab]").expect_match("a");
    try mkcase("[ab]").expect_match("b");
    try mkcase("[ab]").expect_nonmatch("c");
    try mkcase("[ab]").expect_nonmatch("x");
    try mkcase("[ab]").expect_nonmatch("ab");
    try mkcase("[ab]").expect_nonmatch("bx");
    try mkcase("[ab]").expect_nonmatch("xa");
    try mkcase("[ab]").expect_nonmatch("xb");
    try mkcase("[am]").expect_nonmatch("");
    try mkcase("[am]").expect_match("a");
    try mkcase("[am]").expect_match("m");
    try mkcase("[am]").expect_nonmatch("b");
    try mkcase("[am]").expect_nonmatch("l");
    try mkcase("[am]").expect_nonmatch("n");
    try mkcase("[am]").expect_nonmatch("am");
    try mkcase("[am]").expect_nonmatch("mx");
    try mkcase("[am]").expect_nonmatch("x");
    try mkcase("[am]").expect_nonmatch("xa");
    try mkcase("[am]").expect_nonmatch("xm");
    try mkcase("[a-b]").expect_nonmatch("");
    try mkcase("[a-b]").expect_match("a");
    try mkcase("[a-b]").expect_match("b");
    try mkcase("[a-b]").expect_nonmatch("c");
    try mkcase("[a-b]").expect_nonmatch("x");
    try mkcase("[a-m]").expect_nonmatch("");
    try mkcase("[a-m]").expect_match("a");
    try mkcase("[a-m]").expect_match("b");
    try mkcase("[a-m]").expect_match("l");
    try mkcase("[a-m]").expect_match("m");
    try mkcase("[a-m]").expect_nonmatch("n");
    try mkcase("[a-m]").expect_nonmatch("x");
}

test "single-token: antiset" {
    const mkcase = TestCase.init;
    try mkcase("[^a]").expect_nonmatch("");
    try mkcase("[^a]").expect_nonmatch("a");
    try mkcase("[^a]").expect_nonmatch("ab");
    try mkcase("[^a]").expect_match("x");
    try mkcase("[^a]").expect_nonmatch("xa");
    try mkcase("[^ab]").expect_nonmatch("");
    try mkcase("[^ab]").expect_nonmatch("a");
    try mkcase("[^ab]").expect_nonmatch("b");
    try mkcase("[^ab]").expect_match("c");
    try mkcase("[^ab]").expect_match("x");
    try mkcase("[^ab]").expect_nonmatch("ab");
    try mkcase("[^ab]").expect_nonmatch("bx");
    try mkcase("[^ab]").expect_nonmatch("xa");
    try mkcase("[^ab]").expect_nonmatch("xb");
    try mkcase("[^am]").expect_nonmatch("");
    try mkcase("[^am]").expect_nonmatch("a");
    try mkcase("[^am]").expect_match("b");
    try mkcase("[^am]").expect_match("l");
    try mkcase("[^am]").expect_nonmatch("m");
    try mkcase("[^am]").expect_match("n");
    try mkcase("[^am]").expect_match("x");
    try mkcase("[^am]").expect_nonmatch("am");
    try mkcase("[^am]").expect_nonmatch("mx");
    try mkcase("[^am]").expect_nonmatch("xa");
    try mkcase("[^am]").expect_nonmatch("xm");
    try mkcase("[^a-b]").expect_nonmatch("");
    try mkcase("[^a-b]").expect_nonmatch("a");
    try mkcase("[^a-b]").expect_nonmatch("b");
    try mkcase("[^a-b]").expect_match("c");
    try mkcase("[^a-b]").expect_match("x");
    try mkcase("[^a-m]").expect_nonmatch("");
    try mkcase("[^a-m]").expect_nonmatch("a");
    try mkcase("[^a-m]").expect_nonmatch("b");
    try mkcase("[^a-m]").expect_nonmatch("l");
    try mkcase("[^a-m]").expect_nonmatch("m");
    try mkcase("[^a-m]").expect_match("n");
    try mkcase("[^a-m]").expect_match("x");
}

test "two-token patterns" {
    const mkcase = TestCase.init;
    //
    // combinations of 2 tokens:
    //
    //   literal        literal         # impossible
    //   literal        question_mark
    //   literal        asterisk
    //   literal        set
    //   literal        antiset
    try mkcase("ab?").expect_len(2);
    try mkcase("ab?").expect_token_at(0, Token{ .literal = "ab" });
    try mkcase("ab?").expect_token_at(1, Token{ .question_mark = {} });
    try mkcase("ab*").expect_len(2);
    try mkcase("ab*").expect_token_at(0, Token{ .literal = "ab" });
    try mkcase("ab*").expect_token_at(1, Token{ .asterisk = {} });
    try mkcase("ab[d-f]").expect_len(2);
    try mkcase("ab[d-f]").expect_token_at(0, Token{ .literal = "ab" });
    try mkcase("ab[d-f]").expect_token_at(1, Token{ .set = "def" });
    try mkcase("ab[^d-f]").expect_len(2);
    try mkcase("ab[^d-f]").expect_token_at(0, Token{ .literal = "ab" });
    try mkcase("ab[^d-f]").expect_token_at(1, Token{ .antiset = "def" });
    //   question_mark  literal
    //   question_mark  question_mark
    //   question_mark  asterisk
    //   question_mark  set
    //   question_mark  antiset
    try mkcase("?ab").expect_len(2);
    try mkcase("?ab").expect_token_at(0, Token{ .question_mark = {} });
    try mkcase("?ab").expect_token_at(1, Token{ .literal = "ab" });
    try mkcase("??").expect_len(2);
    try mkcase("??").expect_token_at(0, Token{ .question_mark = {} });
    try mkcase("??").expect_token_at(1, Token{ .question_mark = {} });
    try mkcase("?*").expect_len(2);
    try mkcase("?*").expect_token_at(0, Token{ .question_mark = {} });
    try mkcase("?*").expect_token_at(1, Token{ .asterisk = {} });
    try mkcase("?[d-f]").expect_len(2);
    try mkcase("?[d-f]").expect_token_at(0, Token{ .question_mark = {} });
    try mkcase("?[d-f]").expect_token_at(1, Token{ .set = "def" });
    try mkcase("?[^d-f]").expect_len(2);
    try mkcase("?[^d-f]").expect_token_at(0, Token{ .question_mark = {} });
    try mkcase("?[^d-f]").expect_token_at(1, Token{ .antiset = "def" });
    //   asterisk       literal
    //   asterisk       question_mark
    //   asterisk       asterisk        # illegal
    //   asterisk       set
    //   asterisk       antiset
    try mkcase("*ab").expect_len(2);
    try mkcase("*ab").expect_token_at(0, Token{ .asterisk = {} });
    try mkcase("*ab").expect_token_at(1, Token{ .literal = "ab" });
    try mkcase("*?").expect_len(2); // gets swapped by Parser._normalized_tokens()
    try mkcase("*?").expect_token_at(0, Token{ .question_mark = {} });
    try mkcase("*?").expect_token_at(1, Token{ .asterisk = {} });
    try mkcase("**").expect_error(ParsingError.DoubleAsterisk);
    try mkcase("*[d-f]").expect_len(2);
    try mkcase("*[d-f]").expect_token_at(0, Token{ .asterisk = {} });
    try mkcase("*[d-f]").expect_token_at(1, Token{ .set = "def" });
    try mkcase("*[^d-f]").expect_len(2);
    try mkcase("*[^d-f]").expect_token_at(0, Token{ .asterisk = {} });
    try mkcase("*[^d-f]").expect_token_at(1, Token{ .antiset = "def" });
    //   set            literal
    //   set            question_mark
    //   set            asterisk
    //   set            set
    //   set            antiset
    try mkcase("[a-c]ab").expect_len(2);
    try mkcase("[a-c]ab").expect_token_at(0, Token{ .set = "abc" });
    try mkcase("[a-c]ab").expect_token_at(1, Token{ .literal = "ab" });
    try mkcase("[a-c]?").expect_len(2);
    try mkcase("[a-c]?").expect_token_at(0, Token{ .set = "abc" });
    try mkcase("[a-c]?").expect_token_at(1, Token{ .question_mark = {} });
    try mkcase("[a-c]*").expect_len(2);
    try mkcase("[a-c]*").expect_token_at(0, Token{ .set = "abc" });
    try mkcase("[a-c]*").expect_token_at(1, Token{ .asterisk = {} });
    try mkcase("[a-c][d-f]").expect_len(2);
    try mkcase("[a-c][d-f]").expect_token_at(0, Token{ .set = "abc" });
    try mkcase("[a-c][d-f]").expect_token_at(1, Token{ .set = "def" });
    try mkcase("[a-c][^d-f]").expect_len(2);
    try mkcase("[a-c][^d-f]").expect_token_at(0, Token{ .set = "abc" });
    try mkcase("[a-c][^d-f]").expect_token_at(1, Token{ .antiset = "def" });
    //   antiset        literal
    //   antiset        question_mark
    //   antiset        asterisk
    //   antiset        set
    //   antiset        antiset
    try mkcase("[^a-c]ab").expect_len(2);
    try mkcase("[^a-c]ab").expect_token_at(0, Token{ .antiset = "abc" });
    try mkcase("[^a-c]ab").expect_token_at(1, Token{ .literal = "ab" });
    try mkcase("[^a-c]?").expect_len(2);
    try mkcase("[^a-c]?").expect_token_at(0, Token{ .antiset = "abc" });
    try mkcase("[^a-c]?").expect_token_at(1, Token{ .question_mark = {} });
    try mkcase("[^a-c]*").expect_len(2);
    try mkcase("[^a-c]*").expect_token_at(0, Token{ .antiset = "abc" });
    try mkcase("[^a-c]*").expect_token_at(1, Token{ .asterisk = {} });
    try mkcase("[^a-c][d-f]").expect_len(2);
    try mkcase("[^a-c][d-f]").expect_token_at(0, Token{ .antiset = "abc" });
    try mkcase("[^a-c][d-f]").expect_token_at(1, Token{ .set = "def" });
    try mkcase("[^a-c][^d-f]").expect_len(2);
    try mkcase("[^a-c][^d-f]").expect_token_at(0, Token{ .antiset = "abc" });
    try mkcase("[^a-c][^d-f]").expect_token_at(1, Token{ .antiset = "def" });
}

test "three-token patterns: literal,.." {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;

    // group: literal, literal, ..
    //   .. literal          # impossible
    //   .. question_mark    # impossible
    //   .. asterisk         # impossible
    //   .. set              # impossible
    //   .. antiset          # impossible

    // group: literal, question_mark, ..
    //   ..  literal
    //   ..  question_mark
    //   ..  asterisk
    //   ..  set
    //   ..  antiset
    c = mkcase("ab?cd");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .literal = "ab" });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .literal = "cd" });
    try c.expect_nonmatch("abcd");
    try c.expect_match("abAcd");
    try c.expect_nonmatch("XabAcd");
    try c.expect_nonmatch("abAcdX");
    try c.expect_nonmatch("abXYcd");
    c = mkcase("ab??");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .literal = "ab" });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .question_mark = {} });
    try c.expect_nonmatch("ab");
    try c.expect_nonmatch("abc");
    try c.expect_match("abcd");
    try c.expect_nonmatch("abcdX");
    try c.expect_nonmatch("Xabcd");
    try c.expect_nonmatch("abcXd");
    try c.expect_nonmatch("XabcXd");
    try c.expect_nonmatch("abXYcd");
    c = mkcase("ab?*");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .literal = "ab" });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_nonmatch("ab");
    try c.expect_match("abH");
    try c.expect_match("abHD");
    try c.expect_nonmatch("XabHDE");
    try c.expect_nonmatch("XabHDE");
    c = mkcase("ab?[d-f]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .literal = "ab" });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .set = "def" });
    try c.expect_match("abAd");
    try c.expect_match("abAf");
    try c.expect_nonmatch("abAg");
    try c.expect_nonmatch("abAc");
    try c.expect_nonmatch("XabAd");
    try c.expect_nonmatch("abAdX");
    c = mkcase("ab?[^d-f]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .literal = "ab" });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .antiset = "def" });
    try c.expect_match("abAc");
    try c.expect_match("abAg");
    try c.expect_nonmatch("abAd");
    try c.expect_nonmatch("abAf");
    try c.expect_nonmatch("XabAc");
    try c.expect_nonmatch("abAcX");

    // group: literal, asterisk, ..
    //   .. literal
    //   .. question_mark
    //   .. asterisk
    //   .. set
    //   .. antiset
    c = mkcase("ab*cd");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .literal = "ab" });
    try c.expect_token_at(1, Token{ .asterisk = {} });
    try c.expect_token_at(2, Token{ .literal = "cd" });
    try c.expect_match("abcd");
    try c.expect_match("abAcd");
    try c.expect_match("abABcd");
    try c.expect_nonmatch("Xabcd");
    try c.expect_nonmatch("abABcdX");
    c = mkcase("ab*?");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .literal = "ab" });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_match("abQ");
    try c.expect_match("abHQ");
    try c.expect_match("abHIQ");
    try c.expect_nonmatch("XabQ");
    try c.expect_nonmatch("XabHQ");
    try c.expect_nonmatch("XabHIQ");
    c = mkcase("ab**");
    try c.expect_error(ParsingError.DoubleAsterisk);
    c = mkcase("ab*[d-f]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .literal = "ab" });
    try c.expect_token_at(1, Token{ .asterisk = {} });
    try c.expect_token_at(2, Token{ .set = "def" });
    try c.expect_match("abd");
    try c.expect_match("abHd");
    try c.expect_match("abHId");
    try c.expect_nonmatch("abHIg");
    try c.expect_nonmatch("abHIc");
    try c.expect_nonmatch("Xabd");
    try c.expect_nonmatch("abdX");
    c = mkcase("ab*[^d-f]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .literal = "ab" });
    try c.expect_token_at(1, Token{ .asterisk = {} });
    try c.expect_token_at(2, Token{ .antiset = "def" });
    try c.expect_match("abc");
    try c.expect_match("abg");
    try c.expect_match("abHg");
    try c.expect_match("abHIg");
    try c.expect_nonmatch("abHId");
    try c.expect_nonmatch("abHIf");
    try c.expect_nonmatch("Xabd");
    try c.expect_match("abHIgX");

    // group: literal, set, ..
    //   .. literal
    //   .. question_mark
    //   .. asterisk
    //   .. set
    //   .. antiset
    c = mkcase("ab[d-f]mn");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .literal = "ab" });
    try c.expect_token_at(1, Token{ .set = "def" });
    try c.expect_token_at(2, Token{ .literal = "mn" });
    try c.expect_match("abdmn");
    try c.expect_match("abfmn");
    try c.expect_nonmatch("abgmn");
    try c.expect_nonmatch("abgmnX");
    try c.expect_nonmatch("Xabgmn");
    c = mkcase("ab[d-f]?");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .literal = "ab" });
    try c.expect_token_at(1, Token{ .set = "def" });
    try c.expect_token_at(2, Token{ .question_mark = {} });
    try c.expect_match("abdQ");
    try c.expect_match("abfQ");
    try c.expect_nonmatch("abgQ");
    try c.expect_nonmatch("abd");
    try c.expect_nonmatch("abf");
    try c.expect_nonmatch("abdQX");
    try c.expect_nonmatch("XabdQ");
    c = mkcase("ab[d-f]*");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .literal = "ab" });
    try c.expect_token_at(1, Token{ .set = "def" });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_match("abd");
    try c.expect_match("abdH");
    try c.expect_match("abfH");
    try c.expect_match("abfHI");
    try c.expect_nonmatch("abgH");
    try c.expect_nonmatch("Xabd");
    try c.expect_nonmatch("XabdH");
    c = mkcase("ab[d-f][p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .literal = "ab" });
    try c.expect_token_at(1, Token{ .set = "def" });
    try c.expect_token_at(2, Token{ .set = "pqrs" });
    try c.expect_match("abdp");
    try c.expect_match("abfp");
    try c.expect_nonmatch("ab");
    try c.expect_nonmatch("abd");
    try c.expect_nonmatch("abgp");
    try c.expect_nonmatch("abdt");
    try c.expect_nonmatch("Xabdp");
    try c.expect_nonmatch("abdpX");
    c = mkcase("ab[d-f][^p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .literal = "ab" });
    try c.expect_token_at(1, Token{ .set = "def" });
    try c.expect_token_at(2, Token{ .antiset = "pqrs" });
    try c.expect_match("abdt");
    try c.expect_match("abff");
    try c.expect_nonmatch("ab");
    try c.expect_nonmatch("abd");
    try c.expect_nonmatch("abgp");
    try c.expect_nonmatch("abds");
    try c.expect_nonmatch("Xabdt");
    try c.expect_nonmatch("abdtX");
}

test "three-token patterns: question_mark, literal, .." {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    //   .. literal          # impossible
    //   .. question_mark
    c = mkcase("?ab?");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .literal = "ab" });
    try c.expect_token_at(2, Token{ .question_mark = {} });
    try c.expect_match("QabR");
    try c.expect_nonmatch("ab");
    try c.expect_nonmatch("Qab");
    try c.expect_nonmatch("abR");
    try c.expect_nonmatch("XQabR");
    try c.expect_nonmatch("QabRX");
    //   .. asterisk
    c = mkcase("?ab*");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .literal = "ab" });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_match("Qab");
    try c.expect_match("QabH");
    try c.expect_match("QabHI");
    try c.expect_match("Qabab");
    try c.expect_nonmatch("ab");
    try c.expect_nonmatch("abHI");
    try c.expect_nonmatch("abH");
    try c.expect_nonmatch("XQab");
    //   .. set
    c = mkcase("?ab[d-f]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .literal = "ab" });
    try c.expect_token_at(2, Token{ .set = "def" });
    try c.expect_match("Qabd");
    try c.expect_match("Qabf");
    try c.expect_nonmatch("Qabg");
    try c.expect_nonmatch("abf");
    try c.expect_nonmatch("XQabd");
    try c.expect_nonmatch("QabdX");
    //   .. antiset
    c = mkcase("?ab[^d-f]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .literal = "ab" });
    try c.expect_token_at(2, Token{ .antiset = "def" });
    try c.expect_match("Qabc");
    try c.expect_match("Qabg");
    try c.expect_nonmatch("Qabd");
    try c.expect_nonmatch("abc");
    try c.expect_nonmatch("XQabc");
    try c.expect_nonmatch("QabcX");
}

test "three-token patterns: question_mark question_mark, .." {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    //   .. literal
    c = mkcase("??ab");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .literal = "ab" });
    try c.expect_match("QRab");
    try c.expect_match("abab");
    try c.expect_nonmatch("Qab");
    try c.expect_nonmatch("ab");
    try c.expect_nonmatch("XQRab");
    try c.expect_nonmatch("QRabX");
    //   .. question_mark
    c = mkcase("???");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .question_mark = {} });
    try c.expect_match("QRS");
    try c.expect_nonmatch("Q");
    try c.expect_nonmatch("QR");
    try c.expect_nonmatch("QRSX");
    //   .. asterisk
    c = mkcase("??*");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_match("QR");
    try c.expect_match("QRH");
    try c.expect_match("QRHI");
    try c.expect_nonmatch("");
    try c.expect_nonmatch("Q");
    //   .. set
    c = mkcase("??[d-f]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .set = "def" });
    try c.expect_match("QRd");
    try c.expect_match("QRf");
    try c.expect_match("ddd");
    try c.expect_nonmatch("QR");
    try c.expect_nonmatch("QRg");
    try c.expect_nonmatch("QRc");
    try c.expect_nonmatch("QRdX");
    try c.expect_nonmatch("dddd");
    //   .. antiset
    c = mkcase("??[^d-f]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .antiset = "def" });
    try c.expect_match("QRc");
    try c.expect_match("QRg");
    try c.expect_match("ccc");
    try c.expect_nonmatch("QR");
    try c.expect_nonmatch("QRd");
    try c.expect_nonmatch("QRcX");
    try c.expect_nonmatch("cccc");
}

test "three-token patterns: question_mark, asterisk, .." {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    //   .. literal
    c = mkcase("?*ab");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .asterisk = {} });
    try c.expect_token_at(2, Token{ .literal = "ab" });
    try c.expect_match("Qab");
    try c.expect_match("QHab");
    try c.expect_match("QHIab");
    try c.expect_match("QHabVWab"); // worked extra hard for this one!
    try c.expect_nonmatch("ab");
    try c.expect_nonmatch("QabX");
    try c.expect_nonmatch("QHabX");
    try c.expect_nonmatch("QHIabX");
    try c.expect_nonmatch("QRabVWabX");
    //   .. question_mark
    c = mkcase("?*?");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_match("QR");
    try c.expect_match("QHR");
    try c.expect_match("QHIR");
    try c.expect_nonmatch("");
    try c.expect_nonmatch("Q");
    //   .. asterisk
    c = mkcase("?**");
    try c.expect_error(ParsingError.DoubleAsterisk);
    //   .. set
    c = mkcase("?*[d-f]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .asterisk = {} });
    try c.expect_token_at(2, Token{ .set = "def" });
    try c.expect_match("Qd");
    try c.expect_match("Qf");
    try c.expect_match("QHd");
    try c.expect_match("QHId");
    try c.expect_nonmatch("");
    try c.expect_nonmatch("Q");
    try c.expect_nonmatch("g");
    try c.expect_nonmatch("Qg");
    try c.expect_nonmatch("QHg");
    try c.expect_nonmatch("QHIg");
    try c.expect_nonmatch("QHIdX");
    //   .. antiset
    c = mkcase("?*[^d-f]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .asterisk = {} });
    try c.expect_token_at(2, Token{ .antiset = "def" });
    try c.expect_match("Qc");
    try c.expect_match("Qg");
    try c.expect_match("QHc");
    try c.expect_match("QHIc");
    try c.expect_nonmatch("");
    try c.expect_nonmatch("Q");
    try c.expect_nonmatch("d");
    try c.expect_nonmatch("Qd");
    try c.expect_nonmatch("QHd");
    try c.expect_nonmatch("QHId");
}

test "three-token patterns: question_mark, set, .." {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    //   .. literal
    c = mkcase("?[d-f]ab");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .set = "def" });
    try c.expect_token_at(2, Token{ .literal = "ab" });
    try c.expect_match("Qdab");
    try c.expect_match("Qfab");
    try c.expect_nonmatch("dab");
    try c.expect_nonmatch("Qgab");
    try c.expect_nonmatch("XQdab");
    try c.expect_nonmatch("QdabX");
    //   .. question_mark
    c = mkcase("?[d-f]?");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .set = "def" });
    try c.expect_token_at(2, Token{ .question_mark = {} });
    try c.expect_match("QdR");
    try c.expect_match("QfR");
    try c.expect_nonmatch("");
    try c.expect_nonmatch("f");
    try c.expect_nonmatch("Qf");
    try c.expect_nonmatch("QgR");
    try c.expect_nonmatch("XQdR");
    try c.expect_nonmatch("QdRX");
    //   .. asterisk
    c = mkcase("?[d-f]*");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .set = "def" });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_match("Qd");
    try c.expect_match("QdH");
    try c.expect_match("QdHI");
    try c.expect_nonmatch("Q");
    try c.expect_nonmatch("Qg");
    try c.expect_nonmatch("QgH");
    try c.expect_nonmatch("QgHI");
    try c.expect_nonmatch("XQd");
    //   .. set
    c = mkcase("?[d-f][p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .set = "def" });
    try c.expect_token_at(2, Token{ .set = "pqrs" });
    try c.expect_match("Qdp");
    try c.expect_match("Qfp");
    try c.expect_match("Qds");
    try c.expect_nonmatch("");
    try c.expect_nonmatch("Q");
    try c.expect_nonmatch("Qd");
    try c.expect_nonmatch("Qcs");
    try c.expect_nonmatch("Qct");
    try c.expect_nonmatch("QdpX");
    try c.expect_nonmatch("XQdp");
    //   .. antiset
    c = mkcase("?[d-f][^p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .set = "def" });
    try c.expect_token_at(2, Token{ .antiset = "pqrs" });
    try c.expect_match("Qdo");
    try c.expect_match("Qft");
    try c.expect_nonmatch("");
    try c.expect_nonmatch("Q");
    try c.expect_nonmatch("Qd");
    try c.expect_nonmatch("Qdp");
    try c.expect_nonmatch("Qds");
    try c.expect_nonmatch("QdoX");
    try c.expect_nonmatch("XQdo");
}

test "three-token patterns: question_mark, antiset, .." {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    //   .. literal
    c = mkcase("?[^d-f]ab");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .antiset = "def" });
    try c.expect_token_at(2, Token{ .literal = "ab" });
    try c.expect_match("Qcab");
    try c.expect_match("Qgab");
    try c.expect_nonmatch("cab");
    try c.expect_nonmatch("Qc");
    try c.expect_nonmatch("Qab");
    try c.expect_nonmatch("XQcab");
    try c.expect_nonmatch("QcabX");
    //   .. question_mark
    c = mkcase("?[^d-f]?");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .antiset = "def" });
    try c.expect_token_at(2, Token{ .question_mark = {} });
    try c.expect_match("QcR");
    try c.expect_match("QgR");
    try c.expect_nonmatch("QdR");
    try c.expect_nonmatch("Qc");
    try c.expect_nonmatch("cR");
    try c.expect_nonmatch("XQcR");
    try c.expect_nonmatch("QcRX");
    //   .. asterisk
    c = mkcase("?[^d-f]*");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .antiset = "def" });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_match("Qc");
    try c.expect_match("QcH");
    try c.expect_match("QcHI");
    try c.expect_nonmatch("Qd");
    try c.expect_nonmatch("QdH");
    try c.expect_nonmatch("QdHI");
    try c.expect_nonmatch("Q");
    try c.expect_nonmatch("c");
    try c.expect_nonmatch("cd");
    try c.expect_nonmatch("cdH");
    //   .. set
    c = mkcase("?[^d-f][p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .antiset = "def" });
    try c.expect_token_at(2, Token{ .set = "pqrs" });
    try c.expect_match("Qcp");
    try c.expect_match("Qcs");
    try c.expect_nonmatch("Qdp");
    try c.expect_nonmatch("Qco");
    try c.expect_nonmatch("Qc");
    try c.expect_nonmatch("cp");
    try c.expect_nonmatch("QcpX");
    //   .. antiset
    c = mkcase("?[^d-f][^p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .antiset = "def" });
    try c.expect_token_at(2, Token{ .antiset = "pqrs" });
    try c.expect_match("Qco");
    try c.expect_match("Qct");
    try c.expect_match("Qgo");
    try c.expect_match("Qgt");
    try c.expect_nonmatch("Qcp");
    try c.expect_nonmatch("Qdo");
    try c.expect_nonmatch("Qc");
    try c.expect_nonmatch("Q");
    try c.expect_nonmatch("co");
    try c.expect_nonmatch("o");
}

test "three-token patterns: asterisk, literal, .." {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    //   .. literal      # impossible
    //   .. question_mark
    c = mkcase("*ab?");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .asterisk = {} });
    try c.expect_token_at(1, Token{ .literal = "ab" });
    try c.expect_token_at(2, Token{ .question_mark = {} });
    try c.expect_match("abQ");
    try c.expect_match("HabQ");
    try c.expect_match("HIabQ");
    try c.expect_match("ababQ");
    try c.expect_match("HababQ");
    try c.expect_match("HIababQ");
    try c.expect_nonmatch("abQR");
    try c.expect_nonmatch("ababQR");
    //   .. asterisk
    c = mkcase("*ab*");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .asterisk = {} });
    try c.expect_token_at(1, Token{ .literal = "ab" });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_match("ab");
    try c.expect_match("Hab");
    try c.expect_match("abH");
    try c.expect_match("abHI");
    try c.expect_match("Hab");
    try c.expect_match("HIab");
    try c.expect_match("HIabJ");
    try c.expect_match("HIabJK");
    try c.expect_match("abHIabJK");
    try c.expect_match("abab");
    try c.expect_nonmatch("a");
    try c.expect_nonmatch("aH");
    try c.expect_nonmatch("Hb");
    //   .. set
    c = mkcase("*ab[d-f]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .asterisk = {} });
    try c.expect_token_at(1, Token{ .literal = "ab" });
    try c.expect_token_at(2, Token{ .set = "def" });
    try c.expect_match("abd");
    try c.expect_match("abf");
    try c.expect_match("Habd");
    try c.expect_match("HIabd");
    try c.expect_match("abHIabd");
    try c.expect_match("HIabJKabd");
    try c.expect_match("abdHIabd");
    try c.expect_match("HIabdJKabd");
    try c.expect_nonmatch("abc");
    try c.expect_nonmatch("Habc");
    try c.expect_nonmatch("HIabc");
    //   .. antiset
    c = mkcase("*ab[^d-f]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .asterisk = {} });
    try c.expect_token_at(1, Token{ .literal = "ab" });
    try c.expect_token_at(2, Token{ .antiset = "def" });
    try c.expect_match("abc");
    try c.expect_match("abg");
    try c.expect_match("Habc");
    try c.expect_match("HIabc");
    try c.expect_match("abHIabc");
    try c.expect_match("HIabJKabc");
    try c.expect_match("abcHIabc");
    try c.expect_match("HIabcJKabc");
    try c.expect_nonmatch("abd");
    try c.expect_nonmatch("Habd");
    try c.expect_nonmatch("HIabd");
}

test "three-token patterns: asterisk, question_mark.." {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    //   .. literal
    c = mkcase("*?ab");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .asterisk = {} });
    try c.expect_token_at(2, Token{ .literal = "ab" });
    try c.expect_match("Qab");
    try c.expect_match("QHab");
    try c.expect_match("QHIab");
    try c.expect_match("QHabVWab"); // worked extra hard for this one!
    try c.expect_nonmatch("ab");
    try c.expect_nonmatch("QabX");
    try c.expect_nonmatch("QHabX");
    try c.expect_nonmatch("QHIabX");
    try c.expect_nonmatch("QRabVWabX");
    //   .. question_mark
    c = mkcase("*??");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_match("QR");
    try c.expect_match("QHR");
    try c.expect_match("QHIR");
    try c.expect_nonmatch("");
    try c.expect_nonmatch("Q");
    //   .. asterisk
    c = mkcase("*?*"); // sneaks past ParsingError.DoubleAsterisk (it's a weakly held rule though)
    try c.expect_len(2); // yes, see Parser._normalized_tokens()
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .asterisk = {} });
    try c.expect_match("Q");
    try c.expect_match("QH");
    try c.expect_match("QHI");
    try c.expect_nonmatch("");
    //   .. set
    c = mkcase("*?[d-f]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .asterisk = {} });
    try c.expect_token_at(2, Token{ .set = "def" });
    try c.expect_match("Qd");
    try c.expect_match("Qf");
    try c.expect_match("QHd");
    try c.expect_match("QHId");
    try c.expect_nonmatch("");
    try c.expect_nonmatch("Q");
    try c.expect_nonmatch("g");
    try c.expect_nonmatch("Qg");
    try c.expect_nonmatch("QHg");
    try c.expect_nonmatch("QHIg");
    try c.expect_nonmatch("QHIdX");
    //   .. antiset
    c = mkcase("*?[^d-f]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .question_mark = {} });
    try c.expect_token_at(1, Token{ .asterisk = {} });
    try c.expect_token_at(2, Token{ .antiset = "def" });
    try c.expect_match("Qc");
    try c.expect_match("Qg");
    try c.expect_match("QHc");
    try c.expect_match("QHIc");
    try c.expect_nonmatch("");
    try c.expect_nonmatch("Q");
    try c.expect_nonmatch("d");
    try c.expect_nonmatch("Qd");
    try c.expect_nonmatch("QHd");
    try c.expect_nonmatch("QHId");
}

test "three-token patterns: asterisk, asterisk,.." {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    // group: asterisk, asterisk, ..
    //   .. literal          # illegal
    c = mkcase("**ab");
    try c.expect_error(ParsingError.DoubleAsterisk);
    //   .. question_mark    # illegal
    c = mkcase("**?");
    try c.expect_error(ParsingError.DoubleAsterisk);
    //   .. asterisk         # illegal
    c = mkcase("***");
    try c.expect_error(ParsingError.DoubleAsterisk);
    //   .. set              # illegal
    c = mkcase("**[d-f]");
    try c.expect_error(ParsingError.DoubleAsterisk);
    //   .. antiset          # illegal
    c = mkcase("**[^d-f]");
    try c.expect_error(ParsingError.DoubleAsterisk);
}

test "three-token patterns: asterisk, set, .." {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    //   .. literal
    c = mkcase("*[d-f]uv");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .asterisk = {} });
    try c.expect_token_at(1, Token{ .set = "def" });
    try c.expect_token_at(2, Token{ .literal = "uv" });
    try c.expect_match("duv");
    try c.expect_match("fuv");
    try c.expect_match("Hduv");
    try c.expect_match("HIduv");
    try c.expect_nonmatch("cuv");
    try c.expect_nonmatch("Hcuv");
    try c.expect_nonmatch("HIcuv");
    try c.expect_nonmatch("duvX");
    try c.expect_nonmatch("HduvX");
    try c.expect_nonmatch("HIduvX");
    //   .. question_mark
    c = mkcase("*[d-f]?");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .asterisk = {} });
    try c.expect_token_at(1, Token{ .set = "def" });
    try c.expect_token_at(2, Token{ .question_mark = {} });
    try c.expect_match("dQ");
    try c.expect_match("fQ");
    try c.expect_match("HdQ");
    try c.expect_match("HIdQ");
    try c.expect_nonmatch("cQ");
    try c.expect_nonmatch("HcQ");
    try c.expect_nonmatch("HIcQ");
    try c.expect_nonmatch("dQX");
    try c.expect_nonmatch("HdQX");
    try c.expect_nonmatch("HIdQX");
    //   .. asterisk
    c = mkcase("*[d-f]*");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .asterisk = {} });
    try c.expect_token_at(1, Token{ .set = "def" });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_match("d");
    try c.expect_match("f");
    try c.expect_match("Hd");
    try c.expect_match("HId");
    try c.expect_match("dH");
    try c.expect_match("dHI");
    try c.expect_match("HIdJ");
    try c.expect_match("HIdJK");
    try c.expect_nonmatch("");
    try c.expect_nonmatch("g");
    try c.expect_nonmatch("Hg");
    try c.expect_nonmatch("HIg");
    try c.expect_nonmatch("gH");
    try c.expect_nonmatch("gHI");
    try c.expect_nonmatch("HIgJ");
    try c.expect_nonmatch("HIgJK");
    //   .. set
    c = mkcase("*[d-f][p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .asterisk = {} });
    try c.expect_token_at(1, Token{ .set = "def" });
    try c.expect_token_at(2, Token{ .set = "pqrs" });
    try c.expect_match("dp");
    try c.expect_match("fp");
    try c.expect_match("ds");
    try c.expect_match("fs");
    try c.expect_match("Hdp");
    try c.expect_match("HIdp");
    try c.expect_nonmatch("cp");
    try c.expect_nonmatch("do");
    try c.expect_nonmatch("gt");
    try c.expect_nonmatch("Hcp");
    try c.expect_nonmatch("HIcp");
    try c.expect_nonmatch("dpX");
    //   .. antiset
    c = mkcase("*[d-f][^p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .asterisk = {} });
    try c.expect_token_at(1, Token{ .set = "def" });
    try c.expect_token_at(2, Token{ .antiset = "pqrs" });
    try c.expect_match("do");
    try c.expect_match("fo");
    try c.expect_match("dt");
    try c.expect_match("ft");
    try c.expect_match("Hdo");
    try c.expect_match("HIdo");
    try c.expect_nonmatch("co");
    try c.expect_nonmatch("dp");
    try c.expect_nonmatch("gs");
    try c.expect_nonmatch("Hco");
    try c.expect_nonmatch("HIco");
    try c.expect_nonmatch("doX");
}

test "three-token patterns: asterisk, antiset, .." {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    //   .. literal
    c = mkcase("*[^d-f]uv");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .asterisk = {} });
    try c.expect_token_at(1, Token{ .antiset = "def" });
    try c.expect_token_at(2, Token{ .literal = "uv" });
    try c.expect_match("cuv");
    try c.expect_match("guv");
    try c.expect_match("Hcuv");
    try c.expect_match("HIcuv");
    try c.expect_nonmatch("duv");
    try c.expect_nonmatch("Hduv");
    try c.expect_nonmatch("HIduv");
    try c.expect_nonmatch("cuvX");
    try c.expect_nonmatch("cuvX");
    //   .. question_mark
    c = mkcase("*[^d-f]?");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .asterisk = {} });
    try c.expect_token_at(1, Token{ .antiset = "def" });
    try c.expect_token_at(2, Token{ .question_mark = {} });
    try c.expect_match("cQ");
    try c.expect_match("gQ");
    try c.expect_match("HcQ");
    try c.expect_match("HIcQ");
    try c.expect_nonmatch("dQ");
    try c.expect_nonmatch("fQ");
    try c.expect_nonmatch("HdQ");
    try c.expect_nonmatch("HIdQ");
    //   .. asterisk
    c = mkcase("*[^d-f]*");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .asterisk = {} });
    try c.expect_token_at(1, Token{ .antiset = "def" });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_match("c");
    try c.expect_match("g");
    try c.expect_match("Hc");
    try c.expect_match("HIc");
    try c.expect_match("cH");
    try c.expect_match("cHI");
    try c.expect_match("HIcJ");
    try c.expect_match("HIcJK");
    try c.expect_nonmatch("");
    try c.expect_nonmatch("d");
    //   .. set
    c = mkcase("*[^d-f][p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .asterisk = {} });
    try c.expect_token_at(1, Token{ .antiset = "def" });
    try c.expect_token_at(2, Token{ .set = "pqrs" });
    try c.expect_match("cp");
    try c.expect_match("cs");
    try c.expect_match("gp");
    try c.expect_match("gs");
    try c.expect_match("Hcp");
    try c.expect_match("HIcp");
    try c.expect_nonmatch("co");
    try c.expect_nonmatch("ct");
    try c.expect_nonmatch("ds");
    try c.expect_nonmatch("dt");
    try c.expect_nonmatch("Hco");
    try c.expect_nonmatch("HIco");
    //   .. antiset
    c = mkcase("*[^d-f][^p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .asterisk = {} });
    try c.expect_token_at(1, Token{ .antiset = "def" });
    try c.expect_token_at(2, Token{ .antiset = "pqrs" });
    try c.expect_match("co");
    try c.expect_match("ct");
    try c.expect_match("go");
    try c.expect_match("gt");
    try c.expect_match("Hco");
    try c.expect_match("HIco");
    try c.expect_nonmatch("cp");
    try c.expect_nonmatch("cs");
    try c.expect_nonmatch("dt");
    try c.expect_nonmatch("ds");
    try c.expect_nonmatch("Hcp");
    try c.expect_nonmatch("HIcp");
}

test "three-token patterns: set, literal, .." {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    //   .. literal          # impossible
    //   .. question_mark
    c = mkcase("[a-c]fg?");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .literal = "fg" });
    try c.expect_token_at(2, Token{ .question_mark = {} });
    try c.expect_match("afgQ");
    try c.expect_match("cfgQ");
    try c.expect_nonmatch("afg");
    try c.expect_nonmatch("dfgQ");
    try c.expect_nonmatch("XafgQ");
    try c.expect_nonmatch("afgQX");
    //   .. asterisk
    c = mkcase("[a-c]fg*");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .literal = "fg" });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_match("afg");
    try c.expect_match("cfg");
    try c.expect_match("afgH");
    try c.expect_match("afgHI");
    try c.expect_nonmatch("dfg");
    try c.expect_nonmatch("dfgH");
    try c.expect_nonmatch("dfgHI");
    //   .. set
    c = mkcase("[a-c]fg[p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .literal = "fg" });
    try c.expect_token_at(2, Token{ .set = "pqrs" });
    try c.expect_match("afgp");
    try c.expect_match("afgs");
    try c.expect_match("cfgp");
    try c.expect_match("cfgs");
    try c.expect_nonmatch("afgt");
    try c.expect_nonmatch("dfgp");
    try c.expect_nonmatch("dfgt");
    try c.expect_nonmatch("Xafgp");
    try c.expect_nonmatch("fgafgp");
    //   .. antiset
    c = mkcase("[a-c]fg[^p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .literal = "fg" });
    try c.expect_token_at(2, Token{ .antiset = "pqrs" });
    try c.expect_match("afgo");
    try c.expect_match("afgt");
    try c.expect_match("cfgo");
    try c.expect_match("cfgt");
    try c.expect_nonmatch("afgs");
    try c.expect_nonmatch("dfgo");
    try c.expect_nonmatch("dfgs");
    try c.expect_nonmatch("Xafgo");
    try c.expect_nonmatch("fgafgo");
}

test "three-token patterns: set, question_mark, .." {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    //   .. literal
    c = mkcase("[a-c]?uv");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .literal = "uv" });
    try c.expect_match("aQuv");
    try c.expect_match("cQuv");
    try c.expect_nonmatch("dQuv");
    try c.expect_nonmatch("XaQuv");
    try c.expect_nonmatch("aQuvX");
    try c.expect_nonmatch("XaQuvX");
    //   .. question_mark
    c = mkcase("[a-c]??");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .question_mark = {} });
    try c.expect_match("aQR");
    try c.expect_match("cQR");
    try c.expect_nonmatch("dQR");
    try c.expect_nonmatch("aQRX");
    try c.expect_nonmatch("XaQR");
    try c.expect_nonmatch("XaQRX");
    //   .. asterisk
    c = mkcase("[a-c]?*");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_match("aQ");
    try c.expect_match("cQ");
    try c.expect_match("aQH");
    try c.expect_match("aQHI");
    try c.expect_nonmatch("a");
    try c.expect_nonmatch("dQ");
    try c.expect_nonmatch("dQH");
    try c.expect_nonmatch("dQHI");
    //   .. set
    c = mkcase("[a-c]?[p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .set = "pqrs" });
    try c.expect_match("aQp");
    try c.expect_match("aQs");
    try c.expect_match("cQp");
    try c.expect_match("aQp");
    try c.expect_nonmatch("aQo");
    try c.expect_nonmatch("dQp");
    try c.expect_nonmatch("aQo");
    try c.expect_nonmatch("XaQp");
    try c.expect_nonmatch("aQpX");
    //   .. antiset
    c = mkcase("[a-c]?[^p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .antiset = "pqrs" });
    try c.expect_match("aQo");
    try c.expect_match("aQt");
    try c.expect_match("cQo");
    try c.expect_match("aQo");
    try c.expect_nonmatch("aQp");
    try c.expect_nonmatch("aQs");
    try c.expect_nonmatch("dQo");
    try c.expect_nonmatch("dQs");
    try c.expect_nonmatch("XaQo");
    try c.expect_nonmatch("aQoX");
}

test "three-token patterns: set, asterisk, .." {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    //   .. literal
    c = mkcase("[a-c]*uv");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .asterisk = {} });
    try c.expect_token_at(2, Token{ .literal = "uv" });
    try c.expect_match("auv");
    try c.expect_match("cuv");
    try c.expect_match("aHuv");
    try c.expect_match("aHIuv");
    try c.expect_nonmatch("duv");
    try c.expect_nonmatch("dHuv");
    try c.expect_nonmatch("dHIuv");
    //   .. question_mark
    c = mkcase("[a-c]*?");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_match("aQ");
    try c.expect_match("cQ");
    try c.expect_match("aQH");
    try c.expect_match("aQHI");
    try c.expect_nonmatch("a");
    try c.expect_nonmatch("dQ");
    try c.expect_nonmatch("dQH");
    try c.expect_nonmatch("dQHI");
    //   .. asterisk
    c = mkcase("[a-c]**");
    try c.expect_error(ParsingError.DoubleAsterisk);
    //   .. set
    c = mkcase("[a-c]*[p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .asterisk = {} });
    try c.expect_token_at(2, Token{ .set = "pqrs" });
    try c.expect_match("ap");
    try c.expect_match("as");
    try c.expect_match("cp");
    try c.expect_match("cs");
    try c.expect_match("aHp");
    try c.expect_match("aHIp");
    try c.expect_nonmatch("dp");
    try c.expect_nonmatch("dHp");
    try c.expect_nonmatch("dHIp");
    //   .. antiset
    c = mkcase("[a-c]*[^p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .asterisk = {} });
    try c.expect_token_at(2, Token{ .antiset = "pqrs" });
    try c.expect_match("at");
    try c.expect_match("ao");
    try c.expect_match("ct");
    try c.expect_match("co");
    try c.expect_match("aHt");
    try c.expect_match("aHIt");
    try c.expect_nonmatch("dt");
    try c.expect_nonmatch("dHt");
    try c.expect_nonmatch("dHIt");
}

test "three-token patterns: set, set, .." {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    //   .. literal
    c = mkcase("[a-c][p-s]uv");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .set = "pqrs" });
    try c.expect_token_at(2, Token{ .literal = "uv" });
    try c.expect_match("apuv");
    try c.expect_match("asuv");
    try c.expect_match("cpuv");
    try c.expect_match("csuv");
    try c.expect_nonmatch("aouv");
    try c.expect_nonmatch("atuv");
    try c.expect_nonmatch("douv");
    try c.expect_nonmatch("dtuv");
    try c.expect_nonmatch("Xapuv");
    try c.expect_nonmatch("apuvX");
    //   .. question_mark
    c = mkcase("[a-c][p-s]?");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .set = "pqrs" });
    try c.expect_token_at(2, Token{ .question_mark = {} });
    try c.expect_match("apQ");
    try c.expect_match("asQ");
    try c.expect_match("cpQ");
    try c.expect_match("csQ");
    try c.expect_nonmatch("aoQ");
    try c.expect_nonmatch("atQ");
    try c.expect_nonmatch("doQ");
    try c.expect_nonmatch("dtQ");
    try c.expect_nonmatch("XapQ");
    try c.expect_nonmatch("apQX");
    //   .. asterisk
    c = mkcase("[a-c][p-s]*");
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .set = "pqrs" });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_match("ap");
    try c.expect_match("as");
    try c.expect_match("cp");
    try c.expect_match("cs");
    try c.expect_match("apH");
    try c.expect_match("apHI");
    try c.expect_nonmatch("ao");
    try c.expect_nonmatch("at");
    try c.expect_nonmatch("do");
    try c.expect_nonmatch("dt");
    try c.expect_nonmatch("Xap");
    try c.expect_nonmatch("XapH");
    try c.expect_nonmatch("XapHI");
    //   .. set
    c = mkcase("[a-c][p-s][p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .set = "pqrs" });
    try c.expect_token_at(2, Token{ .set = "pqrs" });
    try c.expect_match("app");
    try c.expect_match("aps");
    try c.expect_match("asp");
    try c.expect_match("ass");
    try c.expect_match("cpp");
    try c.expect_match("cps");
    try c.expect_match("csp");
    try c.expect_match("css");
    try c.expect_nonmatch("apo");
    try c.expect_nonmatch("aop");
    try c.expect_nonmatch("dpp");
    try c.expect_nonmatch("Xapp");
    try c.expect_nonmatch("appX");
    //   .. antiset
    c = mkcase("[a-c][p-s][^p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .set = "pqrs" });
    try c.expect_token_at(2, Token{ .antiset = "pqrs" });
    try c.expect_match("apo");
    try c.expect_match("apt");
    try c.expect_match("aso");
    try c.expect_match("ast");
    try c.expect_match("cpo");
    try c.expect_match("cpt");
    try c.expect_match("cso");
    try c.expect_match("cst");
    try c.expect_nonmatch("app");
    try c.expect_nonmatch("aoo");
    try c.expect_nonmatch("dpo");
    try c.expect_nonmatch("Xapo");
    try c.expect_nonmatch("apoX");
}

test "three-token patterns: set, antiset, .." {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    //   .. literal
    c = mkcase("[a-c][^p-s]uv");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .antiset = "pqrs" });
    try c.expect_token_at(2, Token{ .literal = "uv" });
    try c.expect_match("aouv");
    try c.expect_match("atuv");
    try c.expect_match("couv");
    try c.expect_match("ctuv");
    try c.expect_nonmatch("apuv");
    try c.expect_nonmatch("douv");
    try c.expect_nonmatch("dpuv");
    try c.expect_nonmatch("Xaouv");
    try c.expect_nonmatch("aouvX");
    //   .. question_mark
    c = mkcase("[a-c][^p-s]?");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .antiset = "pqrs" });
    try c.expect_token_at(2, Token{ .question_mark = {} });
    try c.expect_match("aoQ");
    try c.expect_match("atQ");
    try c.expect_match("coQ");
    try c.expect_match("ctQ");
    try c.expect_nonmatch("apQ");
    try c.expect_nonmatch("doQ");
    try c.expect_nonmatch("dpQ");
    try c.expect_nonmatch("XaoQ");
    try c.expect_nonmatch("aoQX");
    //   .. asterisk
    c = mkcase("[a-c][^p-s]*");
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .antiset = "pqrs" });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_match("ao");
    try c.expect_match("at");
    try c.expect_match("co");
    try c.expect_match("ct");
    try c.expect_match("aoH");
    try c.expect_match("aoHI");
    try c.expect_nonmatch("ap");
    try c.expect_nonmatch("do");
    try c.expect_nonmatch("dp");
    try c.expect_nonmatch("apH");
    try c.expect_nonmatch("apHI");
    try c.expect_nonmatch("Xao");
    try c.expect_nonmatch("XaoHI");
    //   .. set
    c = mkcase("[a-c][^p-s][p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .antiset = "pqrs" });
    try c.expect_token_at(2, Token{ .set = "pqrs" });
    try c.expect_match("aop");
    try c.expect_match("atp");
    try c.expect_match("aos");
    try c.expect_match("ats");
    try c.expect_match("cop");
    try c.expect_match("ctp");
    try c.expect_match("cos");
    try c.expect_match("cts");
    try c.expect_nonmatch("app");
    try c.expect_nonmatch("aoo");
    try c.expect_nonmatch("dop");
    try c.expect_nonmatch("Xaop");
    try c.expect_nonmatch("aopX");
    //   .. antiset
    c = mkcase("[a-c][^p-s][^p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .set = "abc" });
    try c.expect_token_at(1, Token{ .antiset = "pqrs" });
    try c.expect_token_at(2, Token{ .antiset = "pqrs" });
    try c.expect_match("aoo");
    try c.expect_match("ato");
    try c.expect_match("aot");
    try c.expect_match("att");
    try c.expect_match("coo");
    try c.expect_match("cto");
    try c.expect_match("cot");
    try c.expect_match("ctt");
    try c.expect_nonmatch("aop");
    try c.expect_nonmatch("apo");
    try c.expect_nonmatch("doo");
    try c.expect_nonmatch("Xaoo");
    try c.expect_nonmatch("aooX");
}

test "three-token patterns: antiset, literal, .." {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    //   .. literal          # impossible
    //   .. question_mark
    c = mkcase("[^a-c]fg?");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .literal = "fg" });
    try c.expect_token_at(2, Token{ .question_mark = {} });
    try c.expect_match("_fgQ");
    try c.expect_match("dfgQ");
    try c.expect_nonmatch("_fg");
    try c.expect_nonmatch("afgQ");
    try c.expect_nonmatch("cfgQ");
    try c.expect_nonmatch("X_fgQ");
    try c.expect_nonmatch("_fgQX");
    //   .. asterisk
    c = mkcase("[^a-c]fg*");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .literal = "fg" });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_match("_fg");
    try c.expect_match("dfg");
    try c.expect_match("_fgH");
    try c.expect_match("_fgHI");
    try c.expect_nonmatch("afg");
    try c.expect_nonmatch("afgH");
    try c.expect_nonmatch("afgHI");
    //   .. set
    c = mkcase("[^a-c]fg[p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .literal = "fg" });
    try c.expect_token_at(2, Token{ .set = "pqrs" });
    try c.expect_match("_fgp");
    try c.expect_match("_fgs");
    try c.expect_match("dfgp");
    try c.expect_match("dfgs");
    try c.expect_nonmatch("_fgt");
    try c.expect_nonmatch("afgp");
    try c.expect_nonmatch("afgt");
    try c.expect_nonmatch("X_fgp");
    try c.expect_nonmatch("fg_fgp");
    //   .. antiset
    c = mkcase("[^a-c]fg[^p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .literal = "fg" });
    try c.expect_token_at(2, Token{ .antiset = "pqrs" });
    try c.expect_match("_fgo");
    try c.expect_match("_fgt");
    try c.expect_match("dfgo");
    try c.expect_match("dfgt");
    try c.expect_nonmatch("_fgs");
    try c.expect_nonmatch("cfgo");
    try c.expect_nonmatch("cfgs");
    try c.expect_nonmatch("X_fgo");
    try c.expect_nonmatch("fg_fgo");
}

test "three-token patterns: antiset, question_mark, .." {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    //   .. literal
    c = mkcase("[^a-c]?uv");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .literal = "uv" });
    try c.expect_match("_Quv");
    try c.expect_match("dQuv");
    try c.expect_nonmatch("_");
    try c.expect_nonmatch("aQuv");
    try c.expect_nonmatch("cQuv");
    try c.expect_nonmatch("X_Quv");
    try c.expect_nonmatch("_QuvX");
    //   .. question_mark
    c = mkcase("[^a-c]??");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .question_mark = {} });
    try c.expect_match("_QR");
    try c.expect_match("dQR");
    try c.expect_nonmatch("_");
    try c.expect_nonmatch("aQR");
    try c.expect_nonmatch("cQR");
    try c.expect_nonmatch("X_QR");
    try c.expect_nonmatch("_QRX");
    //   .. asterisk
    c = mkcase("[^a-c]?*");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_match("_Q");
    try c.expect_match("dQ");
    try c.expect_match("_QH");
    try c.expect_match("_QHI");
    try c.expect_nonmatch("_");
    try c.expect_nonmatch("aQ");
    try c.expect_nonmatch("aQH");
    try c.expect_nonmatch("aQHI");
    //   .. set
    c = mkcase("[^a-c]?[p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .set = "pqrs" });
    try c.expect_match("_Qp");
    try c.expect_match("_Qs");
    try c.expect_match("dQp");
    try c.expect_match("dQs");
    try c.expect_nonmatch("_Qt");
    try c.expect_nonmatch("aQp");
    try c.expect_nonmatch("aQt");
    try c.expect_nonmatch("X_Qp");
    try c.expect_nonmatch("_QpX");
    //   .. antiset
    c = mkcase("[^a-c]?[^p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .antiset = "pqrs" });
    try c.expect_match("_Qo");
    try c.expect_match("_Qt");
    try c.expect_match("dQo");
    try c.expect_match("dQt");
    try c.expect_nonmatch("_Qp");
    try c.expect_nonmatch("aQt");
    try c.expect_nonmatch("X_Qo");
    try c.expect_nonmatch("_QoX");
}

test "three-token patterns: antiset, asterisk, .." {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    //   .. literal
    c = mkcase("[^a-c]*uv");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .asterisk = {} });
    try c.expect_token_at(2, Token{ .literal = "uv" });
    try c.expect_match("_uv");
    try c.expect_match("duv");
    try c.expect_match("_Huv");
    try c.expect_match("_HIuv");
    try c.expect_nonmatch("auv");
    try c.expect_nonmatch("aHuv");
    try c.expect_nonmatch("aHIuv");
    //   .. question_mark
    c = mkcase("[^a-c]*?");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .question_mark = {} });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_match("_Q");
    try c.expect_match("dQ");
    try c.expect_match("_HQ");
    try c.expect_match("_HIQ");
    try c.expect_nonmatch("aQ");
    try c.expect_nonmatch("aHQ");
    try c.expect_nonmatch("aHIQ");
    //   .. asterisk
    c = mkcase("[^a-c]**");
    try c.expect_error(ParsingError.DoubleAsterisk);
    //   .. set
    c = mkcase("[^a-c]*[p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .asterisk = {} });
    try c.expect_token_at(2, Token{ .set = "pqrs" });
    try c.expect_match("_p");
    try c.expect_match("_s");
    try c.expect_match("dp");
    try c.expect_match("ds");
    try c.expect_match("_Hp");
    try c.expect_match("_HIp");
    try c.expect_nonmatch("_o");
    try c.expect_nonmatch("_Ho");
    try c.expect_nonmatch("_HIo");
    try c.expect_nonmatch("_pX");
    //   .. antiset
    c = mkcase("[^a-c]*[^p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .asterisk = {} });
    try c.expect_token_at(2, Token{ .antiset = "pqrs" });
    try c.expect_match("_o");
    try c.expect_match("_t");
    try c.expect_match("do");
    try c.expect_match("dt");
    try c.expect_match("_Ho");
    try c.expect_match("_HIo");
    try c.expect_nonmatch("_p");
    try c.expect_nonmatch("_Hp");
    try c.expect_nonmatch("_HIp");
    try c.expect_nonmatch("_pp");
}

test "three-token patterns: antiset, set, .." {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    //   .. literal
    c = mkcase("[^a-c][p-s]uv");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .set = "pqrs" });
    try c.expect_token_at(2, Token{ .literal = "uv" });
    try c.expect_match("_puv");
    try c.expect_match("_suv");
    try c.expect_match("dpuv");
    try c.expect_match("dsuv");
    try c.expect_nonmatch("_ouv");
    try c.expect_nonmatch("_tuv");
    try c.expect_nonmatch("douv");
    try c.expect_nonmatch("dtuv");
    try c.expect_nonmatch("X_puv");
    try c.expect_nonmatch("_puvX");
    //   .. question_mark
    c = mkcase("[^a-c][p-s]?");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .set = "pqrs" });
    try c.expect_token_at(2, Token{ .question_mark = {} });
    try c.expect_match("_pQ");
    try c.expect_match("_sQ");
    try c.expect_match("dpQ");
    try c.expect_match("dsQ");
    try c.expect_nonmatch("_oQ");
    try c.expect_nonmatch("_tQ");
    try c.expect_nonmatch("doQ");
    try c.expect_nonmatch("dtQ");
    try c.expect_nonmatch("X_pQ");
    try c.expect_nonmatch("_pQX");
    //   .. asterisk
    c = mkcase("[^a-c][p-s]*");
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .set = "pqrs" });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_match("_p");
    try c.expect_match("_s");
    try c.expect_match("dp");
    try c.expect_match("ds");
    try c.expect_match("_pH");
    try c.expect_match("_pHI");
    try c.expect_nonmatch("_o");
    try c.expect_nonmatch("cp");
    try c.expect_nonmatch("co");
    try c.expect_nonmatch("X_p");
    try c.expect_nonmatch("X_pH");
    try c.expect_nonmatch("X_pHI");
    //   .. set
    c = mkcase("[^a-c][p-s][p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .set = "pqrs" });
    try c.expect_token_at(2, Token{ .set = "pqrs" });
    try c.expect_match("_pp");
    try c.expect_match("_ps");
    try c.expect_match("_sp");
    try c.expect_match("_ss");
    try c.expect_match("dpp");
    try c.expect_match("dps");
    try c.expect_match("dsp");
    try c.expect_match("dss");
    try c.expect_nonmatch("_po");
    try c.expect_nonmatch("_op");
    try c.expect_nonmatch("cpp");
    try c.expect_nonmatch("X_pp");
    try c.expect_nonmatch("_ppX");
    //   .. antiset
    c = mkcase("[^a-c][p-s][^p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .set = "pqrs" });
    try c.expect_token_at(2, Token{ .antiset = "pqrs" });
    try c.expect_match("_po");
    try c.expect_match("_pt");
    try c.expect_match("_so");
    try c.expect_match("_st");
    try c.expect_match("dpo");
    try c.expect_match("dpt");
    try c.expect_match("dso");
    try c.expect_match("dst");
    try c.expect_nonmatch("_pp");
    try c.expect_nonmatch("_oo");
    try c.expect_nonmatch("cpo");
    try c.expect_nonmatch("X_po");
    try c.expect_nonmatch("_poX");
}

test "three-token patterns: antiset, antiset, .." {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    //   .. literal
    c = mkcase("[^a-c][^p-s]uv");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .antiset = "pqrs" });
    try c.expect_token_at(2, Token{ .literal = "uv" });
    try c.expect_match("_ouv");
    try c.expect_match("_tuv");
    try c.expect_match("douv");
    try c.expect_match("dtuv");
    try c.expect_nonmatch("_puv");
    try c.expect_nonmatch("_suv");
    try c.expect_nonmatch("dpuv");
    try c.expect_nonmatch("dsuv");
    try c.expect_nonmatch("X_ouv");
    try c.expect_nonmatch("_ouvX");
    //   .. question_mark
    c = mkcase("[^a-c][^p-s]?");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .antiset = "pqrs" });
    try c.expect_token_at(2, Token{ .question_mark = {} });
    try c.expect_match("_oQ");
    try c.expect_match("_tQ");
    try c.expect_match("doQ");
    try c.expect_match("dtQ");
    try c.expect_nonmatch("_pQ");
    try c.expect_nonmatch("_sQ");
    try c.expect_nonmatch("dpQ");
    try c.expect_nonmatch("dsQ");
    try c.expect_nonmatch("X_oQ");
    try c.expect_nonmatch("_oQX");
    //   .. asterisk
    c = mkcase("[^a-c][^p-s]*");
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .antiset = "pqrs" });
    try c.expect_token_at(2, Token{ .asterisk = {} });
    try c.expect_match("_o");
    try c.expect_match("_t");
    try c.expect_match("do");
    try c.expect_match("dt");
    try c.expect_match("_oH");
    try c.expect_match("_oHI");
    try c.expect_nonmatch("_p");
    try c.expect_nonmatch("co");
    try c.expect_nonmatch("cp");
    //   .. set
    c = mkcase("[^a-c][^p-s][p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .antiset = "pqrs" });
    try c.expect_token_at(2, Token{ .set = "pqrs" });
    try c.expect_match("_op");
    try c.expect_match("_os");
    try c.expect_match("_tp");
    try c.expect_match("_ts");
    try c.expect_match("dop");
    try c.expect_match("dos");
    try c.expect_match("dtp");
    try c.expect_match("dts");
    try c.expect_nonmatch("_oo");
    try c.expect_nonmatch("_pp");
    try c.expect_nonmatch("cop");
    try c.expect_nonmatch("X_op");
    try c.expect_nonmatch("_opX");
    //   .. antiset
    c = mkcase("[^a-c][^p-s][^p-s]");
    try c.expect_len(3);
    try c.expect_token_at(0, Token{ .antiset = "abc" });
    try c.expect_token_at(1, Token{ .antiset = "pqrs" });
    try c.expect_token_at(2, Token{ .antiset = "pqrs" });
    try c.expect_match("_oo");
    try c.expect_match("_ot");
    try c.expect_match("_to");
    try c.expect_match("_tt");
    try c.expect_match("doo");
    try c.expect_match("dot");
    try c.expect_match("dto");
    try c.expect_match("dtt");
    try c.expect_nonmatch("_op");
    try c.expect_nonmatch("_po");
    try c.expect_nonmatch("coo");
    try c.expect_nonmatch("X_oo");
    try c.expect_nonmatch("_ooX");
}

test "stress from swtch.com few more interesting cases" {
    // see https://research.swtch.com/glob
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    c = mkcase("a*a*a*a*a*a*a*a*b");
    try c.expect_nonmatch("a" ** 20);
    // try c.expect_nonmatch("a" ** 40); FIXME + enable: heavy mem leak/OOM and really slow
    try c.expect_match("a" ** 20 ++ "b");
}

test "few more interesting cases" {
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    c = mkcase("");
    try c.expect_error(ParsingError.PrematureEnd);
    c = mkcase("foo*bar*bar");
    try c.expect_match("foobarbar");
    try c.expect_match("foobarHbar");
    try c.expect_match("foobarHIbar");
    try c.expect_match("foobarXbarXbar");
    try c.expect_match("fooHbarHbar");
    try c.expect_match("fooHbarXbarXbar");
    try c.expect_match("fooXbarXbarHbar");
    try c.expect_match("fooXbarXbarXbarXbar");
}

test "redis KEYS examples" {
    // from https://redis.io/docs/latest/commands/keys/
    //
    // > Supported glob-style patterns:
    // >  *  h?llo matches hello, hallo and hxllo
    // >  *  h*llo matches hllo and heeeello
    // >  *  h[ae]llo matches hello and hallo, but not hillo
    // >  *  h[^e]llo matches hallo, hbllo, ... but not hello
    // >  *  h[a-b]llo matches hallo and hbllo
    //
    const mkcase = TestCase.init;
    var c: TestCase = undefined;
    c = mkcase("h?llo");
    try c.expect_match("hello");
    try c.expect_match("hallo");
    try c.expect_match("hxllo");
    c = mkcase("h*llo");
    try c.expect_match("hllo");
    try c.expect_match("heeeello");
    c = mkcase("h[ae]llo");
    try c.expect_match("hello");
    try c.expect_match("hallo");
    try c.expect_nonmatch("hillo");
    c = mkcase("h[^e]llo");
    try c.expect_match("hallo");
    try c.expect_match("hbllo");
    try c.expect_nonmatch("hello");
    c = mkcase("h[a-b]llo");
    try c.expect_match("hallo");
    try c.expect_match("hbllo");
}

const DebugHelper = struct {
    enabled: bool,
    depth: usize,
    value_buff: ?*std.ArrayList(u8),
    loc: []const u8,
    fn init(comptime loc: []const u8, depth: usize) DebugHelper {
        const al_ptr = std.heap.page_allocator.create(std.ArrayList(u8)) catch return .{
            .loc = loc,
            .enabled = false,
            .depth = depth,
            .value_buff = null,
        };
        al_ptr.* = std.ArrayList(u8).init(std.heap.page_allocator);
        return .{
            .loc = loc,
            .enabled = DEBUG,
            .depth = depth,
            .value_buff = al_ptr,
        };
    }
    fn _add_byte(self: *DebugHelper, byte: u8) void {
        if (!self.enabled) return;
        const value_buff = self.value_buff orelse return;
        var w = value_buff.writer();
        w.writeByte(byte) catch return;
    }
    fn _add_fmt(self: *DebugHelper, comptime fmt: []const u8, args: anytype) void {
        if (!self.enabled) return;
        const value_buff = self.value_buff orelse return;
        var w = value_buff.writer();
        w.print(fmt, args) catch return;
    }
    fn _add_slice(self: *DebugHelper, bytes: []const u8) void {
        if (!self.enabled) return;
        const value_buff = self.value_buff orelse return;
        var w = value_buff.writer();
        w.writeAll(bytes) catch return;
    }
    fn _flush(self: DebugHelper, comptime vname: ?[]const u8) void {
        if (!self.enabled) return;
        const value_buff = self.value_buff orelse return;
        var al = std.ArrayList(u8).init(std.heap.page_allocator);
        defer al.deinit();
        al.appendSlice(self.loc) catch return;
        al.append(':') catch return;
        for (0..self.depth) |_| al.appendSlice("> ") catch return;
        if (vname) |n| {
            al.appendSlice(n) catch return;
            al.append('=') catch return;
        }
        al.appendSlice(value_buff.items) catch return;
        value_buff.shrinkAndFree(0);
        std.log.warn("{s}", .{al.items});
    }
    fn show_message(self: *DebugHelper, msg: []const u8) void {
        self._add_slice(msg);
        self._flush(null);
    }
    fn show_fmt_message(self: *DebugHelper, comptime fmt: []const u8, args: anytype) void {
        self._add_fmt(fmt, args);
        self._flush(null);
    }
    fn show_int(self: *DebugHelper, comptime vname: []const u8, value: usize) void {
        if (!self.enabled) return;
        self._add_fmt("\"{d}\"", .{value});
        self._flush(vname);
    }
    fn show_range(self: *DebugHelper, comptime vname: []const u8, value: PartialResult.Range) void {
        if (!self.enabled) return;
        self._add_fmt("{{.bgn={d},.end={d}}}", .{ value.bgn, value.end });
        self._flush(vname);
    }
    fn show_str(self: *DebugHelper, comptime vname: []const u8, value: []const u8) void {
        if (!self.enabled) return;
        self._add_fmt("\"{s}\"", .{value});
        self._flush(vname);
    }
    fn show_token(self: *DebugHelper, comptime vname: []const u8, value: Token) void {
        if (!self.enabled) return;
        self._add_slice("~|");
        switch (value) {
            .literal => |t| self._add_slice(t),
            .set => |t| self._add_fmt("[{c}-{c}]", .{ t[0], t[t.len - 1] }),
            .antiset => |t| self._add_fmt("[^{c}-{c}]", .{ t[0], t[t.len - 1] }),
            .question_mark => self._add_byte('?'),
            .asterisk => self._add_byte('*'),
        }
        self._add_slice("|");
        self._flush(vname);
    }
    fn show_tokens(self: *DebugHelper, comptime vname: []const u8, items: []const Token) void {
        if (!self.enabled) return;
        self._add_slice("~|");
        for (items) |item| {
            switch (item) {
                .literal => |t| self._add_slice(t),
                .set => |t| self._add_fmt("[{c}-{c}]", .{ t[0], t[t.len - 1] }),
                .antiset => |t| self._add_fmt("[^{c}-{c}]", .{ t[0], t[t.len - 1] }),
                .question_mark => self._add_byte('?'),
                .asterisk => self._add_byte('*'),
            }
        }
        self._add_slice("|");
        self._flush(vname);
    }
    fn show_partial_result(self: *DebugHelper, comptime vname: []const u8, value: PartialResult) void {
        if (!self.enabled) return;
        self._add_slice(".{.hungry_tokens=");
        self._add_slice("~|");
        for (value.hungry_tokens) |tok| {
            switch (tok) {
                .literal => |t| self._add_slice(t),
                .set => |t| self._add_fmt("[{c}-{c}]", .{ t[0], t[t.len - 1] }),
                .antiset => |t| self._add_fmt("[^{c}-{c}]", .{ t[0], t[t.len - 1] }),
                .question_mark => self._add_byte('?'),
                .asterisk => self._add_byte('*'),
            }
        }
        self._add_fmt("|,matched={d},to_match={d},is_fail={}}}", .{ value.matched, value.to_match, value.is_fail });
        self._flush(vname);
    }
};
