const std = @import("std");
const net = std.net;

const util = @import("util.zig");
const redis = @import("redis.zig");

const _KB = 1024;
const _MB = 1024 * _KB;

const HELP_TEXT =
    \\usage: codecrafters-redis [options]
    \\
    \\options:
    \\  --address ADDR     Bind to IP address ADDR.
    \\  --port PORT        Bind to port PORT.
    \\  --command-timeout SECONDS
    \\                     Drop command connection if a peer does not respond under
    \\                     SECONDS.   SECONDS must be decimal with up to millisecond
    \\                     precision. Default: 10 seconds.
    \\  --max-threads THREAD_COUNT
    \\                     Don't create more than THREAD_COUNT threads.  Shared for
    \\                     command threads as well as replication threads.
    \\                     Default: use as many threads as available on CPU.
    \\  --max-replicas REPLICA_COUNT
    \\                     Don't accept more than REPLICA_COUNT replicas.  This must
    \\                     at least one lower than --max-threads, otherwise
    \\                     no commands can be processed.  Default: half of --max-threads,
    \\                     rounded up to integer value.
    \\                     This option is ignored in replica mode.
    \\  --replica-timeout SECONDS
    \\                     Drop replica connection if it does not respond under
    \\                     SECONDS.   Applies to replicas for master and vice
    \\                     versa.  SECONDS must be decimal with up to millisecond
    \\                     precision. Default: 10 seconds.
    \\  --replicaof CONN   Run as replica of CONN; CONN must be in form
    \\                     of 'ADDR PORT'; that is, IP address or hostname
    \\                     followed by space and port number.
    \\  --dir DIRPATH      Store RDB files in DIRPATH.
    \\  --dbfilename NAME  RDB file name.
    \\  --debug            Enable debugging output.
    \\  --logdir DIRPATH   Store logs in directory DIRPATH.  This will create separate
    \\                     files for server and each individual task (connection).
    \\                     (Default: logging to files is disabled).
    \\  --verbose          Enable verbose output.
    \\
    \\
;

const ArgParsingResult = union(enum) {
    options: AppOptions,
    usage_error: []const u8,

    fn unknown_arg(arg: []const u8) ArgParsingResult {
        var arr: [100]u8 = undefined;
        const msg = std.fmt.bufPrint(@as([]u8, &arr), "unknown argument: {s}", .{arg}) catch "unknown argument";
        return ArgParsingResult{ .usage_error = msg };
    }
};

const AppOptions = struct {
    address: []const u8 = "127.0.0.1",
    port: u16 = 6379,
    debug: bool = false,
    verbose: bool = false,
    logdir: ?[]const u8 = null,
    thread_buff_size: usize = 3 * _MB,
    server_buff_size: usize = 10 * _MB,
    replica_buff_size: usize = 640 * _KB,
    replica_timeout_ms: u64 = 10_000,
    command_timeout_ms: u64 = 10_000,
    max_dump_bytes: usize = 10 * _MB,
    max_threads: ?usize = null,
    max_replicas: ?usize = null,
    dir: ?[]const u8 = null,
    dbfilename: []const u8 = "dump.rdb",
    redis_replicaof_host: ?[]const u8 = null,
    redis_replicaof_port: u16 = 6379,

    const ConnInfo = struct {
        addr: []const u8,
        port: u16,
        fn parse(text: []const u8) !ConnInfo {
            if (std.mem.count(u8, text, " ") != 1) return error.InvalidConnStr;
            const space_idx = std.mem.indexOf(u8, text, " ") orelse unreachable;
            return ConnInfo{
                .addr = text[0..space_idx],
                .port = std.fmt.parseInt(u16, text[space_idx + 1 ..], 10) catch return error.InvalidConnStr,
            };
        }
    };

    fn init() AppOptions {
        const parsing_result = AppOptions.parse_args();
        switch (parsing_result) {
            .options => return parsing_result.options,
            .usage_error => {
                _ = std.io.getStdErr().write(HELP_TEXT) catch {};
                _ = std.io.getStdErr().write(parsing_result.usage_error) catch {};
                _ = std.io.getStdErr().write("\n") catch {};
                std.process.exit(2);
            },
        }
    }

    fn parse_args() ArgParsingResult {
        var options = AppOptions{};
        var this: []const u8 = undefined;
        var it = std.process.args();
        var param: ?[]const u8 = null;
        _ = it.next(); // drop program name
        while (true) {
            this = @as(?[]const u8, it.next()) orelse break;
            if (std.mem.eql(u8, this, "--address")) {
                param = @as(?[]const u8, it.next());
                options.address = param orelse return ArgParsingResult{ .usage_error = "no ADDRESS?" };
            } else if (std.mem.eql(u8, this, "--port")) {
                param = @as(?[]const u8, it.next());
                options.port = std.fmt.parseInt(u16, param orelse return ArgParsingResult{ .usage_error = "no PORT?" }, 10) catch {
                    return ArgParsingResult{ .usage_error = "PORT must be integer" };
                };
            } else if (std.mem.eql(u8, this, "--replicaof")) {
                param = @as(?[]const u8, it.next());
                const conn_info = ConnInfo.parse(param orelse return ArgParsingResult{ .usage_error = "no CONN?" }) catch {
                    return ArgParsingResult{ .usage_error = "invalid CONN (must be '<ADDR> <PORT>')" };
                };
                options.redis_replicaof_host = conn_info.addr;
                options.redis_replicaof_port = conn_info.port;
            } else if (std.mem.eql(u8, this, "--thread-buff-size")) {
                param = @as(?[]const u8, it.next());
                options.thread_buff_size = std.fmt.parseInt(usize, param orelse return ArgParsingResult{ .usage_error = "no BYTES?" }, 10) catch {
                    return ArgParsingResult{ .usage_error = "BYTES must be integer" };
                };
            } else if (std.mem.eql(u8, this, "--server-buff-size")) {
                param = @as(?[]const u8, it.next());
                options.server_buff_size = std.fmt.parseInt(usize, param orelse return ArgParsingResult{ .usage_error = "no BYTES?" }, 10) catch {
                    return ArgParsingResult{ .usage_error = "BYTES must be integer" };
                };
            } else if (std.mem.eql(u8, this, "--max-dump-bytes")) {
                param = @as(?[]const u8, it.next());
                options.max_dump_bytes = std.fmt.parseInt(usize, param orelse return ArgParsingResult{ .usage_error = "no BYTES?" }, 10) catch {
                    return ArgParsingResult{ .usage_error = "BYTES must be integer" };
                };
            } else if (std.mem.eql(u8, this, "--max-threads")) {
                param = @as(?[]const u8, it.next());
                options.max_threads = std.fmt.parseInt(usize, param orelse return ArgParsingResult{ .usage_error = "no THREAD_COUNT?" }, 10) catch {
                    return ArgParsingResult{ .usage_error = "THREAD_COUNT must be integer" };
                };
                if (options.max_threads == 0) {
                    return ArgParsingResult{ .usage_error = "THREAD_COUNT must be greater than 0" };
                }
            } else if (std.mem.eql(u8, this, "--max-replicas")) {
                param = @as(?[]const u8, it.next());
                options.max_replicas = std.fmt.parseInt(usize, param orelse return ArgParsingResult{ .usage_error = "no REPLICA_COUNT?" }, 10) catch {
                    return ArgParsingResult{ .usage_error = "REPLICA_COUNT must be integer" };
                };
            } else if (std.mem.eql(u8, this, "--replica-timeout")) {
                param = @as(?[]const u8, it.next());
                const replica_timeout_s = std.fmt.parseFloat(f32, param orelse return ArgParsingResult{ .usage_error = "no SECONDS?" }) catch {
                    return ArgParsingResult{ .usage_error = "SECONDS must be decimal" };
                };
                options.replica_timeout_ms = @intFromFloat(replica_timeout_s * std.time.ms_per_s);
            } else if (std.mem.eql(u8, this, "--command-timeout")) {
                param = @as(?[]const u8, it.next());
                const command_timeout_s = std.fmt.parseFloat(f32, param orelse return ArgParsingResult{ .usage_error = "no SECONDS?" }) catch {
                    return ArgParsingResult{ .usage_error = "SECONDS must be decimal" };
                };
                options.command_timeout_ms = @intFromFloat(command_timeout_s * std.time.ms_per_s);
            } else if (std.mem.eql(u8, this, "--dir")) {
                param = @as(?[]const u8, it.next());
                options.dir = param orelse return ArgParsingResult{ .usage_error = "no DIRPATH?" };
            } else if (std.mem.eql(u8, this, "--dbfilename")) {
                param = @as(?[]const u8, it.next());
                options.dbfilename = param orelse return ArgParsingResult{ .usage_error = "no NAME?" };
            } else if (std.mem.eql(u8, this, "--debug")) {
                options.debug = true;
            } else if (std.mem.eql(u8, this, "--verbose")) {
                options.verbose = true;
            } else if (std.mem.eql(u8, this, "--logdir")) {
                param = @as(?[]const u8, it.next());
                options.logdir = param orelse return ArgParsingResult{ .usage_error = "no DIRPATH?" };
            } else {
                return ArgParsingResult.unknown_arg(this);
            }
        }
        return ArgParsingResult{ .options = options };
    }
};

const Memory = struct {
    allocator: std.mem.Allocator,
    server_buff: []u8,
    rdbload_buff: []u8,
    threads_buff: []u8,
    replicas_buff: []u8,
    server_fba: std.heap.FixedBufferAllocator,
    rdbload_fba: std.heap.FixedBufferAllocator,
    threads_fba: std.heap.FixedBufferAllocator,
    replicas_fba: std.heap.FixedBufferAllocator,

    fn create(options: AppOptions, thread_count: usize, replica_count: usize, allocator: std.mem.Allocator) !Memory {
        const server_buff = try allocator.alloc(u8, options.server_buff_size);
        const threads_buff_size = try std.math.mul(usize, options.thread_buff_size, thread_count);
        const threads_buff = try allocator.alloc(u8, threads_buff_size);
        const replicas_buff_size = try std.math.mul(usize, options.replica_buff_size, replica_count);
        const replicas_buff = try allocator.alloc(u8, replicas_buff_size);
        const rdbload_buff_size = try std.math.mul(usize, options.max_dump_bytes, 2);
        const rdbload_buff = try allocator.alloc(u8, rdbload_buff_size);
        return Memory{
            .allocator = allocator,
            .server_buff = server_buff,
            .server_fba = std.heap.FixedBufferAllocator.init(server_buff),
            .threads_buff = threads_buff,
            .threads_fba = std.heap.FixedBufferAllocator.init(threads_buff),
            .replicas_buff = replicas_buff,
            .replicas_fba = std.heap.FixedBufferAllocator.init(replicas_buff),
            .rdbload_buff = rdbload_buff,
            .rdbload_fba = std.heap.FixedBufferAllocator.init(rdbload_buff),
        };
    }

    fn destroy(self: *Memory) void {
        self.server_fba.reset();
        self.allocator.free(self.server_buff);
        self.threads_fba.reset();
        self.replicas_fba.reset();
        self.allocator.free(self.threads_buff);
        self.rdbload_fba.reset();
        self.allocator.free(self.rdbload_buff);
    }
};

pub fn main() !void {
    const app_options = AppOptions.init();
    if (app_options.verbose) {
        std.log.info("PID: {}", .{std.os.linux.getpid()});
        std.log.info("getCpuCount: {}", .{try std.Thread.getCpuCount()});
        std.log.info("app_options.address: {s}", .{app_options.address});
        std.log.info("app_options.port: {d}", .{app_options.port});
        std.log.info("app_options.debug: {}", .{app_options.debug});
        std.log.info("app_options.verbose: {}", .{app_options.verbose});
        std.log.info("app_options.logdir: {?s}", .{app_options.logdir});
        std.log.info("app_options.thread_buff_size: {d}", .{app_options.thread_buff_size});
        std.log.info("app_options.server_buff_size: {d}", .{app_options.server_buff_size});
        std.log.info("app_options.replica_buff_size: {d}", .{app_options.replica_buff_size});
        std.log.info("app_options.replica_timeout_ms: {d}", .{app_options.replica_timeout_ms});
        std.log.info("app_options.command_timeout_s: {d}", .{app_options.command_timeout_ms});
        std.log.info("app_options.max_dump_bytes: {d}", .{app_options.max_dump_bytes});
        std.log.info("app_options.max_threads: {?d}", .{app_options.max_threads});
        std.log.info("app_options.max_replicas: {?d}", .{app_options.max_replicas});
        std.log.info("app_options.dir: {?s}", .{app_options.dir});
        std.log.info("app_options.dbfilename: {s}", .{app_options.dbfilename});
        std.log.info("app_options.redis_replicaof_host: {?s}", .{app_options.redis_replicaof_host});
        std.log.info("app_options.redis_replicaof_port: {d}", .{app_options.redis_replicaof_port});
    }

    const thread_count = app_options.max_threads orelse @as(usize, @max(1, try std.Thread.getCpuCount()));
    const replica_count = counting: {
        if (app_options.redis_replicaof_host != null) break :counting 1;
        if (app_options.max_replicas != null) break :counting app_options.max_replicas.?;
        if (app_options.redis_replicaof_host != null and app_options.max_replicas != null) {
            std.log.warn("ignoring --max-replicas in replica mode", .{});
        }
        break :counting (thread_count + 1) / 2;
    };

    var gpa = std.heap.GeneralPurposeAllocator(.{}){};
    var memory = try Memory.create(app_options, thread_count, replica_count, gpa.allocator());
    defer memory.destroy();

    const address = try net.Address.resolveIp(app_options.address, app_options.port);
    var listener = try address.listen(.{
        .reuse_address = true,
    });
    defer listener.deinit();

    var server = try redis.server.Server.create(
        redis.server.Config{
            .dir = app_options.dir,
            .dbfilename = app_options.dbfilename,
            .max_dump_bytes = app_options.max_dump_bytes,
            .replication_host = app_options.redis_replicaof_host,
            .replication_port = app_options.redis_replicaof_port,
            .replica_timeout_ms = app_options.replica_timeout_ms,
            .command_timeout_ms = app_options.command_timeout_ms,
            .thread_count = thread_count,
            .replica_count = replica_count,
            .replica_buff_size = app_options.replica_buff_size,
            .port = app_options.port,
            .debug = app_options.debug,
            .verbose = app_options.verbose,
            .logdir = app_options.logdir,
        },
        memory.threads_buff,
        memory.replicas_buff,
        memory.server_fba.allocator(),
    );
    defer server.destroy();

    try server.maybe_load_dump(memory.rdbload_fba.allocator());
    try server.serve(&listener);
}

test {
    std.testing.refAllDecls(@This());
}
