const std = @import("std");

const glob = @import("../glob.zig");
const util = @import("../util.zig");

const core = @import("core.zig");
const resp = @import("resp.zig");
const replication = @import("replication.zig");
const store = @import("store.zig");

const Elem = core.Elem;
const Logger = @import("logger.zig").Logger;
const Server = @import("server.zig").Server;
const Task = @import("task.zig").Task;

const ELEM_MAX_DEPTH: usize = 128;

const SyntaxError = error{
    InvalidArgument,
    InvalidCommand,
    InvalidKey,
    InvalidValue,
    NoCommand,
    TooFewArguments,
    TooManyArguments,
    UnknownArgument,
    UnknownCommand,
    UnknownValue,
};

const RunError = error{
    InternalError,
    NotAMaster,
    NotAReplica,
};

const ParseCtx = struct {
    args_stack: *ElemStack,
    logger: Logger,
    allocator: std.mem.Allocator,
};

pub const Result = struct {
    elem: Elem,
    ok: bool = true,

    const NBS = Result{ .elem = Elem{ .nbs = null } };

    pub fn from_str(str: []const u8) Result {
        return Result{ .elem = Elem{ .str = str } };
    }

    pub fn from_err(err: anyerror) Result {
        return Result{ .elem = Elem{ .err = @errorName(err) }, .ok = false };
    }

    pub fn from_int(int: i64) Result {
        return Result{ .elem = Elem{ .int = int }, .ok = false };
    }

    pub fn from_arr(arr: []const Elem) Result {
        return Result{ .elem = Elem{ .arr = arr } };
    }

    pub fn format(value: Result, comptime fmt: []const u8, options: std.fmt.FormatOptions, writer: anytype) !void {
        _ = fmt;
        _ = options;
        const w: std.io.AnyWriter = writer;
        try w.print("Result{{.elem={any}}}", .{&value.elem});
    }
};

const _TestCommand = struct {
    subcommand: Subcmd,

    const Subcmd = union(enum) {
        queuing,
        slowset: SetCommand,
        slowset_p: SetCommand,
    };

    fn parse(ctx: ParseCtx) !Command {
        const subcmd = b: {
            if (try ctx.args_stack.pop_literal("SLOWSET")) {
                const command = try SetCommand.parse(ctx);
                break :b Subcmd{ .slowset = command.set };
            } else if (try ctx.args_stack.pop_literal("SLOWSET_P")) {
                const command = try SetCommand.parse(ctx);
                break :b Subcmd{ .slowset = command.set };
            } else if (try ctx.args_stack.pop_literal("QUEUING")) {
                break :b Subcmd{ .queuing = {} };
            } else return SyntaxError.UnknownValue;
        };
        return Command{ ._test = _TestCommand{ .subcommand = subcmd } };
    }

    fn _run_queuing(self: @This(), task: *Task, server: *Server) !Result {
        _ = self;
        _ = server;
        const timeout_ns = 5_000_000;
        const ms = try task.require_master_state();
        const data = "*3\r\n+SET\r\n+_test\r\n+foo\r\n";
        var fba = std.io.fixedBufferStream(data);
        const et = try resp.parse_et(fba.reader().any(), task.allocator);
        const batch1 = try ms.replicas.bq.queue_batch(.{ .propagate = et }, timeout_ns);
        ms.replicas.bq._dump("_TEST QUEUING batch1 queued");
        const batch2 = try ms.replicas.bq.queue_batch(.{ .propagate = et }, timeout_ns);
        ms.replicas.bq._dump("_TEST QUEUING batch2 queued");
        const batch3 = try ms.replicas.bq.queue_batch(.{ .propagate = et }, timeout_ns);
        ms.replicas.bq._dump("_TEST QUEUING batch3 queued");
        while (true) {
            const item = batch1.next();
            std.log.debug("batch1.next()={any}", .{item});
            if (item == null) break;
        }
        ms.replicas.bq._dump("_TEST QUEUING batch1 exhausted");
        batch1.deinit();
        ms.replicas.bq._dump("_TEST QUEUING batch1 deinited");
        while (true) {
            const item = batch2.next();
            std.log.debug("batch2.next()={any}", .{item});
            if (item == null) break;
        }
        ms.replicas.bq._dump("_TEST QUEUING batch2 exhausted");
        batch2.deinit();
        ms.replicas.bq._dump("_TEST QUEUING batch2 deinited");
        while (true) {
            const item = batch3.next();
            std.log.debug("batch3.next()={any}", .{item});
            if (item == null) break;
        }
        ms.replicas.bq._dump("_TEST QUEUING batch3 exhausted");
        batch3.deinit();
        ms.replicas.bq._dump("_TEST QUEUING batch3 deinited");
        task.logger.info("_TestCommand().run():et={any}", .{et});
        return Result.from_str("najs");
    }

    fn _run_slowset(self: @This(), task: *Task, server: *Server) !Result {
        task.sleep(5_000);
        return try self.subcommand.slowset.run(task, server);
    }

    fn run(self: @This(), task: *Task, server: *Server) !Result {
        return switch (self.subcommand) {
            .slowset_p => try self._run_slowset(task, server),
            .slowset => try self._run_slowset(task, server),
            .queuing => try self._run_queuing(task, server),
        };
    }
};

const ConfigCommand = struct {
    subcommand: Subcmd,

    const Subcmd = union(enum) {
        get: struct { key: []const u8 },
    };

    fn parse(ctx: ParseCtx) !Command {
        var subcmd: Subcmd = undefined;
        if (try ctx.args_stack.pop_literal("GET")) {
            const key = try ctx.args_stack.pop_str();
            subcmd = Subcmd{ .get = .{ .key = key } };
        } else return SyntaxError.UnknownValue;
        return Command{ .config = .{ .subcommand = subcmd } };
    }

    fn _run_get(self: @This(), server: *Server, allocator: std.mem.Allocator) !Result {
        var elems_al = std.ArrayList(Elem).init(allocator);
        const key = self.subcommand.get.key;
        const key_elem = Elem{ .str = key };
        const value_elem = blk: {
            if (std.mem.eql(u8, key, "dir")) break :blk Elem{
                .str = server.config.dir orelse break :blk core.NBS,
            };
            if (std.mem.eql(u8, key, "dbfilename")) break :blk Elem{
                .str = server.config.dbfilename,
            };
            return Result.NBS;
        };
        try elems_al.append(key_elem);
        try elems_al.append(value_elem);
        return Result.from_arr(elems_al.items);
    }

    fn run(self: @This(), task: *Task, server: *Server) !Result {
        switch (self.subcommand) {
            .get => return try self._run_get(server, task.allocator),
        }
    }
};

const EchoCommand = struct {
    message: Elem,

    fn parse(ctx: ParseCtx) !Command {
        const message = try ctx.args_stack.pop_elem();
        return Command{ .echo = EchoCommand{ .message = message } };
    }

    fn run(self: @This(), task: *Task, server: *Server) !Result {
        _ = task;
        _ = server;
        return Result{ .elem = self.message };
    }
};

const GetCommand = struct {
    key: []const u8,

    fn parse(ctx: ParseCtx) !Command {
        const key = try ctx.args_stack.pop_str();
        return Command{ .get = GetCommand{ .key = key } };
    }

    fn run(self: @This(), task: *Task, server: *Server) !Result {
        _ = task;
        server.store.m.lock();
        defer server.store.m.unlock();
        const maybe_elem = try server.store.get_elem(self.key);
        return Result{ .elem = maybe_elem orelse core.NBS };
    }
};

const InfoCommand = struct {
    sections: std.AutoHashMap(Section, void),

    const Section = enum { replication };
    const KnownSection = struct {
        section: Section,
        label: []const u8,
    };

    const KNOWN_SECTIONS: []const KnownSection = &.{
        KnownSection{ .section = .replication, .label = "Replication" },
    };

    const Buff = struct {
        al: std.ArrayList(u8),
        fn init(allocator: std.mem.Allocator) Buff {
            return Buff{ .al = std.ArrayList(u8).init(allocator) };
        }
        fn append_cmt(self: *Buff, text: []const u8) !void {
            const w = self.al.writer();
            try std.fmt.format(w, "# {s}\n", .{text});
        }
        fn append_replid(self: *Buff, key: []const u8, value: u160) !void {
            const w = self.al.writer();
            try std.fmt.format(w, "{s}:{x:040}\n", .{ key, value });
        }
        fn append_int(self: *Buff, key: []const u8, value: usize) !void {
            const w = self.al.writer();
            try std.fmt.format(w, "{s}:{d}\n", .{ key, value });
        }
        fn append_str(self: *Buff, key: []const u8, value: []const u8) !void {
            const w = self.al.writer();
            try std.fmt.format(w, "{s}:{s}\n", .{ key, value });
        }
        fn append_crep(self: *Buff, r: replication.Replica, s: *Server) !void {
            const connection = try s.connection_pool.require(r.connection_h);
            const w = self.al.writer();
            try std.fmt.format(w, "# connected replica: connection_handle:{any} addr:{any} port:{?d} capabilities:", .{
                r.connection_h,
                connection.address,
                r.port,
            });
            var wrote = false;
            if (r.capabilities.psync2) {
                if (wrote) try w.writeByte(',');
                wrote = true;
                try w.writeAll("psync2");
            }
            try w.writeByte('\n');
        }
    };

    fn parse(ctx: ParseCtx) !Command {
        var sections = std.AutoHashMap(Section, void).init(ctx.allocator);
        while (ctx.args_stack.has_more()) {
            const section_name = try ctx.args_stack.pop_str();
            const section: Section = std.meta.stringToEnum(Section, section_name) orelse {
                ctx.logger.err("unknown section name: '{s}'", .{section_name});
                return SyntaxError.UnknownValue;
            };
            try sections.put(section, {});
        }
        return Command{ .info = InfoCommand{ .sections = sections } };
    }

    fn run(self: @This(), task: *Task, server: *Server) !Result {
        var buff = Buff.init(task.allocator);
        for (KNOWN_SECTIONS) |known_section| {
            if (!self.sections.contains(known_section.section)) continue;
            try buff.append_cmt(known_section.label);
            switch (known_section.section) {
                .replication => {
                    switch (server.replication_state) {
                        .master => |v| {
                            try buff.append_str("role", "master");
                            try buff.append_replid("master_replid", v.replid);
                            try buff.append_int("master_repl_offset", @intCast(v.repl_offset));
                            var itr = v.replicas.ptr_iterator();
                            while (itr.next()) |ptr| try buff.append_crep(ptr.*, server);
                        },
                        .replica => |v| {
                            try buff.append_str("role", "slave");
                            try buff.append_str("master_host", v.master_host);
                            try buff.append_int("master_port", @intCast(v.master_port));
                        },
                    }
                },
            }
        }
        return Result.from_str(buff.al.items);
    }
};

const KeysCommand = struct {
    glob: glob.Glob,
    _pattern: []const u8,

    fn parse(ctx: ParseCtx) !Command {
        const pattern = try ctx.args_stack.pop_str();
        const glob_ = glob.parse_glob(pattern, ctx.allocator) catch |e| {
            ctx.logger.err("invalid pattern: {} when parsing '{s}'", .{ e, pattern });
            return SyntaxError.InvalidArgument;
        };
        return Command{ .keys = KeysCommand{ .glob = glob_, ._pattern = pattern } };
    }

    fn run(self: @This(), task: *Task, server: *Server) !Result {
        var elems_al = std.ArrayList(Elem).init(task.allocator);
        server.store.m.lock();
        defer server.store.m.unlock();
        var key_iter = server.store.elems_db.keyIterator();
        while (key_iter.next()) |key| {
            const is_match = self.glob.match(key.*) catch |e| {
                task.logger.err("internal error: {} when matching key: '{s}' against glob pattern '{s}'", .{
                    e,
                    key.*,
                    self._pattern,
                });
                return e;
            };
            if (!is_match) continue;
            try elems_al.append(Elem{ .str = key.* });
        }
        return Result.from_arr(elems_al.items);
    }
};

const PingCommand = struct {
    fn parse(task: ParseCtx) !Command {
        _ = task;
        return Command{ .ping = PingCommand{} };
    }

    fn run(self: @This(), task: *Task, server: *Server) !Result {
        _ = self;
        _ = task;
        _ = server;
        return Result.from_str("PONG");
    }
};

const PsyncCommand = struct {
    replid: ?u160,
    offset: isize,

    fn parse(ctx: ParseCtx) !Command {
        return Command{ .psync = PsyncCommand{
            .replid = blk: {
                const word = try ctx.args_stack.pop_str();
                if (word.len == 1 and word[0] == '?') break :blk null;
                break :blk std.fmt.parseInt(u160, word, 16) catch return SyntaxError.InvalidValue;
            },
            .offset = blk: {
                const word = try ctx.args_stack.pop_str();
                break :blk std.fmt.parseInt(isize, word, 10) catch return SyntaxError.InvalidValue;
            },
        } };
    }

    fn run(self: @This(), task: *Task, server: *Server) !Result {
        _ = self;
        _ = server;
        const ms = try task.require_master_state();
        var cmd_al = std.ArrayList(u8).init(task.allocator);
        try cmd_al.writer().print(
            "FULLRESYNC {x} {d}",
            .{ ms.replid, 0 },
        );
        return Result{ .elem = Elem{ .str = try cmd_al.toOwnedSlice() } };
    }
};

const ReplconfCommand = struct {
    subcommand: Subcmd,

    const Subcmd = union(enum) {
        listening_port: u16,
        capa: []replication.Capability,
        getack: void,
    };

    fn parse(ctx: ParseCtx) !Command {
        var subcmd: Subcmd = undefined;
        if (try ctx.args_stack.pop_literal("listening-port")) {
            const port_str = try ctx.args_stack.pop_str();
            const port = std.fmt.parseInt(u16, port_str, 10) catch return SyntaxError.InvalidValue;
            subcmd = Subcmd{ .listening_port = port };
        } else if (try ctx.args_stack.pop_literal("capa")) {
            var capas_al = std.ArrayList(replication.Capability).init(ctx.allocator);
            if (try ctx.args_stack.pop_literal("psync2")) {
                try capas_al.append(.psync2);
            } else return SyntaxError.InvalidValue;
            subcmd = Subcmd{ .capa = capas_al.items };
        } else if (try ctx.args_stack.pop_literal("getack")) {
            _ = try ctx.args_stack.pop_literal("*"); // no explanation for this argument
            subcmd = Subcmd{ .getack = {} };
        } else return SyntaxError.UnknownValue;
        return Command{ .replconf = .{ .subcommand = subcmd } };
    }

    fn run(self: @This(), task: *Task, server: *Server) !Result {
        _ = server;
        switch (self.subcommand) {
            .listening_port => |port| {
                const ms = try task.require_master_state();
                const replica = try ms.mkreplica(task.connection_h);
                replica.port = port;
                return Result.from_str("OK");
            },
            .capa => |new_capas| {
                const ms = try task.require_master_state();
                const replica = try ms.mkreplica(task.connection_h);
                replica.set_capas(new_capas);
                return Result.from_str("OK");
            },
            .getack => {
                const rs = try task.require_replica_state();
                var eb = ElemBuilder.init(task.allocator);
                try eb.append_str("REPLCONF");
                try eb.append_str("ACK");
                try eb.append_fmt("{d}", .{rs.master_last_offset orelse 0});
                return Result.from_arr(try eb.al.toOwnedSlice());
            },
        }
    }
};

const SetCommand = struct {
    key: []const u8,
    elem: Elem,
    xc: ExistenceConstraint,
    px_update: store.PxUpdate,
    do_get: bool,

    const ExistenceConstraint = enum { yes, no, any };

    fn parse(ctx: ParseCtx) !Command {
        const key = try ctx.args_stack.pop_str();
        const elem = try ctx.args_stack.pop_elem();
        var do_get: bool = false;
        var xc: ExistenceConstraint = .any;
        var px_update: store.PxUpdate = .unset;
        while (ctx.args_stack.has_more()) {
            if (try ctx.args_stack.pop_literal("NX")) {
                xc = .no;
            } else if (try ctx.args_stack.pop_literal("XX")) {
                xc = .yes;
            } else if (try ctx.args_stack.pop_literal("GET")) {
                do_get = true;
            } else if (try ctx.args_stack.pop_literal("EX")) {
                const param_str = try ctx.args_stack.pop_str();
                const offset_s = std.fmt.parseInt(i64, param_str, 10) catch return SyntaxError.InvalidValue;
                const offset_ms = std.math.mul(i64, offset_s, 1000) catch return SyntaxError.InvalidValue;
                const ts_ms = std.math.add(i64, offset_ms, std.time.milliTimestamp()) catch return SyntaxError.InvalidValue;
                px_update = .{ .new = ts_ms };
            } else if (try ctx.args_stack.pop_literal("PX")) {
                const param_str = try ctx.args_stack.pop_str();
                const offset_ms = std.fmt.parseInt(i64, param_str, 10) catch return SyntaxError.InvalidValue;
                const ts_ms = std.math.add(i64, offset_ms, std.time.milliTimestamp()) catch return SyntaxError.InvalidValue;
                px_update = .{ .new = ts_ms };
            } else if (try ctx.args_stack.pop_literal("EXAT")) {
                const param_str = try ctx.args_stack.pop_str();
                const ts_s = std.fmt.parseInt(i64, param_str, 10) catch return SyntaxError.InvalidValue;
                const ts_ms = std.math.mul(i64, ts_s, 1000) catch return SyntaxError.InvalidValue;
                px_update = .{ .new = ts_ms };
            } else if (try ctx.args_stack.pop_literal("PXAT")) {
                const param_str = try ctx.args_stack.pop_str();
                const ts_ms = std.fmt.parseInt(i64, param_str, 10) catch return SyntaxError.InvalidValue;
                px_update = .{ .new = ts_ms };
            } else if (try ctx.args_stack.pop_literal("KEEPTTL")) {
                px_update = .keep;
            } else {
                return SyntaxError.UnknownArgument;
            }
        }
        return Command{ .set = SetCommand{
            .key = key,
            .elem = elem,
            .do_get = do_get,
            .xc = xc,
            .px_update = px_update,
        } };
    }

    fn _put_get_elem(self: @This(), server: *Server) !Elem {
        var ok_elem = Elem{ .str = "OK" };
        if (self.do_get) {
            ok_elem = if (try server.store.get_elem(self.key)) |prev| prev else core.NBS;
        }
        try server.store.put_elem(self.key, self.elem, self.px_update);
        return ok_elem;
    }

    fn run(self: @This(), task: *Task, server: *Server) !Result {
        server.store.m.lock();
        defer server.store.m.unlock();
        switch (self.xc) {
            .yes => if (!server.store.elems_db.contains(self.key)) return Result.NBS,
            .no => if (server.store.elems_db.contains(self.key)) return Result.NBS,
            .any => {},
        }
        return Result{ .elem = self._put_get_elem(server) catch |e| {
            task.logger.warn("error when storing element: {}", .{e});
            return e;
        } };
    }
};

const WaitCommand = struct {
    numreplicas: usize,
    timeout_ms: u32,

    fn parse(ctx: ParseCtx) !Command {
        const numreplicas = try ctx.args_stack.pop_int(usize);
        const timeout_ms = try ctx.args_stack.pop_int(u32);
        return Command{ .wait = WaitCommand{
            .numreplicas = numreplicas,
            .timeout_ms = timeout_ms,
        } };
    }

    fn run(self: @This(), task: *Task, server: *Server) !Result {
        _ = server;
        const ok_count = try task.check_replicas(self.numreplicas, self.timeout_ms);
        return Result.from_int(@intCast(ok_count));
    }
};

pub const Command = union(enum) {
    _test: _TestCommand,
    config: ConfigCommand,
    echo: EchoCommand,
    get: GetCommand,
    info: InfoCommand,
    keys: KeysCommand,
    ping: PingCommand,
    psync: PsyncCommand,
    replconf: ReplconfCommand,
    set: SetCommand,
    wait: WaitCommand,

    pub const Flag = enum { propagate, start_replica, ack_master };

    pub fn parse(elem: ?Elem, logger: Logger, allocator: std.mem.Allocator) !Command {
        const tokens = elem orelse return SyntaxError.NoCommand;
        if (tokens != .arr) return SyntaxError.InvalidCommand;
        if (tokens.arr.len == 0) return SyntaxError.InvalidCommand;
        var args_stack = ElemStack{ .elems = tokens.arr };
        const ctx = ParseCtx{ .args_stack = &args_stack, .logger = logger, .allocator = allocator };
        var command: Command = undefined;
        if (try args_stack.pop_literal("_TEST")) {
            command = try _TestCommand.parse(ctx);
        } else if (try args_stack.pop_literal("ECHO")) {
            command = try EchoCommand.parse(ctx);
        } else if (try args_stack.pop_literal("GET")) {
            command = try GetCommand.parse(ctx);
        } else if (try args_stack.pop_literal("INFO")) {
            command = try InfoCommand.parse(ctx);
        } else if (try args_stack.pop_literal("KEYS")) {
            command = try KeysCommand.parse(ctx);
        } else if (try args_stack.pop_literal("PING")) {
            command = try PingCommand.parse(ctx);
        } else if (try args_stack.pop_literal("PSYNC")) {
            command = try PsyncCommand.parse(ctx);
        } else if (try args_stack.pop_literal("REPLCONF")) {
            command = try ReplconfCommand.parse(ctx);
        } else if (try args_stack.pop_literal("SET")) {
            command = try SetCommand.parse(ctx);
        } else if (try args_stack.pop_literal("WAIT")) {
            command = try WaitCommand.parse(ctx);
        } else if (try args_stack.pop_literal("CONFIG")) {
            command = try ConfigCommand.parse(ctx);
        } else {
            return SyntaxError.UnknownCommand;
        }
        try args_stack.assert_done();
        return command;
    }

    pub fn run(
        self: @This(),
        task: *Task,
        server: *Server,
    ) Result {
        return switch (self) {
            inline else => |c| c.run(task, server) catch |e| Result.from_err(e),
        };
    }

    pub fn should(self: Command, flag: Flag) bool {
        return switch (flag) {
            .propagate => switch (self) {
                .set => true,
                ._test => |impl| switch (impl.subcommand) {
                    .slowset => true,
                    else => false,
                },
                else => false,
            },
            .start_replica => switch (self) {
                .psync => true,
                else => false,
            },
            .ack_master => switch (self) {
                .replconf => |c| c.subcommand == .getack,
                else => false,
            },
        };
    }
};

const ElemStack = struct {
    elems: []const Elem,
    cur: usize = 0,

    fn _step(self: *ElemStack, n: usize) void {
        //
        // [|]      (0) +0 => 0 < 0     // ok
        // [|]      (0) +1 => 0 < 1     // unreachable
        // [|]      (0) +2 => 0 < 2     // unreachable
        // [ ]!     (1) +0 => 0 < 1     // unreachable
        // [ ]!     (1) +1 => 0 < 2     // unreachable
        // [ ]!     (1) +2 => 0 < 3     // unreachable
        // [ ]!     (2) +0 => 0 < 2     // unreachable
        // [ ]!     (2) +1 => 0 < 3     // unreachable
        // [ ]!     (2) +2 => 0 < 4     // unreachable
        //
        // [|_ ]    (0) +0 => 1 < 0     // ok
        // [|_ ]    (0) +1 => 1 < 1     // ok
        // [|_ ]    (0) +2 => 1 < 2     // unreachable
        // [ _|]    (1) +0 => 1 < 1     // ok
        // [ _|]    (1) +1 => 1 < 2     // unreachable
        // [ _|]    (1) +2 => 1 < 3     // unreachable
        // [ _ ]!   (2) +0 => 1 < 2     // unreachable
        // [ _ ]!   (2) +1 => 1 < 3     // unreachable
        // [ _ ]!   (2) +2 => 1 < 4     // unreachable
        //
        // [|_ _ ]  (0) +0 => 2 < 0     // ok
        // [|_ _ ]  (0) +1 => 2 < 1     // ok
        // [|_ _ ]  (0) +2 => 2 < 2     // ok
        // [ _|_ ]  (1) +0 => 2 < 1     // ok
        // [ _|_ ]  (1) +1 => 2 < 2     // ok
        // [ _|_ ]  (1) +2 => 2 < 3     // unreachable
        // [ _ _|]  (2) +0 => 2 < 2     // ok
        // [ _ _|]  (2) +1 => 2 < 3     // unreachable
        // [ _ _|]  (2) +2 => 2 < 4     // unreachable
        //
        if (self.elems.len < (self.cur + n)) unreachable;
        self.cur = self.cur + n;
    }
    fn assert_done(self: ElemStack) !void {
        if (self.has_more()) return SyntaxError.TooManyArguments;
        return;
    }
    fn has_more(self: ElemStack) bool {
        const rest = self.elems[self.cur..];
        return rest.len > 0;
    }
    fn _maybe_next(self: ElemStack) ?Elem {
        if (self.cur >= self.elems.len) return null;
        return self.elems[self.cur];
    }
    fn pop_literal(self: *ElemStack, str: []const u8) !bool {
        const next = self._maybe_next() orelse return SyntaxError.TooFewArguments;
        if (next != .str) return SyntaxError.InvalidValue;
        if (!util.eql_i(next.str, str)) return false;
        self._step(1);
        return true;
    }
    fn pop_elem(self: *ElemStack) !Elem {
        const next = self._maybe_next() orelse return SyntaxError.TooFewArguments;
        self._step(1);
        return next;
    }
    fn pop_int(self: *ElemStack, comptime T: type) !T {
        const elem = try self.pop_elem();
        const ti: std.builtin.Type = @typeInfo(T);
        if (ti != .Int) @compileError("not an integer type: " ++ @typeName(T) ++ " is " ++ @tagName(ti));
        switch (elem) {
            .int => |n| return @intCast(n),
            .str => |s| return std.fmt.parseInt(T, s, 10) catch return SyntaxError.InvalidValue,
            else => return SyntaxError.InvalidValue,
        }
    }
    fn pop_str(self: *ElemStack) ![]const u8 {
        const next = self._maybe_next() orelse return SyntaxError.TooFewArguments;
        if (next != .str) return SyntaxError.InvalidValue;
        self._step(1);
        return next.str;
    }
};

test "ElemStack" {
    const Case = struct {
        stack: *ElemStack,
        const Case = @This();
        fn init(elems: []const Elem) !Case {
            var elems_al = std.ArrayList(Elem).init(std.heap.page_allocator);
            try elems_al.appendSlice(elems);
            const stack_ptr = try std.heap.page_allocator.create(ElemStack);
            stack_ptr.elems = elems_al.items;
            stack_ptr.cur = 0;
            return .{ .stack = stack_ptr };
        }
        fn expect_fullness(self: *Case) !void {
            try std.testing.expectEqual(true, self.stack.has_more());
            try std.testing.expectEqual(SyntaxError.TooManyArguments, self.stack.assert_done());
        }
        fn expect_emptiness(self: *Case) !void {
            try std.testing.expectEqual(false, self.stack.has_more());
            try self.stack.assert_done();
            try std.testing.expectEqual(SyntaxError.TooFewArguments, self.stack.pop_literal("FOO"));
            try std.testing.expectEqual(SyntaxError.TooFewArguments, self.stack.pop_str());
            try std.testing.expectEqual(SyntaxError.TooFewArguments, self.stack.pop_elem());
        }
    };

    var c: Case = undefined;
    const mkcase = Case.init;

    c = try mkcase(&.{});
    try c.expect_emptiness();

    c = try mkcase(&.{Elem{ .null = null }});
    try c.expect_fullness();
    try std.testing.expectEqual(SyntaxError.InvalidValue, c.stack.pop_literal("FOO"));
    try std.testing.expectEqual(SyntaxError.InvalidValue, c.stack.pop_str());
    try std.testing.expectEqualDeep(Elem{ .null = null }, try c.stack.pop_elem());
    try c.expect_emptiness();

    c = try mkcase(&.{Elem{ .int = 3 }});
    try c.expect_fullness();
    try std.testing.expectEqual(SyntaxError.InvalidValue, c.stack.pop_literal("FOO"));
    try std.testing.expectEqual(SyntaxError.InvalidValue, c.stack.pop_str());
    try std.testing.expectEqualDeep(Elem{ .int = 3 }, try c.stack.pop_elem());
    try c.expect_emptiness();

    c = try mkcase(&.{Elem{ .str = "" }});
    try c.expect_fullness();
    try std.testing.expectEqual(false, c.stack.pop_literal("FOO"));
    try std.testing.expectEqual("", c.stack.pop_str());
    try c.expect_emptiness();

    c = try mkcase(&.{Elem{ .str = "" }});
    try c.expect_fullness();
    try std.testing.expectEqualDeep(Elem{ .str = "" }, try c.stack.pop_elem());
    try c.expect_emptiness();

    c = try mkcase(&.{Elem{ .str = "FOO" }});
    try c.expect_fullness();
    try std.testing.expectEqual(true, c.stack.pop_literal("FOO"));
    try c.expect_emptiness();

    c = try mkcase(&.{Elem{ .str = "FOO" }});
    try c.expect_fullness();
    try std.testing.expectEqualDeep(Elem{ .str = "FOO" }, try c.stack.pop_elem());
    try c.expect_emptiness();

    c = try mkcase(&.{ Elem{ .str = "FOO" }, Elem{ .null = null } });
    try c.expect_fullness();
    try std.testing.expectEqual(true, c.stack.pop_literal("FOO"));
    try std.testing.expectEqual(SyntaxError.InvalidValue, c.stack.pop_str());
    try std.testing.expectEqualDeep(Elem{ .null = null }, try c.stack.pop_elem());
    try c.expect_emptiness();

    c = try mkcase(&.{ Elem{ .str = "FOO" }, Elem{ .null = null } });
    try c.expect_fullness();
    try std.testing.expectEqual(true, c.stack.pop_literal("FOO"));
    try std.testing.expectEqualDeep(Elem{ .null = null }, try c.stack.pop_elem());
    try c.expect_emptiness();

    c = try mkcase(&.{ Elem{ .str = "FOO" }, Elem{ .int = 42 } });
    try c.expect_fullness();
    try std.testing.expectEqual(true, c.stack.pop_literal("FOO"));
    try std.testing.expectEqual(SyntaxError.InvalidValue, c.stack.pop_str());
    try std.testing.expectEqualDeep(Elem{ .int = 42 }, try c.stack.pop_elem());
    try c.expect_emptiness();

    c = try mkcase(&.{ Elem{ .str = "FOO" }, Elem{ .int = 42 } });
    try c.expect_fullness();
    try std.testing.expectEqual(true, c.stack.pop_literal("FOO"));
    try std.testing.expectEqualDeep(Elem{ .int = 42 }, try c.stack.pop_elem());
    try c.expect_emptiness();

    c = try mkcase(&.{ Elem{ .str = "FOO" }, Elem{ .str = "bar" } });
    try c.expect_fullness();
    try std.testing.expectEqual(true, c.stack.pop_literal("FOO"));
    try std.testing.expectEqual("bar", c.stack.pop_str());
    try c.expect_emptiness();

    c = try mkcase(&.{ Elem{ .str = "FOO" }, Elem{ .str = "bar" } });
    try c.expect_fullness();
    try std.testing.expectEqual(true, c.stack.pop_literal("FOO"));
    try std.testing.expectEqualDeep(Elem{ .str = "bar" }, try c.stack.pop_elem());
    try c.expect_emptiness();
}

const ElemBuilder = struct {
    al: std.ArrayList(Elem),
    allocator: std.mem.Allocator,

    fn init(allocator: std.mem.Allocator) ElemBuilder {
        return ElemBuilder{ .al = std.ArrayList(Elem).init(allocator), .allocator = allocator };
    }

    fn append_str(self: *ElemBuilder, str: []const u8) !void {
        try self.al.append(Elem{ .str = try self.allocator.dupe(u8, str) });
    }

    fn append_fmt(self: *ElemBuilder, comptime fmt: []const u8, args: anytype) !void {
        var str = std.ArrayList(u8).init(self.allocator);
        try str.writer().print(fmt, args);
        try self.al.append(Elem{ .str = try str.toOwnedSlice() });
    }

    fn to_arr(self: *ElemBuilder) !Elem {
        return Elem{ .arr = try self.al.toOwnedSlice() };
    }
};
