const std = @import("std");

const util = @import("../util.zig");

pub const ElemType = enum { null, nbs, bool, int, double, err, str, bin, arr };

pub const Elem = union(ElemType) {
    null: ?void,
    nbs: ?void,
    bool: bool,
    int: i64,
    double: f64,
    err: []const u8,
    str: []const u8,
    bin: []const u8,
    arr: []const Elem,
    pub fn format(self: Elem, comptime fmt: []const u8, options: std.fmt.FormatOptions, writer: anytype) !void {
        _ = fmt;
        _ = options;
        try deep_fmt_elem(&self, writer, ELEM_MAX_DEPTH);
    }
};

pub const ElemTree = struct {
    root: ?*Elem,
    raw_bytes: []const u8,
    arena: std.heap.ArenaAllocator,

    pub const Error = error{
        ElemIsABinaryString,
        ElemIsAScalar,
        ElemIsAString,
        ElemIsAnArray,
        ElemIsAnError,
        ElemIsAnInteger,
        ElemNotABinaryString,
        ElemNotAString,
        ElemNotAnArray,
        ElemNotAnError,
        ElemNotAnInteger,
        ElemNotExpected,
        ElemStringNotAnInt,
        ElemTooLong,
        ElemTooShort,
        ElemValueTooHigh,
        ElemValueTooLow,
        NoElem,
    };

    pub fn deinit(self: ElemTree) void {
        self.arena.deinit();
    }

    pub fn format(self: ElemTree, comptime fmt: []const u8, options: std.fmt.FormatOptions, writer: anytype) !void {
        _ = fmt;
        _ = options;
        if (self.root) |_| {
            try writer.print("ElemTree{{.raw_bytes=({d}B),.root={?any}}}", .{ self.raw_bytes.len, self.root });
        } else {
            try writer.writeAll("ElemTree{}(STREAM_END)");
        }
    }

    const Condition = struct {
        idx: ?usize = null,

        isa_arr: ?bool = null,
        isa_bin: ?bool = null,
        isa_str: ?bool = null,
        isa_err: ?bool = null,
        isa_int: ?bool = null,

        len_gt: ?usize = null,
        len_ge: ?usize = null,
        len_eq: ?usize = null,
        len_le: ?usize = null,
        len_lt: ?usize = null,

        int_gt: ?i64 = null,
        int_ge: ?i64 = null,
        int_eq: ?i64 = null,
        int_le: ?i64 = null,
        int_lt: ?i64 = null,

        sint_gt: ?i64 = null,
        sint_ge: ?i64 = null,
        sint_eq: ?i64 = null,
        sint_le: ?i64 = null,
        sint_lt: ?i64 = null,

        str_neq: ?[]const u8 = null,
        str_eq: ?[]const u8 = null,
        str_startswith: ?[]const u8 = null,
        str_neq_i: ?[]const u8 = null,
        str_eq_i: ?[]const u8 = null,
        str_startswith_i: ?[]const u8 = null,

        fn _needs_len(self: Condition) bool {
            if (self.len_gt) |_| return true;
            if (self.len_ge) |_| return true;
            if (self.len_eq) |_| return true;
            if (self.len_le) |_| return true;
            if (self.len_lt) |_| return true;
            return false;
        }

        fn _needs_int(self: Condition) bool {
            if (self.int_gt) |_| return true;
            if (self.int_ge) |_| return true;
            if (self.int_eq) |_| return true;
            if (self.int_le) |_| return true;
            if (self.int_lt) |_| return true;
            return false;
        }

        fn _needs_sint(self: Condition) bool {
            if (self.sint_gt) |_| return true;
            if (self.sint_ge) |_| return true;
            if (self.sint_eq) |_| return true;
            if (self.sint_le) |_| return true;
            if (self.sint_lt) |_| return true;
            return false;
        }

        fn _needs_str(self: Condition) bool {
            if (self.str_neq) |_| return true;
            if (self.str_eq) |_| return true;
            if (self.str_startswith) |_| return true;
            if (self.str_neq_i) |_| return true;
            if (self.str_eq_i) |_| return true;
            if (self.str_startswith_i) |_| return true;
            return false;
        }
    };

    pub fn _assert_cond(condition: Condition, elem: Elem, idx: ?usize) !void {
        if (idx) |i| {
            if (elem != .arr) return Error.ElemNotAnArray;
            if (i >= elem.arr.len) return Error.ElemTooShort;
            return _assert_cond(condition, elem.arr[i], null);
        }

        if (condition.isa_err) |goal| {
            if (goal and elem != .err) return Error.ElemNotAnError;
            if (!goal and elem == .err) return Error.ElemIsAnError;
        }
        if (condition.isa_str) |goal| {
            if (goal and elem != .str) return Error.ElemNotAString;
            if (!goal and elem == .str) return Error.ElemIsAString;
        }
        if (condition.isa_arr) |goal| {
            if (goal and elem != .arr) return Error.ElemNotAnArray;
            if (!goal and elem == .arr) return Error.ElemIsAnArray;
        }
        if (condition.isa_bin) |goal| {
            if (goal and elem != .bin) return Error.ElemNotABinaryString;
            if (!goal and elem == .bin) return Error.ElemIsABinaryString;
        }
        if (condition.isa_int) |goal| {
            if (goal and elem != .arr) return Error.ElemNotAnInteger;
            if (!goal and elem == .arr) return Error.ElemIsAnInteger;
        }

        if (condition._needs_len()) {
            const val = switch (elem) {
                .str => |e| e.len,
                .bin => |e| e.len,
                .err => |e| e.len,
                .arr => |e| e.len,
                else => return Error.ElemIsAScalar,
            };
            if (condition.len_gt) |goal| {
                if (val <= goal) return Error.ElemTooShort;
            }
            if (condition.len_ge) |goal| {
                if (val < goal) return Error.ElemTooShort;
            }
            if (condition.len_eq) |goal| {
                if (val > goal) return Error.ElemTooLong;
                if (val < goal) return Error.ElemTooShort;
            }
            if (condition.len_le) |goal| {
                if (val > goal) return Error.ElemTooLong;
            }
            if (condition.len_lt) |goal| {
                if (val >= goal) return Error.ElemTooLong;
            }
        }

        if (condition._needs_int()) {
            if (elem != .int) return Error.ElemNotAnInteger;
            const val = elem.int;
            if (condition.int_gt) |goal| {
                if (val <= goal) return Error.ElemValueTooLow;
            }
            if (condition.int_ge) |goal| {
                if (val < goal) return Error.ElemValueTooLow;
            }
            if (condition.int_eq) |goal| {
                if (val > goal) return Error.ElemValueTooHigh;
                if (val < goal) return Error.ElemValueTooLow;
            }
            if (condition.int_le) |goal| {
                if (val > goal) return Error.ElemValueTooHigh;
            }
            if (condition.int_lt) |goal| {
                if (val >= goal) return Error.ElemValueTooHigh;
            }
        }

        if (condition._needs_sint()) {
            if (elem != .str) return Error.ElemNotAString;
            const val = std.fmt.parseInt(i64, elem.str, 10) catch return Error.ElemStringNotAnInt;
            if (condition.sint_gt) |goal| {
                if (val <= goal) return Error.ElemValueTooLow;
            }
            if (condition.sint_ge) |goal| {
                if (val < goal) return Error.ElemValueTooLow;
            }
            if (condition.sint_eq) |goal| {
                if (val > goal) return Error.ElemValueTooHigh;
                if (val < goal) return Error.ElemValueTooLow;
            }
            if (condition.sint_le) |goal| {
                if (val > goal) return Error.ElemValueTooHigh;
            }
            if (condition.sint_lt) |goal| {
                if (val >= goal) return Error.ElemValueTooHigh;
            }
        }

        if (condition._needs_str()) {
            if (elem != .str) return Error.ElemNotAString;
            const val = elem.str;
            if (condition.str_eq) |goal| {
                if (!std.mem.eql(u8, val, goal)) return Error.ElemNotExpected;
            }
            if (condition.str_neq) |goal| {
                if (std.mem.eql(u8, val, goal)) return Error.ElemNotExpected;
            }
            if (condition.str_startswith) |goal| {
                if (!std.mem.startsWith(u8, val, goal)) return Error.ElemNotExpected;
            }
            if (condition.str_eq_i) |goal| {
                if (!util.eql_i(goal, val)) return Error.ElemNotExpected;
            }
            if (condition.str_neq_i) |goal| {
                if (util.eql_i(goal, val)) return Error.ElemNotExpected;
            }
            if (condition.str_startswith_i) |goal| {
                if (!util.startswith_i(goal, val)) return Error.ElemNotExpected;
            }
        }
    }

    pub fn assert_cond(self: ElemTree, condition: Condition) !void {
        const elem = self.root orelse return Error.NoElem;
        try _assert_cond(condition, elem.*, condition.idx);
    }

    pub fn assert_arr_len(self: ElemTree, len: usize) !void {
        const elem = self.root orelse return Error.NoElem;
        if (elem.* != .arr) return Error.ElemNotAnArray;
        if (elem.arr.len < len) return Error.ElemTooShort;
        if (elem.arr.len > len) return Error.ElemTooLong;
    }

    pub fn assert_bin_len_ge(self: ElemTree, len: usize) !void {
        const elem = self.root orelse return Error.NoElem;
        if (elem.* != .bin) return Error.ElemNotAString;
        if (elem.arr.len < len) return Error.ElemTooShort;
    }

    pub fn assert_eql_at(self: ElemTree, idx: usize, text: []const u8) !void {
        const elem = self.root orelse return Error.NoElem;
        if (elem.* != .arr) return Error.ElemNotAnArray;
        if (idx >= elem.arr.len) return Error.ElemTooShort;
        const e = elem.arr[idx];
        if (e != .str) return Error.ElemNotAString;
        if (!util.eql_i(text, e.str)) return Error.ElemNotExpected;
    }

    pub fn assert_ok(self: ElemTree) !void {
        try self.assert_cond(.{ .isa_err = false });
    }

    pub fn assert_parsed_int_eql_at(self: ElemTree, idx: usize, number: i64) !void {
        const elem = self.root orelse return Error.NoElem;
        if (elem.* != .arr) return Error.ElemNotAnArray;
        if (idx >= elem.arr.len) return Error.ElemTooShort;
        const e = elem.arr[idx];
        if (e != .str) return Error.ElemNotAnArray;
        const value = try std.fmt.parseInt(i64, e.str, 10);
        if (value < number) return Error.ElemValueTooLow;
        if (value > number) return Error.ElemValueTooHigh;
    }

    pub fn assert_startswith(self: ElemTree, text: []const u8) !void {
        const elem = self.root orelse return Error.NoElem;
        if (elem.* != .str) return Error.ElemNotAString;
        if (!util.startswith_i(text, elem.str)) return Error.ElemNotExpected;
    }

    pub fn parse_int_at(self: ElemTree, idx: usize) !i64 {
        try self.assert_cond(.{ .isa_str = true, .idx = idx });
        return try std.fmt.parseInt(i64, self.root.?.*.arr[idx].str, 10);
    }
};

pub const NBS = Elem{ .nbs = null };

pub const ELEM_MAX_DEPTH: usize = 128;

pub fn deep_copy_elem(elem: *const Elem, allocator: std.mem.Allocator, frames: usize) !*Elem {
    if (frames < 1) return error.ValueTooComplex;
    const new_ptr = try allocator.create(Elem);
    switch (elem.*) {
        .null => |value| new_ptr.* = Elem{ .null = value },
        .nbs => |value| new_ptr.* = Elem{ .nbs = value },
        .bool => |value| new_ptr.* = Elem{ .bool = value },
        .int => |value| new_ptr.* = Elem{ .int = value },
        .double => |value| new_ptr.* = Elem{ .double = value },
        .str => |values| new_ptr.* = Elem{ .str = try allocator.dupe(u8, values) },
        .bin => |values| new_ptr.* = Elem{ .bin = try allocator.dupe(u8, values) },
        .err => |values| new_ptr.* = Elem{ .err = try allocator.dupe(u8, values) },
        .arr => |values| {
            const items = try allocator.alloc(Elem, values.len);
            for (values, 0..) |*value, idx| {
                const new_item_ptr = try deep_copy_elem(value, allocator, frames - 1);
                items[idx] = new_item_ptr.*;
            }
            new_ptr.* = Elem{ .arr = items };
        },
    }
    return new_ptr;
}

pub fn deep_fmt_elem(elem: *const Elem, writer: std.io.AnyWriter, frames: usize) !void {
    if (frames < 1) return error.ValueTooComplex;
    if (frames == ELEM_MAX_DEPTH) {
        try writer.writeAll("Elem{");
    } else {
        try writer.writeAll(".{");
    }
    switch (elem.*) {
        .null => try writer.print(".null", .{}),
        .nbs => try writer.print(".nbs", .{}),
        .bool => |value| try writer.print(".bool={}", .{value}),
        .int => |value| try writer.print(".int={d}", .{value}),
        .double => |value| try writer.print(".double={d}", .{value}),
        .str => |value| try util.qstr(value, writer),
        .bin => |value| try util.qblob(value, writer),
        .err => |value| try writer.print(".err=\"{s}\"", .{value}),
        .arr => |values| {
            try writer.writeByte('[');
            for (values, 0..) |*value, i| {
                if (i > 0) try writer.writeByte(',');
                try deep_fmt_elem(value, writer, frames - 1);
            }
            try writer.print("]", .{});
        },
    }
    try writer.writeByte('}');
}

pub fn deep_free_elem(elem: *const Elem, allocator: std.mem.Allocator, frames: usize) !void {
    if (frames < 1) return error.ValueTooComplex;
    switch (elem.*) {
        .str => |values| allocator.free(values),
        .bin => |values| allocator.free(values),
        .err => |values| allocator.free(values),
        .arr => |values| for (values) |*value_ptr| try deep_free_elem(value_ptr, allocator, frames - 1),
        else => {},
    }
    allocator.destroy(elem);
}

const Overstep = enum {
    data_inside_prohibited_buff,
    data_outside_required_buff,
    data_mixed_layout,
    items_inside_prohibited_buff,
    items_outside_required_buff,
    items_mixed_layout,
    none,
};

fn find_overstep_sli(
    comptime T: type,
    buff_rng: util.Range,
    comptime should_own: bool,
    sli: []const T,
) Overstep {
    const items_rng = util.range_from_sli(T, sli) orelse return .none;
    const items_rel = util.range_rel(items_rng, buff_rng) catch unreachable;
    switch (items_rel) {
        .inside, .same => if (!should_own) return .items_inside_prohibited_buff,
        .left, .right => if (should_own) return .items_outside_required_buff,
        else => return .items_outside_required_buff,
    }
    if (T == Elem) {
        for (sli) |*item| {
            const overstep = find_overstep_elem(buff_rng, should_own, item);
            if (overstep != .none) return overstep;
        }
    }
    return .none;
}

fn find_overstep_elem(
    buff_rng: util.Range,
    comptime should_own: bool,
    elem: *const Elem,
) Overstep {
    const data_rng = util.range_from_ptr(Elem, elem);
    // first, check elem's own memory
    const own_rel = util.range_rel(data_rng, buff_rng) catch unreachable;
    switch (own_rel) {
        .inside, .same => if (!should_own) return .data_inside_prohibited_buff,
        .left, .right => if (should_own) return .data_outside_required_buff,
        else => return .data_mixed_layout,
    }
    // then, check any memory of elems
    switch (elem.*) {
        .nbs, .null, .bool, .int, .double => return .none,
        .err, .str, .bin => |values| return find_overstep_sli(u8, buff_rng, should_own, values),
        .arr => |values| return find_overstep_sli(Elem, buff_rng, should_own, values),
    }
    return .none;
}

fn assert_elem_ownership(
    buff: []u8,
    comptime should_own: bool,
    elem: *const Elem,
) !void {
    const buff_rng = util.range_from_sli(u8, buff) orelse return error.InvalidBuff;
    const overstep = find_overstep_elem(buff_rng, should_own, elem);
    return switch (overstep) {
        .data_inside_prohibited_buff => error.DataInsideProhibitedBuff,
        .data_outside_required_buff => error.DataOutsideRequiredBuff,
        .data_mixed_layout => error.DataMixedLayout,
        .items_inside_prohibited_buff => error.ItemInsideProhibitedBuff,
        .items_outside_required_buff => error.ItemOutsideRequiredBuff,
        .items_mixed_layout => error.ItemMixedLayout,
        .none => {},
    };
}

fn assert_clones_ok(elem: *const Elem) !void {
    const buff_a = try std.heap.page_allocator.alloc(u8, 1024);
    const buff_b = try std.heap.page_allocator.alloc(u8, 1024);
    defer std.heap.page_allocator.free(buff_a);
    defer std.heap.page_allocator.free(buff_b);
    var fba_a = std.heap.FixedBufferAllocator.init(buff_a);
    var fba_b = std.heap.FixedBufferAllocator.init(buff_b);
    const inter = try deep_copy_elem(elem, fba_a.allocator(), 10);
    const clone = try deep_copy_elem(inter, fba_b.allocator(), 10);
    try assert_elem_ownership(buff_b, true, clone);
    try assert_elem_ownership(buff_a, false, clone);
    try std.testing.expectEqualDeep(elem, clone);
}

test "deep copy" {
    try assert_clones_ok(&Elem{ .null = {} });
    try assert_clones_ok(&Elem{ .nbs = {} });
    try assert_clones_ok(&Elem{ .bool = false });
    try assert_clones_ok(&Elem{ .int = 123 });
    try assert_clones_ok(&Elem{ .double = 123.45 });
    try assert_clones_ok(&Elem{ .err = "" });
    try assert_clones_ok(&Elem{ .err = "F" });
    try assert_clones_ok(&Elem{ .err = "FOO" });
    try assert_clones_ok(&Elem{ .str = "" });
    try assert_clones_ok(&Elem{ .str = "F" });
    try assert_clones_ok(&Elem{ .str = "FOO" });
    try assert_clones_ok(&Elem{ .arr = &.{} });
    try assert_clones_ok(&Elem{ .arr = &.{Elem{ .null = {} }} });
    try assert_clones_ok(&Elem{ .arr = &.{Elem{ .arr = &.{Elem{ .null = {} }} }} });
    try assert_clones_ok(&Elem{
        .arr = &.{
            Elem{ .int = 123 },
            Elem{ .arr = &.{} },
            Elem{ .arr = &.{
                Elem{ .str = "" },
                Elem{ .str = "def" },
            } },
            Elem{ .str = "abc" },
        },
    });
}
