const std = @import("std");

const Server = @import("server.zig").Server;
const Role = @import("task.zig").Role;

const Logfile = struct {
    name: []const u8,
    fh: std.fs.File,
    allocator: ?std.mem.Allocator,
    ctx: Logger.Ctx,

    fn deinit(self: Logfile) void {
        if (self.allocator) |a| a.free(self.name);
        self.fh.close();
    }

    fn _open_fh(dir: []const u8, name: []const u8) !std.fs.File {
        const dh = try std.fs.cwd().makeOpenPath(dir, .{});
        return dh.createFile(name, .{}) catch |e| switch (e) {
            error.PathAlreadyExists => return try dh.openFile(name, .{}),
            else => return e,
        };
    }

    fn try_init(dir: []const u8, ctx: Logger.Ctx, allocator: ?std.mem.Allocator) ?Logfile {
        const task_fn_fmt = "task-{s}{d:0>2}.log";
        const name = switch (ctx) {
            .root => "root.log",
            .task => |c| std.fmt.allocPrint(allocator.?, task_fn_fmt, .{ c.task_id_pfx, c.task_id }) catch |e| {
                std.log.warn("[LOGGER] could not format log filename: fmt={s}, task_id={d} e={!}", .{ task_fn_fmt, c.task_id, e });
                return null;
            },
        };
        const fh = _open_fh(dir, name) catch |e| {
            std.log.warn("[LOGGER] could not open file: {s}/{s}, e={!}", .{ dir, name, e });
            return null;
        };
        return .{
            .name = name,
            .fh = fh,
            .allocator = allocator,
            .ctx = ctx,
        };
    }

    fn try_write(self: Logfile, comptime sigil: u8, msg: []const u8) void {
        const w = self.fh.writer();
        w.print("{c}: [{any}] {s}\n", .{ sigil, self.ctx, msg }) catch {
            std.log.warn("[LOGGER] failed to write to log file: name='{s}'", .{self.name});
        };
    }
};

pub const Logger = struct {
    options: Options,
    ctx: Ctx,
    logfile: ?Logfile,

    const Options = struct {
        debug: bool = false,
        verbose: bool = false,
        logdir: ?[]const u8,
    };

    pub const TaskCtx = struct {
        task_id: usize,
        task_id_pfx: []const u8,
        thread_id: ?std.Thread.Id,
        handle: Server.ConnectionPool.Handle,
        address: std.net.Address,
        stream: std.net.Stream,
        role: Role,
    };

    pub const Ctx = union(enum) {
        task: TaskCtx,
        root: void,
        pub fn format(self: Ctx, comptime fmt: []const u8, options: std.fmt.FormatOptions, writer: anytype) !void {
            _ = fmt;
            _ = options;
            switch (self) {
                .root => try writer.writeAll("t=ROOT"),
                .task => |ctx| {
                    try writer.print("t={s}{d}", .{ ctx.task_id_pfx, ctx.task_id });
                    if (ctx.thread_id) |thread_id| {
                        try writer.print(" T={any}", .{thread_id});
                    } else {
                        try writer.writeAll(" T=MAIN");
                    }
                    try writer.print(
                        " c={any} a={any} h={any} r={s}",
                        .{
                            ctx.handle,
                            ctx.address,
                            ctx.stream.handle,
                            @tagName(ctx.role),
                        },
                    );
                },
            }
        }
    };

    pub fn init_for_task(options: Options, ctx: TaskCtx, allocator: std.mem.Allocator) Logger {
        const ctx_ = .{ .task = ctx };
        return .{
            .options = options,
            .ctx = ctx_,
            .logfile = if (options.logdir) |logdir| Logfile.try_init(logdir, ctx_, allocator) else null,
        };
    }

    pub fn init_root(options: Options) Logger {
        const ctx = .{ .root = {} };
        return .{
            .options = options,
            .ctx = ctx,
            .logfile = if (options.logdir) |logdir| Logfile.try_init(logdir, ctx, null) else null,
        };
    }

    pub fn deinit(self: Logger) void {
        if (self.logfile) |logfile| logfile.deinit();
    }

    pub fn set_thread_id(self: *Logger, thread_id: std.Thread.Id) void {
        switch (self.ctx) {
            .root => unreachable,
            .task => self.ctx.task.thread_id = thread_id,
        }
    }

    fn try_write(self: Logger, comptime sigil: u8, msg: []const u8) void {
        if (self.logfile == null) return;
        self.logfile.?.try_write(sigil, msg);
    }

    pub fn debug(self: Logger, comptime fmt: []const u8, args: anytype) void {
        if (!self.options.debug) return;
        var arr: [1024]u8 = undefined;
        const msg = std.fmt.bufPrint(@as([]u8, &arr), fmt, args) catch return;
        self.try_write('D', msg);
        std.log.debug("[{any}] {s}", .{ self.ctx, msg });
    }

    pub fn err(self: Logger, comptime fmt: []const u8, args: anytype) void {
        var arr: [1024]u8 = undefined;
        const msg = std.fmt.bufPrint(@as([]u8, &arr), fmt, args) catch return;
        self.try_write('E', msg);
        std.log.err("[{any}] {s}", .{ self.ctx, msg });
    }

    pub fn info(self: Logger, comptime fmt: []const u8, args: anytype) void {
        if (!self.options.verbose) return;
        var arr: [1024]u8 = undefined;
        const msg = std.fmt.bufPrint(@as([]u8, &arr), fmt, args) catch return;
        self.try_write('I', msg);
        std.log.info("[{any}] {s}", .{ self.ctx, msg });
    }

    pub fn warn(self: Logger, comptime fmt: []const u8, args: anytype) void {
        var arr: [1024]u8 = undefined;
        const msg = std.fmt.bufPrint(@as([]u8, &arr), fmt, args) catch return;
        self.try_write('W', msg);
        std.log.warn("[{any}] {s}", .{ self.ctx, msg });
    }
};
