const std = @import("std");

const core = @import("core.zig");

const Elem = core.Elem;

pub const EMPTY_RDB = &.{
    0x52, 0x45, 0x44, 0x49, 0x53, 0x30, 0x30, 0x31, 0x31, 0xfa, 0x09, 0x72, 0x65, 0x64, 0x69, 0x73,
    0x2d, 0x76, 0x65, 0x72, 0x05, 0x37, 0x2e, 0x32, 0x2e, 0x30, 0xfa, 0x0a, 0x72, 0x65, 0x64, 0x69,
    0x73, 0x2d, 0x62, 0x69, 0x74, 0x73, 0xc0, 0x40, 0xfa, 0x05, 0x63, 0x74, 0x69, 0x6d, 0x65, 0xc2,
    0x6d, 0x08, 0xbc, 0x65, 0xfa, 0x08, 0x75, 0x73, 0x65, 0x64, 0x2d, 0x6d, 0x65, 0x6d, 0xc2, 0xb0,
    0xc4, 0x10, 0x00, 0xfa, 0x08, 0x61, 0x6f, 0x66, 0x2d, 0x62, 0x61, 0x73, 0x65, 0xc0, 0x00, 0xff,
    0xf0, 0x6e, 0x3b, 0xfe, 0xc0, 0xff, 0x5a, 0xa2,
};

const ParsingError = error{
    BadHeader,
    BadOpCode,
    BadStringEncodingFlag,
    BadValueTypeFlag,
    ExpiryOutOfRange,
    ExtraData,
    ExtraItems,
    MissingData,
    MissingItems,
    MissingSizeInfo,
    IllegalSizeInfo,
    OutOfMemory,
    UnexpectedEncodingFlag,
    UnexpectedOpcode,
    UnsupportedFormat,
    UnsupportedStringEncoding,
    UnsupportedValueType,
    ValueTooLarge,
    ValueTooSmall,
};

const OpCode = enum(u8) {
    eof = 0xFF,
    db_sub = 0xFE,
    expiry_s = 0xFD,
    expiry_ms = 0xFC,
    db_sub_sizeinfo = 0xFB,
    attribute = 0xFA,

    fn select(val: u8) ParsingError!OpCode {
        return std.meta.intToEnum(OpCode, val) catch return ParsingError.BadOpCode;
    }
};

const ValueTypeFlag = enum(u8) {
    string = 0,
    list = 1,
    set = 2,
    sorted_set = 3,
    hash = 4,
    zipmap = 9,
    intset = 10,
    sorted_set_in_ziplist = 11,
    hashmap_in_ziplist = 12,
    list_in_quicklist = 13,

    fn select(val: u8) ParsingError!ValueTypeFlag {
        return std.meta.intToEnum(ValueTypeFlag, val) catch return ParsingError.BadValueTypeFlag;
    }
};

const Enc = enum(u8) {
    length_int6,
    length_int14,
    length_int32,
    string_int8,
    string_int16,
    string_int32,
    string_lzf,
};

const Byte0 = struct {
    head: u2,
    tail: u6,
    enc: Enc,

    fn _select_enc(head: u2, tail: u8) ParsingError!Enc {
        return switch (head) {
            0b00 => .length_int6,
            0b01 => .length_int14,
            0b10 => .length_int32,
            0b11 => switch (tail) {
                0 => .string_int8,
                1 => .string_int16,
                2 => .string_int32,
                3 => .string_lzf,
                else => return ParsingError.BadStringEncodingFlag,
            },
        };
    }

    fn from_byte(value: u8) ParsingError!Byte0 {
        const head: u2 = @intCast((value & 0b1100_0000) >> 6);
        const tail: u6 = @intCast(value & 0b0011_1111);
        return .{
            .head = head,
            .tail = tail,
            .enc = try Byte0._select_enc(head, tail),
        };
    }

    fn to_byte(self: Byte0) u8 {
        var val: u8 = self.head;
        val <<= 6;
        val += self.tail;
        return val;
    }
};

const Header = struct {
    format_version: u16,
};

const Footer = struct {
    checksum: ?[]const u8,
};

const MetadataAttribute = struct {
    name: []const u8,
    value: []const u8,
};

const DatabaseSub = struct {
    index: usize,
    values: std.StringHashMap(Elem),
    pxats: std.StringHashMap(i64),
};

const DatabaseSizeInfo = struct {
    values: usize,
    pxats: usize,
};

pub const File = struct {
    header: Header,
    metadata: []const MetadataAttribute,
    database: []const DatabaseSub,
    footer: Footer,
};

fn assert_range(comptime T: type, n: usize) ParsingError!void {
    const max = std.math.maxInt(T);
    const min = std.math.minInt(T);
    if (n > max) return ParsingError.ValueTooLarge;
    if (n < min) return ParsingError.ValueTooSmall;
}

fn assert_item_number(ref: usize, val: usize) ParsingError!void {
    if (val > ref) return ParsingError.ExtraItems;
    if (val < ref) return ParsingError.MissingItems;
}

fn str_contains_only_digits(haystack: []const u8) bool {
    const idx = std.mem.indexOfNone(u8, haystack, "0123456789");
    return (idx == null);
}

fn str_contains_only(haystack: []const u8, needle: []const u8) bool {
    const idx = std.mem.indexOfNone(u8, haystack, needle);
    return (idx == null);
}

const DatabaseRecord = struct {
    key: []const u8,
    elem: Elem,
    pxat: ?i64,
};

const IntFmt = union(enum) {
    int8,
    int16le,
    int16be,
    int32le,
    int32be,
    int64le,
    int64be,

    fn size(self: IntFmt) usize {
        return switch (self) {
            .int8 => 1,
            .int16le => 2,
            .int16be => 2,
            .int32le => 4,
            .int32be => 4,
            .int64le => 8,
            .int64be => 8,
        };
    }

    fn T(self: IntFmt) type {
        return switch (self) {
            .int8 => u8,
            .int16le => u16,
            .int16be => u16,
            .int32le => u32,
            .int32be => u32,
            .int64le => u64,
            .int64be => u64,
        };
    }
};

pub const Parser = struct {
    allocator: std.mem.Allocator,
    buff: []const u8,
    cur: usize = 0,

    pub fn init(buff: []const u8, allocator: std.mem.Allocator) Parser {
        return Parser{
            .allocator = allocator,
            .buff = buff,
        };
    }

    fn _assert_done(self: Parser) ParsingError!void {
        if (self.buff[self.cur..].len == 0) return;
        return ParsingError.ExtraData;
    }

    fn _debug_state(self: Parser, comptime loc: []const u8) void {
        std.log.debug("Parser.{s}():self.cur={d} self.buff={any}", .{ loc, self.cur, self.buff });
    }

    fn _debug(self: Parser, comptime fmt: []const u8, args: anytype) void {
        _ = self;
        std.log.debug(fmt, args);
    }

    fn _drop_magic(self: *Parser) ParsingError!void {
        const bytes = try self._pop_bytes(5);
        if (!std.mem.eql(u8, bytes, "REDIS")) return ParsingError.BadHeader;
    }

    fn _matchpop_opcode(self: *Parser, op_code: OpCode) bool {
        const value = @intFromEnum(op_code);
        if (self.buff.len <= self.cur) return false; // no more bytes
        if (self.buff[self.cur] != value) return false;
        self.cur += 1;
        return true;
    }

    fn _maybe_pop_extra_junk(self: *Parser) void {
        // apparently test error; see my comment:
        //    https://app.codecrafters.io/courses/redis/stages/jz6?repo=ac4e24cb-a3d0-4720-b62d-ab09905https://app.codecrafters.io/courses/redis/stages/jz6?repo=ac4e24cb-a3d0-4720-b62d-ab09905516d8516d8
        const rest = self.buff[self.cur..];
        if (rest.len == 1 and rest[0] == '\n') self.cur += 1;
    }

    fn _pop_byte0(self: *Parser) ParsingError!Byte0 {
        const byte = try self._pop_int(.int8);
        self._debug("_pop_byte0():byte={b:0>8}", .{byte});
        const b0 = Byte0.from_byte(byte);
        return b0;
    }

    fn _pop_bytes(self: *Parser, len: usize) ParsingError![]const u8 {
        if (self.buff.len < (self.cur + len)) return ParsingError.MissingData;
        const out = self.buff[self.cur..(self.cur + len)];
        self.cur = self.cur + len;
        return out;
    }

    fn _pop_int(self: *Parser, comptime fmt: IntFmt) ParsingError!fmt.T() {
        const bytes = try self._pop_bytes(fmt.size());
        switch (fmt) {
            .int8 => return bytes[0],
            .int16le => return std.mem.readVarInt(u16, bytes, .little),
            .int16be => return std.mem.readVarInt(u16, bytes, .big),
            .int32le => return std.mem.readVarInt(u32, bytes, .little),
            .int32be => return std.mem.readVarInt(u32, bytes, .big),
            .int64le => return std.mem.readVarInt(u64, bytes, .little),
            .int64be => return std.mem.readVarInt(u64, bytes, .big),
        }
    }

    fn _pop_opcode(self: *Parser) !OpCode {
        const byte = try self._pop_int(.int8);
        return try OpCode.select(byte);
    }

    fn _pop_vtf(self: *Parser) !ValueTypeFlag {
        const byte = try self._pop_int(.int8);
        return try ValueTypeFlag.select(byte);
    }

    fn _popelem_attribute(self: *Parser) ParsingError!MetadataAttribute {
        self._debug_state("_popelem_attribute");
        return MetadataAttribute{
            .name = try self._popfield_string(),
            .value = try self._popfield_string(),
        };
    }

    fn _popelem_database_record(self: *Parser) ParsingError!DatabaseRecord {
        const pxat = try self._popelem_pxat();
        const vtf = try self._pop_vtf();
        const key = try self._popfield_string();
        const elem = switch (vtf) {
            .string => Elem{ .str = try self._popfield_string() },
            else => return ParsingError.UnsupportedValueType,
        };
        return DatabaseRecord{
            .key = key,
            .elem = elem,
            .pxat = pxat,
        };
    }

    fn _popelem_database_sizeinfo(self: *Parser) ParsingError!DatabaseSizeInfo {
        const opcode = try self._pop_opcode();
        if (opcode != .db_sub_sizeinfo) return ParsingError.UnexpectedOpcode;
        self._debug_state("_popelem_database_sizeinfo");
        const values = try self._popfield_length();
        self._debug("_popelem_database_sizeinfo():values={}", .{values});
        const pxats = try self._popfield_length();
        self._debug("_popelem_database_sizeinfo():pxats={}", .{pxats});
        if (pxats > values) return ParsingError.IllegalSizeInfo;
        return DatabaseSizeInfo{
            .values = values,
            .pxats = pxats,
        };
    }

    fn _popelem_database_sub(self: *Parser) ParsingError!DatabaseSub {
        const index = try self._popfield_length();
        const sizeinfo = try self._popelem_database_sizeinfo();
        var elem_map = std.StringHashMap(Elem).init(self.allocator);
        var pxat_map = std.StringHashMap(i64).init(self.allocator);
        var values_seen: usize = 0;
        var pxats_seen: usize = 0;
        var record: DatabaseRecord = undefined;
        for (0..sizeinfo.values) |_| {
            record = try self._popelem_database_record();
            values_seen += 1;
            elem_map.put(record.key, record.elem) catch return ParsingError.OutOfMemory;
            if (record.pxat) |pxat| {
                pxats_seen += 1;
                pxat_map.put(record.key, pxat) catch return ParsingError.OutOfMemory;
            }
        }
        try assert_item_number(sizeinfo.values, values_seen);
        try assert_item_number(sizeinfo.pxats, pxats_seen);
        return DatabaseSub{
            .index = index,
            .values = elem_map,
            .pxats = pxat_map,
        };
    }

    fn _popelem_pxat(self: *Parser) ParsingError!?i64 {
        if (self._matchpop_opcode(.expiry_ms)) {
            const ms = try self._pop_int(.int64le);
            try assert_range(i64, ms);
            return @intCast(ms);
        }
        if (self._matchpop_opcode(.expiry_s)) {
            const s = try self._pop_int(.int32le);
            const ms = std.math.mul(u63, s, 1000) catch return ParsingError.ExpiryOutOfRange;
            try assert_range(i64, ms);
            return @intCast(ms);
        }
        return null;
    }

    fn _popfield_length(self: *Parser) ParsingError!usize {
        const byte0 = try self._pop_byte0();
        switch (byte0.enc) {
            .length_int6 => {
                return byte0.tail;
            },
            .length_int14 => {
                const bytes: [2]u8 = .{ byte0.tail, try self._pop_int(.int8) };
                return std.mem.readVarInt(usize, &bytes, .big);
            },
            .length_int32 => {
                return try self._pop_int(.int32be);
            },
            else => return ParsingError.UnexpectedEncodingFlag,
        }
    }

    fn _popfield_string(self: *Parser) ParsingError![]const u8 {
        const byte0 = try self._pop_byte0();
        var n: usize = 0;
        self._debug(
            "Byte0[_popfield_string()]:self.head={b:0>2} self.tail={b:0>6}, self.enc={}",
            .{ byte0.head, byte0.tail, byte0.enc },
        );
        switch (byte0.enc) {
            .string_int8, .string_int16, .string_int32 => {
                const int = switch (byte0.enc) {
                    .string_int8 => try self._pop_int(.int8),
                    .string_int16 => try self._pop_int(.int16le),
                    .string_int32 => try self._pop_int(.int32le),
                    else => unreachable,
                };
                var string_al = std.ArrayList(u8).init(self.allocator);
                try std.fmt.formatInt(int, 10, .lower, .{}, string_al.writer());
                const string = string_al.items;
                self._debug("Parser._popfield_string():string={s}", .{string});
                return string;
            },
            .length_int6 => {
                n = byte0.tail;
            },
            .length_int14 => {
                const bytes: [2]u8 = .{ byte0.tail, try self._pop_int(.int8) };
                n = std.mem.readVarInt(usize, &bytes, .big);
            },
            .length_int32 => {
                n = try self._pop_int(.int32be);
            },
            .string_lzf => return ParsingError.UnsupportedStringEncoding,
        }
        self._debug("Parser._popfield_string():n={d}", .{n});
        const str = try self._pop_bytes(n);
        self._debug("Parser._popfield_string():str={s}", .{str});
        return str;
    }

    fn _popsct_database(self: *Parser) ParsingError![]const DatabaseSub {
        var dbs_al = std.ArrayList(DatabaseSub).init(self.allocator);
        while (self._matchpop_opcode(.db_sub)) {
            const db_sub = try self._popelem_database_sub();
            dbs_al.append(db_sub) catch return ParsingError.OutOfMemory;
        }
        return dbs_al.items;
    }

    fn _popsct_footer(self: *Parser) ParsingError!Footer {
        const opcode = try self._pop_opcode();
        if (opcode != .eof) return ParsingError.UnexpectedOpcode;
        return Footer{
            .checksum = blk: {
                const bytes = try self._pop_bytes(8);
                if (str_contains_only(bytes, &.{0x00})) break :blk null;
                break :blk bytes;
            },
        };
    }

    fn _popsct_header(self: *Parser) ParsingError!Header {
        try self._drop_magic();
        const format_version = try self._pop_version();
        if (format_version != 11 and format_version != 10) return ParsingError.UnsupportedFormat;
        return .{
            .format_version = format_version,
        };
    }

    fn _popsct_metadata(self: *Parser) ParsingError![]const MetadataAttribute {
        var attrs_al = std.ArrayList(MetadataAttribute).init(self.allocator);
        while (self._matchpop_opcode(.attribute)) {
            const attribute = try self._popelem_attribute();
            attrs_al.append(attribute) catch return ParsingError.OutOfMemory;
        }
        return attrs_al.items;
    }

    fn _pop_version(self: *Parser) ParsingError!u16 {
        const bytes = try self._pop_bytes(4);
        if (!str_contains_only_digits(bytes)) return ParsingError.BadHeader;
        const num = std.fmt.parseInt(u16, bytes, 10) catch return ParsingError.BadHeader;
        return num;
    }

    pub fn parse(self: *Parser) ParsingError!File {
        const file = File{
            .header = try self._popsct_header(),
            .metadata = try self._popsct_metadata(),
            .database = try self._popsct_database(),
            .footer = try self._popsct_footer(),
        };
        self._maybe_pop_extra_junk();
        try self._assert_done();
        return file;
    }
};

const Writer = struct {
    allocator: std.mem.Allocator(),
    writer: std.io.AnyWriter,

    fn write_header(self: Writer) !usize {
        const bytes = "REDIS0011";
        try self.writer.writeAll(bytes);
        return bytes.len;
    }

    //  fn write_metadata(self: Writer, attributes: []const MetadataAttribute) !usize {
    //      var bytes_written = 0;
    //      for (attributes) |attribute| {
    //          var bytes: []u8 = encode_attribute(attribute, self.allocator);
    //          bytes_written+=bytes.len
    //          try self._write_attribute(attribute);
    //      }
    //  }

};

fn encode_size(value: u32, allocator: std.mem.Allocator) []const u8 {
    const MAX_6BIT = std.math.maxInt(u6);
    const MAX_14BIT = std.math.maxInt(u14);
    const MAX_32BIT = std.math.maxInt(u32);
    if (value <= MAX_6BIT) {
        var bytes = try allocator.alloc(u8, 1);
        const byte0 = Byte0{ .head = 0x00, .tail = value };
        bytes[0] = byte0.to_byte();
        return bytes;
    }
    if (value <= MAX_14BIT) {
        var bytes = try allocator.alloc(u8, 2);
        std.mem.writePackedInt(u14, &bytes, 2, value, .big);
        //   0010   0110
        // | 0100   0100
        //   0110   0100
        const byte0_tail: u6 = (bytes[0] & 0b0011_1111) << 2;
        const byte0 = Byte0{ .head = 0b01, .tail = byte0_tail };
        return &.{ byte0.to_byte(), bytes[1] };
    }
    if (value <= MAX_32BIT) {
        var bytes: [5]u8 = .{0x00} ** 5;
        var payload_bytes: [4]u8 = .{0x00} ** 4;
        std.mem.writePackedInt(u32, &payload_bytes, 0, value, .big);
        bytes[0] == 0b1000_0000;
        inline for (payload_bytes, 0..) |byte, idx| bytes[idx + 1] = byte;
        return &payload_bytes;
    }
    unreachable;
}

const ParserTestCase = union(enum) {
    int_p: ParsingError!usize,
    length_f: ParsingError!usize,
    sint_f: ParsingError!usize,
    string_f: ParsingError![]const u8,
    metadata_s: ParsingError![]const MetadataAttribute,
    header_s: ParsingError!Header,
    database_s: ParsingError![]const DatabaseSub,
    database_sizeinfo_e: ParsingError!DatabaseSizeInfo,
    database_record_e: ParsingError!DatabaseRecord,
    footer_s: ParsingError!Footer,
    file_x: ParsingError!File,

    /// Utility to create easy memory-irresponsible parser instance
    fn _init_test_parser(buff: []const u8) Parser {
        return Parser.init(
            buff,
            std.heap.page_allocator,
        );
    }

    // header_s

    fn init_header_s(buff: []const u8) ParserTestCase {
        var parser = ParserTestCase._init_test_parser(buff);
        return ParserTestCase{ .header_s = parser._popsct_header() };
    }

    fn expect_header(self: ParserTestCase, oracle: Header) !void {
        switch (self) {
            .header_s => |rv| try std.testing.expectEqualDeep(oracle, try rv),
            else => unreachable,
        }
    }

    // metadata_s

    fn init_metadata_s(buff: []const u8) ParserTestCase {
        var parser = ParserTestCase._init_test_parser(buff);
        return ParserTestCase{ .metadata_s = parser._popsct_metadata() };
    }

    fn expect_attribute_at(self: ParserTestCase, idx: usize, oracle: MetadataAttribute) !void {
        switch (self) {
            .metadata_s => |rv| {
                const attrs = try rv;
                std.testing.expect(attrs.len > idx) catch return error.NotEnoughElements;
                try std.testing.expectEqualDeep(oracle, attrs[idx]);
            },
            else => unreachable,
        }
    }

    // database_sizeinfo_e

    fn init_database_sizeinfo_e(buff: []const u8) ParserTestCase {
        var parser = ParserTestCase._init_test_parser(buff);
        return ParserTestCase{ .database_sizeinfo_e = parser._popelem_database_sizeinfo() };
    }

    fn expect_sizeinfo(self: ParserTestCase, oracle: DatabaseSizeInfo) !void {
        switch (self) {
            .database_sizeinfo_e => |rv| try std.testing.expectEqualDeep(oracle, rv),
            else => unreachable,
        }
    }

    // database_record_e

    fn init_database_record_e(buff: []const u8) ParserTestCase {
        var parser = ParserTestCase._init_test_parser(buff);
        return ParserTestCase{ .database_record_e = parser._popelem_database_record() };
    }

    fn expect_record(self: ParserTestCase, oracle: DatabaseRecord) !void {
        switch (self) {
            .database_record_e => |rv| try std.testing.expectEqualDeep(oracle, rv),
            else => unreachable,
        }
    }

    // database_s

    fn init_database_s(buff: []const u8) ParserTestCase {
        var parser = ParserTestCase._init_test_parser(buff);
        return ParserTestCase{ .metadata_s = parser._popsct_database() };
    }

    fn expect_record_at(self: ParserTestCase, idx: usize, oracle: DatabaseRecord) !void {
        switch (self) {
            .database_s => |rv| {
                const recs = try rv;
                std.testing.expect(recs.len > idx) catch return error.NotEnoughElements;
                try std.testing.expectEqualDeep(oracle, recs[idx]);
            },
            else => unreachable,
        }
    }

    // footer_s

    fn init_footer_s(buff: []const u8) ParserTestCase {
        var parser = ParserTestCase._init_test_parser(buff);
        return ParserTestCase{ .footer_s = parser._popsct_footer() };
    }

    fn expect_footer(self: ParserTestCase, oracle: Footer) !void {
        switch (self) {
            .footer_s => |rv| try std.testing.expectEqualDeep(oracle, try rv),
            else => unreachable,
        }
    }

    // file_x

    fn init_file_x(buff: []const u8) ParserTestCase {
        var parser = ParserTestCase._init_test_parser(buff);
        return ParserTestCase{ .file_x = parser.parse() };
    }

    fn expect_file(self: ParserTestCase, oracle: File) !void {
        switch (self) {
            .file_x => |rv| try std.testing.expectEqualDeep(oracle, try rv),
            else => unreachable,
        }
    }

    // string_f

    fn init_string_f(buff: []const u8) ParserTestCase {
        var parser = ParserTestCase._init_test_parser(buff);
        return ParserTestCase{ .string_f = parser._popfield_string() };
    }

    // int_p

    fn init_int_p(comptime fmt: IntFmt, buff: []const u8) ParserTestCase {
        var parser = ParserTestCase._init_test_parser(buff);
        const rv = parser._pop_int(fmt);
        const num = rv catch |e| return ParserTestCase{ .int_p = e };
        return ParserTestCase{ .int_p = @intCast(num) };
    }

    // length_f

    fn init_length_f(buff: []const u8) ParserTestCase {
        var parser = ParserTestCase._init_test_parser(buff);
        return ParserTestCase{ .length_f = parser._popfield_length() };
    }

    // COMMON

    fn expect_error(self: ParserTestCase, oracle: ParsingError) !void {
        switch (self) {
            inline else => |rv| try std.testing.expectEqual(oracle, rv),
        }
    }

    fn expect_int(self: ParserTestCase, oracle: usize) !void {
        switch (self) {
            .int_p, .sint_f, .length_f => |rv| try std.testing.expectEqual(oracle, try rv),
            else => unreachable,
        }
    }

    fn expect_len(self: ParserTestCase, oracle: usize) !void {
        switch (self) {
            .metadata_s => |rv| try std.testing.expectEqual(oracle, (try rv).len),
            .string_f => |rv| try std.testing.expectEqual(oracle, (try rv).len),
            .database_s => |rv| try std.testing.expectEqual(oracle, (try rv).len),
            inline else => unreachable,
        }
    }

    fn expect_str(self: ParserTestCase, oracle: []const u8) !void {
        switch (self) {
            .string_f => |rv| try std.testing.expectEqualStrings(oracle, try rv),
            inline else => unreachable,
        }
    }
};

test "Parser._pop_int" {
    const mkcase = ParserTestCase.init_int_p;

    try mkcase(.int8, &.{}).expect_error(ParsingError.MissingData);
    try mkcase(.int8, &.{0x00}).expect_int(0);
    try mkcase(.int8, &.{0x01}).expect_int(1);
    try mkcase(.int8, &.{0xFE}).expect_int(254);
    try mkcase(.int8, &.{0xFF}).expect_int(255);

    try mkcase(.int16be, &.{}).expect_error(ParsingError.MissingData);
    try mkcase(.int16be, &.{0x00}).expect_error(ParsingError.MissingData);
    try mkcase(.int16be, &.{ 0x00, 0x00 }).expect_int(0);
    try mkcase(.int16be, &.{ 0x00, 0x01 }).expect_int(1);
    try mkcase(.int16be, &.{ 0xFF, 0xFE }).expect_int(65534);
    try mkcase(.int16be, &.{ 0xFF, 0xFF }).expect_int(65535);

    try mkcase(.int16le, &.{}).expect_error(ParsingError.MissingData);
    try mkcase(.int16le, &.{0x00}).expect_error(ParsingError.MissingData);
    try mkcase(.int16le, &.{ 0x00, 0x00 }).expect_int(0);
    try mkcase(.int16le, &.{ 0x01, 0x00 }).expect_int(1);
    try mkcase(.int16le, &.{ 0xFE, 0xFF }).expect_int(65534);
    try mkcase(.int16le, &.{ 0xFF, 0xFF }).expect_int(65535);

    try mkcase(.int32be, &.{}).expect_error(ParsingError.MissingData);
    try mkcase(.int32be, &.{0x00}).expect_error(ParsingError.MissingData);
    try mkcase(.int32be, &.{ 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(.int32be, &.{ 0x00, 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(.int32be, &.{ 0x00, 0x00, 0x00, 0x00 }).expect_int(0);
    try mkcase(.int32be, &.{ 0x00, 0x00, 0x00, 0x01 }).expect_int(1);
    try mkcase(.int32be, &.{ 0xFF, 0xFF, 0xFF, 0xFE }).expect_int(4294967294);
    try mkcase(.int32be, &.{ 0xFF, 0xFF, 0xFF, 0xFF }).expect_int(4294967295);

    try mkcase(.int32le, &.{}).expect_error(ParsingError.MissingData);
    try mkcase(.int32le, &.{0x00}).expect_error(ParsingError.MissingData);
    try mkcase(.int32le, &.{ 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(.int32le, &.{ 0x00, 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(.int32le, &.{ 0x00, 0x00, 0x00, 0x00 }).expect_int(0);
    try mkcase(.int32le, &.{ 0x01, 0x00, 0x00, 0x00 }).expect_int(1);
    try mkcase(.int32le, &.{ 0xFE, 0xFF, 0xFF, 0xFF }).expect_int(4294967294);
    try mkcase(.int32le, &.{ 0xFF, 0xFF, 0xFF, 0xFF }).expect_int(4294967295);

    try mkcase(.int64be, &.{0x00}).expect_error(ParsingError.MissingData);
    try mkcase(.int64be, &.{ 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(.int64be, &.{ 0x00, 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(.int64be, &.{ 0x00, 0x00, 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(.int64be, &.{ 0x00, 0x00, 0x00, 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(.int64be, &.{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(.int64be, &.{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(.int64be, &.{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }).expect_int(0);
    try mkcase(.int64be, &.{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01 }).expect_int(1);
    try mkcase(.int64be, &.{ 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFE }).expect_int(18446744073709551614);
    try mkcase(.int64be, &.{ 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF }).expect_int(18446744073709551615);

    try mkcase(.int64le, &.{0x00}).expect_error(ParsingError.MissingData);
    try mkcase(.int64le, &.{ 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(.int64le, &.{ 0x00, 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(.int64le, &.{ 0x00, 0x00, 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(.int64le, &.{ 0x00, 0x00, 0x00, 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(.int64le, &.{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(.int64le, &.{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(.int64le, &.{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }).expect_int(0);
    try mkcase(.int64le, &.{ 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }).expect_int(1);
    try mkcase(.int64le, &.{ 0xFE, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF }).expect_int(18446744073709551614);
    try mkcase(.int64le, &.{ 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF }).expect_int(18446744073709551615);
}

test "Parser._popfield_length" {
    const mkcase = ParserTestCase.init_length_f;
    try mkcase(&.{}).expect_error(ParsingError.MissingData);
    try mkcase(&.{0b0000_0000}).expect_int(0);
    try mkcase(&.{0b0000_0001}).expect_int(1);
    try mkcase(&.{0b0011_1110}).expect_int(62);
    try mkcase(&.{0b0011_1111}).expect_int(63);
    try mkcase(&.{0b0100_0000}).expect_error(ParsingError.MissingData);
    try mkcase(&.{ 0b0100_0000, 0b0000_0000 }).expect_int(0);
    try mkcase(&.{ 0b0100_0000, 0b0000_0001 }).expect_int(1);
    try mkcase(&.{ 0b0111_1111, 0b1111_1110 }).expect_int(16382);
    try mkcase(&.{ 0b0111_1111, 0b1111_1111 }).expect_int(16383);
    try mkcase(&.{ 0b1000_0000, 0b0000_0000 }).expect_error(ParsingError.MissingData);
    try mkcase(&.{ 0b1000_0000, 0b0000_0000, 0b0000_0000 }).expect_error(ParsingError.MissingData);
    try mkcase(&.{ 0b1000_0000, 0b0000_0000, 0b0000_0000, 0b0000_0000 }).expect_error(ParsingError.MissingData);
    try mkcase(&.{ 0b1000_0000, 0b0000_0000, 0b0000_0000, 0b0000_0000, 0b0000_0000 }).expect_int(0);
    try mkcase(&.{ 0b1000_0000, 0b0000_0000, 0b0000_0000, 0b0000_0000, 0b0000_0001 }).expect_int(1);
    try mkcase(&.{ 0b1000_0000, 0b1111_1111, 0b1111_1111, 0b1111_1111, 0b1111_1110 }).expect_int(4294967294);
    try mkcase(&.{ 0b1000_0000, 0b1111_1111, 0b1111_1111, 0b1111_1111, 0b1111_1111 }).expect_int(4294967295);
    try mkcase(&.{0b1100_0000}).expect_error(ParsingError.UnexpectedEncodingFlag);
    try mkcase(&.{0b1111_1111}).expect_error(ParsingError.BadStringEncodingFlag);
}

test "Parser._popfield_sint" {}

const TestData = struct {
    data: []const u8,
    oracle: []const u8,

    fn init(comptime p: []const u8, comptime n: comptime_int) !TestData {
        var heap = std.heap.ArenaAllocator.init(std.heap.page_allocator);
        const base = [_]u8{ 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J' };
        var data_al = std.ArrayList(u8).init(heap.allocator());
        var oracle_al = std.ArrayList(u8).init(heap.allocator());
        try data_al.appendSlice(p);
        for (0..n) |idx| {
            const char_idx = @mod(idx, 10);
            try data_al.append(base[char_idx]);
            try oracle_al.append(base[char_idx]);
        }
        return TestData{
            .data = data_al.items,
            .oracle = oracle_al.items,
        };
    }
};

test "Parser._popfield_string" {
    const mkcase = ParserTestCase.init_string_f;
    try mkcase(&.{}).expect_error(ParsingError.MissingData);
    try mkcase(&.{0b0000_0000}).expect_str("");
    try mkcase(&.{ 0b0000_0001, 'A' }).expect_str("A");
    try mkcase(&.{ 0b0000_0010, 'A', 'B' }).expect_str("AB");
    try mkcase(&.{ 0b0000_0010, 'A' }).expect_error(ParsingError.MissingData);

    try mkcase(&.{0b1100_0000}).expect_error(ParsingError.MissingData);
    try mkcase(&.{ 0b1100_0000, 0b0000_0000 }).expect_str("0");
    try mkcase(&.{ 0b1100_0000, 0b0000_0001 }).expect_str("1");
    try mkcase(&.{ 0b1100_0000, 0b1111_1110 }).expect_str("254");
    try mkcase(&.{ 0b1100_0000, 0b1111_1111 }).expect_str("255");

    try mkcase(&.{0b1100_0001}).expect_error(ParsingError.MissingData);
    try mkcase(&.{ 0b1100_0001, 0b0000_0000 }).expect_error(ParsingError.MissingData);
    try mkcase(&.{ 0b1100_0001, 0b0000_0000, 0b0000_0000 }).expect_str("0");
    try mkcase(&.{ 0b1100_0001, 0b0000_0001, 0b0000_0000 }).expect_str("1");
    try mkcase(&.{ 0b1100_0001, 0b1111_1110, 0b1111_1111 }).expect_str("65534");
    try mkcase(&.{ 0b1100_0001, 0b1111_1111, 0b1111_1111 }).expect_str("65535");

    try mkcase(&.{0b1100_0001}).expect_error(ParsingError.MissingData);
    try mkcase(&.{ 0b1100_0010, 0b0000_0000 }).expect_error(ParsingError.MissingData);
    try mkcase(&.{ 0b1100_0010, 0b0000_0000, 0b0000_0000 }).expect_error(ParsingError.MissingData);
    try mkcase(&.{ 0b1100_0010, 0b0000_0000, 0b0000_0000, 0b0000_0000 }).expect_error(ParsingError.MissingData);
    try mkcase(&.{ 0b1100_0010, 0b0000_0000, 0b0000_0000, 0b0000_0000, 0b0000_0000 }).expect_str("0");
    try mkcase(&.{ 0b1100_0010, 0b0000_0001, 0b0000_0000, 0b0000_0000, 0b0000_0000 }).expect_str("1");
    try mkcase(&.{ 0b1100_0010, 0b1111_1110, 0b1111_1111, 0b1111_1111, 0b1111_1111 }).expect_str("4294967294");
    try mkcase(&.{ 0b1100_0010, 0b1111_1111, 0b1111_1111, 0b1111_1111, 0b1111_1111 }).expect_str("4294967295");

    try mkcase(&.{0b1100_0011}).expect_error(ParsingError.UnsupportedStringEncoding); // Compressed Strings (not supported)

    var td: TestData = undefined;
    td = try TestData.init(&.{0b0011_1111}, 63);
    try mkcase(td.data).expect_str(td.oracle);
}

test "Parser._popsct_header" {
    const mkcase = ParserTestCase.init_header_s;
    try mkcase("").expect_error(ParsingError.MissingData);
    try mkcase("X").expect_error(ParsingError.MissingData);
    try mkcase("XY").expect_error(ParsingError.MissingData);
    try mkcase("FOOOO").expect_error(ParsingError.BadHeader);
    try mkcase("REDIS").expect_error(ParsingError.MissingData);
    try mkcase("REDISX").expect_error(ParsingError.MissingData);
    try mkcase("REDISXY").expect_error(ParsingError.MissingData);
    try mkcase("REDISXYZ0").expect_error(ParsingError.BadHeader);
    try mkcase("REDIS9999").expect_error(ParsingError.UnsupportedFormat);
    try mkcase("REDIS0011").expect_header(Header{ .format_version = 11 });
}

test "Parser._popsct_metadata" {
    const mkcase = ParserTestCase.init_metadata_s;
    try mkcase("").expect_len(0);
    try mkcase("X").expect_len(0);
    try mkcase("XY").expect_len(0);
    try mkcase(&.{0xFA}).expect_error(ParsingError.MissingData); // by spec, there must be at least 1 attr
    try mkcase(&.{ 0xFA, 0x00 }).expect_error(ParsingError.MissingData);

    try mkcase(&.{ 0xFA, 0x00, 0x00 }).expect_attribute_at(0, MetadataAttribute{ .name = "", .value = "" });
    try mkcase(&.{ 0xFA, 0x00, 0x01, 'A' }).expect_attribute_at(0, MetadataAttribute{ .name = "", .value = "A" });
    try mkcase(&.{ 0xFA, 0x00, 0x02, 'A', 'B' }).expect_attribute_at(0, MetadataAttribute{ .name = "", .value = "AB" });

    try mkcase(&.{ 0xFA, 0x01, 'A', 0x00 }).expect_attribute_at(0, MetadataAttribute{ .name = "A", .value = "" });
    try mkcase(&.{ 0xFA, 0x01, 'A', 0x01, 'A' }).expect_attribute_at(0, MetadataAttribute{ .name = "A", .value = "A" });
    try mkcase(&.{ 0xFA, 0x01, 'A', 0x02, 'A', 'B' }).expect_attribute_at(0, MetadataAttribute{ .name = "A", .value = "AB" });
    try mkcase(&.{ 0xFA, 0x01, 'A', 0xC0, 0x7B }).expect_attribute_at(0, MetadataAttribute{ .name = "A", .value = "123" });
    try mkcase(&.{ 0xFA, 0x01, 'A', 0xC1, 0x39, 0x30 }).expect_attribute_at(0, MetadataAttribute{ .name = "A", .value = "12345" });
    try mkcase(&.{ 0xFA, 0x01, 'A', 0xC2, 0x87, 0xD6, 0x12, 0x00 }).expect_attribute_at(0, MetadataAttribute{ .name = "A", .value = "1234567" });

    try mkcase(&.{ 0xFA, 0x02, 'A', 'B', 0x00 }).expect_attribute_at(0, MetadataAttribute{ .name = "AB", .value = "" });
    try mkcase(&.{ 0xFA, 0x02, 'A', 'B', 0x01, 'A' }).expect_attribute_at(0, MetadataAttribute{ .name = "AB", .value = "A" });
    try mkcase(&.{ 0xFA, 0x02, 'A', 'B', 0x02, 'A', 'B' }).expect_attribute_at(0, MetadataAttribute{ .name = "AB", .value = "AB" });

    try mkcase(&.{ 0xFA, 0x01, 'A', 'X', 0x00 }).expect_error(ParsingError.MissingData); // 'X' (0x58) is interpreted as 14bit size of next element being .. a lot.

    const complex_a1 = &[_]u8{ 0xFA, 0x01, 'A', 0xC0, 0x7B };
    const complex_a2 = &[_]u8{ 0xFA, 0x02, 'B', 'C', 0x02, 'D', 'E' };
    const complex_a3 = &[_]u8{ 0xFA, 0xC0, 0x11, 0x02, 'F', 'G' };
    const complex = complex_a1 ++ complex_a2 ++ complex_a3;
    try mkcase(complex).expect_len(3);
    try mkcase(complex).expect_attribute_at(0, MetadataAttribute{ .name = "A", .value = "123" });
    try mkcase(complex).expect_attribute_at(1, MetadataAttribute{ .name = "BC", .value = "DE" });
    try mkcase(complex).expect_attribute_at(2, MetadataAttribute{ .name = "17", .value = "FG" });
}

test "Parser._popelem_database_sizeinfo" {
    const mkcase = ParserTestCase.init_database_sizeinfo_e;

    try mkcase("").expect_error(ParsingError.MissingData);
    try mkcase("X").expect_error(ParsingError.BadOpCode);
    try mkcase(&.{0xFD}).expect_error(ParsingError.UnexpectedOpcode);
    try mkcase(&.{0xFB}).expect_error(ParsingError.MissingData);

    try mkcase(&.{ 0xFB, 0x00, 0x00 }).expect_sizeinfo(DatabaseSizeInfo{ .values = 0, .pxats = 0 });
    try mkcase(&.{ 0xFB, 0x00, 0x01 }).expect_error(ParsingError.IllegalSizeInfo);
    try mkcase(&.{ 0xFB, 0x01, 0x00 }).expect_sizeinfo(DatabaseSizeInfo{ .values = 1, .pxats = 0 });
    try mkcase(&.{ 0xFB, 0x01, 0x01 }).expect_sizeinfo(DatabaseSizeInfo{ .values = 1, .pxats = 1 });

    const max_length: []const u8 = &.{ 0x80, 0xFF, 0xFF, 0xFF, 0xFF };
    const min_length: []const u8 = &.{0x00};
    try mkcase(&.{ 0xFB, 0x00, 0x00 }).expect_sizeinfo(DatabaseSizeInfo{ .values = 0, .pxats = 0 });
    try mkcase(&[_]u8{0xFB} ++ min_length ++ max_length).expect_error(ParsingError.IllegalSizeInfo);
    try mkcase(&[_]u8{0xFB} ++ max_length ++ min_length).expect_sizeinfo(DatabaseSizeInfo{ .values = 4294967295, .pxats = 0 });
    try mkcase(&[_]u8{0xFB} ++ max_length ++ max_length).expect_sizeinfo(DatabaseSizeInfo{ .values = 4294967295, .pxats = 4294967295 });
}

test "Parser._popelem_database_record" {
    const mkcase = ParserTestCase.init_database_record_e;
    try mkcase("").expect_error(ParsingError.MissingData);
    try mkcase("X").expect_error(ParsingError.BadValueTypeFlag);
    try mkcase(&.{0x00}).expect_error(ParsingError.MissingData);
    try mkcase(&.{0x01}).expect_error(ParsingError.MissingData); // UnsupportedValueType would be raised later

    // empty key
    try mkcase(
        &.{ 0x00, 0x00, 0x00 },
    ).expect_record(DatabaseRecord{ .key = "", .elem = Elem{ .str = "" }, .pxat = null });
    try mkcase(
        &.{ 0x00, 0x00, 0x01, 'B' },
    ).expect_record(DatabaseRecord{ .key = "", .elem = Elem{ .str = "B" }, .pxat = null });

    // "normal" data
    try mkcase(
        &.{ 0x00, 0x01, 'A', 0x00 },
    ).expect_record(DatabaseRecord{ .key = "A", .elem = Elem{ .str = "" }, .pxat = null });
    try mkcase(
        &.{ 0x00, 0x01, 'A', 0x01, 'B' },
    ).expect_record(DatabaseRecord{ .key = "A", .elem = Elem{ .str = "B" }, .pxat = null });
    try mkcase(
        &.{ 0x00, 0x01, 'A', 0x02, 'B', 'C' },
    ).expect_record(DatabaseRecord{ .key = "A", .elem = Elem{ .str = "BC" }, .pxat = null });
    try mkcase(
        &.{ 0x00, 0x02, 'A', 'B', 0x02, 'C', 'D' },
    ).expect_record(DatabaseRecord{ .key = "AB", .elem = Elem{ .str = "CD" }, .pxat = null });

    // expiry_s   vvvvvvvvvvvvvvvvvvvvvv (int16le)
    try mkcase(
        &.{ 0xFD, 0x01, 0x00, 0x00, 0x00, 0x00, 0x01, 'A', 0x01, 'B' },
    ).expect_record(DatabaseRecord{ .key = "A", .elem = Elem{ .str = "B" }, .pxat = 1000 });

    // expiry_ms  vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv (int32le)
    try mkcase(
        &.{ 0xFC, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 'A', 0x01, 'B' },
    ).expect_record(
        DatabaseRecord{ .key = "A", .elem = Elem{ .str = "B" }, .pxat = 1 },
    );
}

test "Parser._popsct_footer" {
    const mkcase = ParserTestCase.init_footer_s;
    try mkcase("").expect_error(ParsingError.MissingData);
    try mkcase("X").expect_error(ParsingError.BadOpCode);
    try mkcase(&.{0xFB}).expect_error(ParsingError.UnexpectedOpcode);
    try mkcase(&.{0xFF}).expect_error(ParsingError.MissingData);
    try mkcase(&.{ 0xFF, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(&.{ 0xFF, 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(&.{ 0xFF, 0x00, 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(&.{ 0xFF, 0x00, 0x00, 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(&.{ 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(&.{ 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(&.{ 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }).expect_error(ParsingError.MissingData);
    try mkcase(&.{ 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }).expect_footer(Footer{ .checksum = null });
    try mkcase(&.{ 0xFF, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48 }).expect_footer(Footer{ .checksum = "ABCDEFGH" });
}

fn mkfile1() !File {
    const checksum = &.{ 0xa0, 0x8a, 0x4c, 0xf6, 0x71, 0x24, 0xb8, 0x79 };
    var heap = std.heap.ArenaAllocator.init(std.heap.page_allocator);
    var metadata_al = std.ArrayList(MetadataAttribute).init(heap.allocator());
    var database_al = std.ArrayList(DatabaseSub).init(heap.allocator());
    var values_map = std.StringHashMap(Elem).init(heap.allocator());
    var pxats_map = std.StringHashMap(i64).init(heap.allocator());
    try metadata_al.append(MetadataAttribute{ .name = "redis-ver", .value = "7.0.15" });
    try metadata_al.append(MetadataAttribute{ .name = "redis-bits", .value = "64" });
    try metadata_al.append(MetadataAttribute{ .name = "ctime", .value = "1729390922" });
    try metadata_al.append(MetadataAttribute{ .name = "used-mem", .value = "925712" });
    try metadata_al.append(MetadataAttribute{ .name = "aof-base", .value = "0" });
    try values_map.put("defed", Elem{ .str = "defeddefed" });
    try pxats_map.put("foo", 1729811285000);
    try database_al.append(DatabaseSub{
        .index = 0,
        .values = values_map,
        .pxats = pxats_map,
    });
    return File{
        .header = Header{ .format_version = 11 },
        .metadata = metadata_al.items,
        .database = database_al.items,
        .footer = Footer{ .checksum = checksum },
    };
}

fn countValueKeys(database_sub: DatabaseSub) usize {
    var n: usize = 0;
    var it = database_sub.values.keyIterator();
    while (it.next()) |_| n += 1;
    return n;
}

fn countPxatKeys(database_sub: DatabaseSub) usize {
    var n: usize = 0;
    var it = database_sub.pxats.keyIterator();
    while (it.next()) |_| n += 1;
    return n;
}

test "Parser.parse" {
    const mkcase = ParserTestCase.init_file_x;
    try mkcase("").expect_error(ParsingError.MissingData);

    const data1 = &.{
        0x52, 0x45, 0x44, 0x49, 0x53, 0x30, 0x30, 0x31, 0x31, 0xfa, 0x09, 0x72, 0x65, 0x64, 0x69, 0x73,
        0x2d, 0x76, 0x65, 0x72, 0x06, 0x37, 0x2e, 0x30, 0x2e, 0x31, 0x35, 0xfa, 0x0a, 0x72, 0x65, 0x64,
        0x69, 0x73, 0x2d, 0x62, 0x69, 0x74, 0x73, 0xc0, 0x40, 0xfa, 0x05, 0x63, 0x74, 0x69, 0x6d, 0x65,
        0xc2, 0x4a, 0x69, 0x14, 0x67, 0xfa, 0x08, 0x75, 0x73, 0x65, 0x64, 0x2d, 0x6d, 0x65, 0x6d, 0xc2,
        0x10, 0x20, 0x0e, 0x00, 0xfa, 0x08, 0x61, 0x6f, 0x66, 0x2d, 0x62, 0x61, 0x73, 0x65, 0xc0, 0x00,
        0xfe, 0x00, 0xfb, 0x01, 0x00, 0x00, 0x05, 0x64, 0x65, 0x66, 0x65, 0x64, 0x0a, 0x64, 0x65, 0x66,
        0x65, 0x64, 0x64, 0x65, 0x66, 0x65, 0x64, 0xff, 0xa0, 0x8a, 0x4c, 0xf6, 0x71, 0x24, 0xb8, 0x79,
    };
    const checksum1 = &.{ 0xa0, 0x8a, 0x4c, 0xf6, 0x71, 0x24, 0xb8, 0x79 };
    var parser1 = ParserTestCase._init_test_parser(data1);
    const file1: File = try parser1.parse();
    try std.testing.expectEqual(11, file1.header.format_version);
    try std.testing.expectEqual(5, file1.metadata.len);
    try std.testing.expectEqualStrings("redis-ver", file1.metadata[0].name);
    try std.testing.expectEqualStrings("7.0.15", file1.metadata[0].value);
    try std.testing.expectEqualStrings("redis-bits", file1.metadata[1].name);
    try std.testing.expectEqualStrings("64", file1.metadata[1].value);
    try std.testing.expectEqualStrings("ctime", file1.metadata[2].name);
    try std.testing.expectEqualStrings("1729390922", file1.metadata[2].value);
    try std.testing.expectEqualStrings("used-mem", file1.metadata[3].name);
    try std.testing.expectEqualStrings("925712", file1.metadata[3].value);
    try std.testing.expectEqualStrings("aof-base", file1.metadata[4].name);
    try std.testing.expectEqualStrings("0", file1.metadata[4].value);
    try std.testing.expectEqual(1, file1.database.len);
    try std.testing.expectEqual(0, file1.database[0].index);
    try std.testing.expectEqual(1, countValueKeys(file1.database[0]));
    try std.testing.expectEqual(0, countPxatKeys(file1.database[0]));
    try std.testing.expectEqualDeep(Elem{ .str = "defeddefed" }, file1.database[0].values.get("defed"));
    try std.testing.expectEqualStrings(checksum1, file1.footer.checksum orelse "NONE");
}
