const std = @import("std");

const core = @import("core.zig");
const util = @import("../util.zig");

const ConnectionHandle = @import("server.zig").Server.ConnectionPool.Handle;
const Server = @import("server.zig").Server;
const ServerConfig = @import("server.zig").Config;
const Task = @import("task.zig").Task;
const HandlerResult = @import("worker.zig").HandlerResult;
const EMPTY_RDB = @import("rdb.zig").EMPTY_RDB;

pub const Capability = enum {
    psync2,
};

const Error = error{ TooManyReplicas, MemoryError, AlreadyExists, QueueFull, QueueEmpty };

const CapabilitySet = struct {
    psync2: bool = false,
};

pub const Request = union(enum) {
    propagate: core.ElemTree,
    getack: struct { elem: *const core.Elem, timeout_ms: u64 },
    shutdown: void,

    pub fn format(self: Request, comptime fmt: []const u8, options: std.fmt.FormatOptions, writer: anytype) !void {
        _ = fmt;
        _ = options;
        try writer.writeAll("Request{");
        switch (self) {
            .propagate => |et| try writer.print(".propagate={s}", .{util.qb(et.raw_bytes)}),
            .getack => |v| try writer.print(".getack={{timeout_ms={d}}}", .{v.timeout_ms}),
            .shutdown => try writer.writeAll(".shutdown"),
        }
        try writer.writeByte('}');
    }
};

const Response = union(enum) {
    err: anyerror,
    ack_offset: u64,
    sent: void,

    pub fn format(self: Response, comptime fmt: []const u8, options: std.fmt.FormatOptions, writer: anytype) !void {
        _ = fmt;
        _ = options;
        try writer.writeAll("Response{");
        switch (self) {
            .err => |e| try writer.print(".err={!}", .{e}),
            .ack_offset => |o| try writer.print(".ack_offset={d}", .{o}),
            .sent => try writer.writeAll(".sent"),
        }
        try writer.writeByte('}');
    }
};

pub const Replica = struct {
    id: Id,
    connection_h: ConnectionHandle,
    capabilities: CapabilitySet,
    port: ?u16,
    last_propagation: ?i64 = null,

    pub const Id = ConnectionHandle;

    pub fn set_capas(self: *Replica, capas: []Capability) void {
        for (capas) |capa| switch (capa) {
            .psync2 => self.capabilities.psync2 = true,
        };
    }

    pub fn set_port(self: *Replica, port: u16) !void {
        self.port = port;
    }
};

pub const MasterState = struct {
    replid: u160,
    repl_offset: u64,
    replicas: ReplicaPool,

    pub fn mkreplica(self: *MasterState, connection_h: ConnectionHandle) !*Replica {
        if (self.replicas.get_ptr(connection_h)) |ptr| return ptr;
        const new_id = try self.replicas.create(connection_h);
        return self.replicas.get_ptr(new_id) orelse return Error.MemoryError;
    }
};

pub const ReplicaState = struct {
    master_host: []const u8,
    master_port: u16,
    master_last_replid: ?u160 = null,
    master_last_offset: ?u64 = null,
};

pub const State = union(enum) {
    master: MasterState,
    replica: ReplicaState,

    pub fn init(config: ServerConfig, allocator: std.mem.Allocator) !State {
        if (config.replication_host) |h| {
            return .{ .replica = .{
                .master_host = h,
                .master_port = config.replication_port,
            } };
        } else {
            const r = std.crypto.random.int(u160);
            return .{ .master = .{
                .replid = r,
                .repl_offset = 0,
                .replicas = try ReplicaPool.init(
                    config.replica_count,
                    allocator,
                ),
            } };
        }
    }
    pub fn deinit(self: State) void {
        switch (self) {
            .master => |m| m.replicas.deinit(),
            .replica => {},
        }
    }
};

pub const ReplicaPool = struct {
    buff: []?Replica,
    allocator: std.mem.Allocator,
    bq: BatchQueue,

    pub const BatchQueue = util.BatchQueue(Request, Response, Replica.Id);

    pub const PtrIterator = struct {
        cur: usize = 0,
        pool: ReplicaPool,
        pub fn next(self: *PtrIterator) ?*Replica {
            while (self.cur < self.pool.buff.len) {
                if (self.pool.buff[self.cur]) |*item_ptr| {
                    self.cur += 1;
                    return item_ptr;
                }
                self.cur += 1;
            }
            return null;
        }
    };

    fn init(max_replicas: usize, allocator: std.mem.Allocator) !ReplicaPool {
        const buff = try allocator.alloc(?Replica, max_replicas);
        for (0..max_replicas) |i| buff[i] = null;
        return ReplicaPool{
            .buff = buff,
            .allocator = allocator,
            .bq = try BatchQueue.init(max_replicas, allocator),
        };
    }

    fn deinit(self: ReplicaPool) void {
        self.allocator.free(self.buff);
        self.bq.deinit();
    }

    fn _locate(self: ReplicaPool, id: Replica.Id) ?usize {
        for (0..self.buff.len) |idx| {
            if (self.buff[idx] == null) continue;
            const item_ptr = &self.buff[idx].?;
            if (item_ptr.id == id) return idx;
        }
        return null;
    }

    fn _locate_free(self: ReplicaPool) ?usize {
        for (0..self.buff.len) |idx| {
            if (self.buff[idx] == null) return idx;
        }
        return null;
    }

    pub fn ptr_iterator(self: ReplicaPool) PtrIterator {
        return PtrIterator{ .pool = self };
    }

    pub fn count(self: ReplicaPool) usize {
        var out: usize = 0;
        for (self.buff) |item| out += if (item == null) 0 else 1;
        return out;
    }

    fn create(self: *ReplicaPool, connection_h: ConnectionHandle) !Replica.Id {
        const id = @as(Replica.Id, connection_h);
        if (self._locate(id)) |_| return Error.AlreadyExists;
        const free_idx = self._locate_free() orelse return Error.TooManyReplicas;
        try self.bq.add_ctx(id);
        self.buff[free_idx] = Replica{
            .id = id,
            .connection_h = connection_h,
            .capabilities = .{},
            .port = null,
            .last_propagation = null,
        };
        return id;
    }

    pub fn get_ptr(self: ReplicaPool, id: Replica.Id) ?*Replica {
        if (self._locate(id)) |idx| return &self.buff[idx].?;
        return null;
    }

    pub fn remove(self: *ReplicaPool, id: Replica.Id) void {
        const idx = self._locate(id) orelse return;
        self.buff[idx] = null;
        self.bq.delete_ctx(id);
    }
};

/// Handshake with master and start handling incoming commands,
/// mostly propagations.
pub fn handle_replica2master(task: *Task, server: *Server) !HandlerResult {
    task.logger.info("* replica2master handler listening", .{});
    try task.handshake();
    while (true) {
        const et = try task.rcv_et();
        defer et.deinit();
        if (et.root == null) break; // connection closed by peeer
        const cmd = try task.parse_cmd_or_reject(et.root.?) orelse break; // we rejected
        const cmd_result = cmd.run(task, server);
        if (cmd.should(.ack_master)) {
            try task.send_elem(cmd_result.elem);
        }
        try task.bump_replica_offset(et.raw_bytes.len);
    }
    return .{};
}

/// Finalize handshake by sending RDB and begin handling requests
/// for this replica.
///
/// The replica to serve is identified by connection handler
/// *task.connection_h*; once the connection is removed from the connection
/// pool, the handler will no longer resolve, ie. we assume replica has been
/// removed and we quit.
pub fn handle_master2replica(task: *Task, server: *Server) !HandlerResult {
    _ = server;
    const ms = try task.require_master_state();
    var reqs_done: usize = 0;
    try task.send_bin_elem(EMPTY_RDB);
    task.logger.info("* master2replica handler ready", .{});
    while (ms.replicas.get_ptr(task.connection_h)) |replica| { // "while we still have replica to care about"
        task.logger.info("* waiting on replication request pipe", .{});
        const ctx = try ms.replicas.bq.wait_request(replica.id);
        task.logger.info("* received replication request: {any}", .{ctx.request});
        switch (ctx.request) {
            .shutdown => break,
            .propagate => |et| {
                task.send_blob(et.raw_bytes) catch |e| {
                    task.logger.err("! failed to send element: {any} due to {!}", .{ et, e });
                    try ctx.respond(.{ .err = e });
                    break;
                };
                try ctx.respond(.{ .sent = {} });
            },
            .getack => |req| {
                const offset = task.getack_one(req.elem) catch |e| {
                    task.logger.err("! failed to get ACK within: {any} ns due to {!}", .{ req.timeout_ms, e });
                    try ctx.respond(.{ .err = e });
                    break;
                };
                try ctx.respond(.{ .ack_offset = @intCast(offset) });
            },
        }
        task.logger.info("* processed replication request: {any}", .{ctx.request});
        reqs_done +|= 1;
    }
    task.logger.debug(": reqs_done={d}", .{reqs_done});
    task.logger.info("* shutting down master2replica propagation handler", .{});
    return .{};
}
