const std = @import("std");

const util = @import("../util.zig");

const Elem = @import("core.zig").Elem;
const ElemTree = @import("core.zig").ElemTree;
const ElemType = @import("core.zig").ElemType;

pub const Sigil = enum {
    a_bulk_string, // bulk string (Elem.str) or null bulk string (Elem.nbs)
    binary_string,
    // ^^ Elem.bin: encoded like bulk string but without \r\n at the end (sigh...).
    //
    // This one is impossible to detect from the sigil so caller must know to expect
    // binary string and use parse_binary_string() function.   (This happens eg. with
    // the PSYNC2 command.)
    simple_error,
    simple_string,
    integer,
    null,
    array,
    nothing,
};

fn str_contains_only_digits(haystack: []const u8) bool {
    const idx = std.mem.indexOfNone(u8, haystack, "0123456789");
    return (idx == null);
}

fn str_contains_any(haystack: []const u8, needles: []const u8) bool {
    const idx = std.mem.indexOfAny(u8, haystack, needles);
    return (idx != null);
}

const MAX_READ_BYTES = 1024 * 1024;

const Parser = struct {
    reader: std.io.AnyReader,
    arena: std.heap.ArenaAllocator,
    raw: std.ArrayList(u8),
    // ^^ we could be smarter and have all .err, .str and .bin refer to .raw,
    // saving lots of memory, but it would require rewrite of some of the parser
    // logic

    pub const Error = error{
        _NotImplemented,
        ReaderError,
        BadSigil,
        BadSize,
        EmptyCommand,
        MissingData,
        MissingBlob,
        MissingSigil,
        MissingArrayElement,
        MissingArrayValue,
        MissingSize,
        MissingCRLF,
        BadIntegerValue,
        OutOfMemory,
        InternalError,
    };

    fn init(reader: std.io.AnyReader, allocator: std.mem.Allocator) Parser {
        var p: Parser = undefined;
        p.arena = std.heap.ArenaAllocator.init(allocator);
        p.raw = std.ArrayList(u8).init(p.arena.allocator());
        p.reader = reader;
        return p;
    }

    fn _pop_size(self: *Parser) Error!?usize {
        // Size is positive int except for special-case of `-1`, which
        // is only valid after bulk string sigil (`$`) and signifies
        // that the element is "null bulk string" (apparently an older
        // solution for "null").
        // For simplicity we just report size as `null`.
        const word = try self._stream_until_crlf();
        defer self.arena.allocator().free(word);
        if (std.mem.eql(u8, word, "-1")) return null;
        if (!str_contains_only_digits(word)) return Error.BadSize;
        const num = std.fmt.parseInt(usize, word, 10) catch return Error.BadSize;
        return num;
    }

    fn _load_blob(self: *Parser, len: usize) Error![]u8 {
        const dest = try self.arena.allocator().alloc(u8, len);
        var rcvd_bytes_n: usize = 0;
        while (rcvd_bytes_n < dest.len) {
            const new_bytes_n = self.reader.read(dest[rcvd_bytes_n..]) catch |e| {
                std.log.err("Parser._load_blob():e={}", .{e});
                return Error.ReaderError;
            };
            if (new_bytes_n == 0) break;
            rcvd_bytes_n = std.math.add(usize, rcvd_bytes_n, new_bytes_n) catch return Error.InternalError;
        }
        if (rcvd_bytes_n < dest.len) return Error.MissingData;
        self.raw.appendSlice(dest) catch return Error.OutOfMemory;
        return dest;
    }

    fn _stream_until_crlf(self: *Parser) Error![]u8 {
        var dest = std.ArrayList(u8).init(self.arena.allocator());
        var got_crlf = false;
        while (true) {
            // .streamUntilDelimiter() will drop the LF
            self.reader.streamUntilDelimiter(dest.writer(), '\n', null) catch |e| switch (e) {
                error.EndOfStream => break,
                else => {
                    std.log.err("Parser._stream_until_crlf():e={}", .{e});
                    return Error.ReaderError;
                },
            };
            if (dest.items.len > 0 and dest.getLast() == '\r') {
                // CRLF confirmed; drop the CR as well
                _ = dest.pop();
                got_crlf = true;
                break;
            }
            // this was just LF: put it back in
            dest.append('\n') catch return Error.OutOfMemory;
        }
        if (!got_crlf) return Error.MissingCRLF;
        self.raw.appendSlice(dest.items) catch return Error.OutOfMemory;
        self.raw.appendSlice("\r\n") catch return Error.OutOfMemory;
        return dest.toOwnedSlice() catch return Error.OutOfMemory;
    }

    fn __require_crlf(self: *Parser) Error!void {
        const b1 = self.reader.readByte() catch return Error.MissingCRLF;
        const b2 = self.reader.readByte() catch return Error.MissingCRLF;
        self.raw.append(b1) catch return Error.OutOfMemory;
        self.raw.append(b2) catch return Error.OutOfMemory;
        if (b1 == '\r' and b2 == '\n') return;
        return Error.MissingCRLF;
    }

    fn _popelem_array(self: *Parser) Error!Elem {
        const len = try self._pop_size() orelse return Error.BadSize;
        var togo = len;
        var elem_al = std.ArrayList(Elem).init(self.arena.allocator());
        while (togo > 0) {
            const elem = try self._popelem() orelse return Error.MissingArrayElement;
            try elem_al.append(elem);
            togo = togo - 1;
        }
        return Elem{ .arr = elem_al.toOwnedSlice() catch return Error.OutOfMemory };
    }

    fn _popelem(self: *Parser) Error!?Elem {
        switch (try self.__pop_sigil()) {
            .nothing => return null,
            .array => return try self._popelem_array(),
            .integer => return try self._popelem_integer(),
            .null => return try self._popelem_null(),
            .simple_error => return try self._popelem_simple_error(),
            .a_bulk_string => return try self._popelem_a_bulk_string(),
            .simple_string => return try self._popelem_simple_string(),
            .binary_string => unreachable,
        }
    }

    fn _popelem_a_bulk_string(self: *Parser) Error!Elem {
        if (try self._pop_size()) |len| {
            // proper bulk_string with actual data
            const buff = try self._load_blob(len);
            try self.__require_crlf();
            return Elem{ .str = buff };
        } else {
            // size was `-1` -> NBS
            return Elem{ .nbs = null };
        }
    }

    fn _popelem_binary_string(self: *Parser) Error!?Elem {
        const sigil = try self.__pop_sigil();
        if (sigil == .nothing) return null;
        if (sigil != .a_bulk_string) return Error.BadSigil;
        const len = try self._pop_size() orelse return Error.BadSize;
        const buff = try self._load_blob(len);
        return Elem{ .bin = buff };
    }

    fn _popelem_integer(self: *Parser) Error!Elem {
        const bytes = try self._stream_until_crlf();
        const int = std.fmt.parseInt(i64, bytes, 10) catch return Error.BadIntegerValue;
        return Elem{ .int = int };
    }

    fn _popelem_null(self: *Parser) Error!Elem {
        try self.__require_crlf();
        return Elem{ .null = null };
    }

    fn _popelem_simple_error(self: *Parser) Error!Elem {
        const bytes = try self._stream_until_crlf();
        return Elem{ .err = bytes };
    }

    fn _popelem_simple_string(self: *Parser) Error!Elem {
        const bytes = try self._stream_until_crlf();
        return Elem{ .str = bytes };
    }

    fn __pop_sigil(self: *Parser) Error!Sigil {
        const sigil = self.reader.readByte() catch return .nothing;
        try self.raw.append(sigil);
        switch (sigil) {
            '*' => return .array,
            '+' => return .simple_string,
            '-' => return .simple_error,
            '$' => return .a_bulk_string,
            ':' => return .integer,
            '_' => return .null,
            else => return Error.BadSigil,
        }
    }

    fn _as_tree(self: *Parser, root: ?Elem) Error!ElemTree {
        return ElemTree{
            .root = r: {
                if (root == null) break :r null;
                const ptr = self.arena.allocator().create(Elem) catch return Error.OutOfMemory;
                ptr.* = root.?;
                break :r ptr;
            },
            .raw_bytes = self.raw.toOwnedSlice() catch return Error.OutOfMemory,
            .arena = self.arena,
        };
    }

    fn parse(self: *Parser) Error!ElemTree {
        //      const EP = self.arena.allocator().create(Elem) catch return Error.OutOfMemory;
        //      std.log.err("EP: {*}", .{EP});
        return try self._as_tree(try self._popelem());
    }

    fn parse_binary_string(self: *Parser) Error!ElemTree {
        return try self._as_tree(try self._popelem_binary_string());
    }
};

pub fn parse_et(reader: std.io.AnyReader, allocator: std.mem.Allocator) Parser.Error!ElemTree {
    var parser = Parser.init(reader, allocator);
    return try parser.parse();
}

/// measure how much bytes would *elem* take encoded
pub fn predict_write(elem: Elem, allocator: std.mem.Allocator) !usize {
    var al = std.ArrayList(u8).init(allocator);
    defer al.deinit();
    try write(elem, al.writer());
    return al.items.len;
}

pub const ElemStreamTimer = struct {
    allocator: std.mem.Allocator,
    poller: Poller,

    const StreamEnum = enum { file };
    const Poller = std.io.Poller(StreamEnum);

    pub const Error = error{ StreamClosed, StreamTimeout };
    const ParseFn = *const fn (std.io.AnyReader, std.mem.Allocator) anyerror!ElemTree;

    pub fn init(stream: std.net.Stream, allocator: std.mem.Allocator) ElemStreamTimer {
        return .{
            .poller = std.io.poll(
                allocator,
                StreamEnum,
                .{ .file = std.fs.File{ .handle = stream.handle } },
            ),
            .allocator = allocator,
        };
    }

    pub fn deinit(self: *ElemStreamTimer) void {
        self.poller.deinit();
    }

    fn _get_slice(self: *ElemStreamTimer, timeout_ns: u64) ![]const u8 {
        if (!(try self.poller.pollTimeout(timeout_ns))) return Error.StreamClosed;
        const fifo = self.poller.fifo(.file);
        if (fifo.readableLength() == 0) return Error.StreamTimeout;
        return fifo.readableSlice(0);
    }

    fn _get_et(self: *ElemStreamTimer, parse_fn: ParseFn, timeout_ns: u64, allocator: std.mem.Allocator) !ElemTree {
        var remaining_ns = timeout_ns;
        while (true) {
            if (remaining_ns == 0) return Error.StreamTimeout;
            const start_nst = std.time.nanoTimestamp();
            const data = try self._get_slice(remaining_ns);
            var stream = std.io.fixedBufferStream(data);
            const et = parse_fn(stream.reader().any(), allocator) catch continue;
            const dur_ns = std.time.nanoTimestamp() - start_nst;
            remaining_ns -|= @intCast(dur_ns);
            if (et.root == null) unreachable; // .get_slice should never return slice of 0;
            return et;
        }
    }

    pub fn get_bin_et(self: *ElemStreamTimer, timeout_ns: u64, allocator: std.mem.Allocator) !ElemTree {
        return self._get_et(
            parse_binary_string_et,
            timeout_ns,
            allocator,
        );
    }

    pub fn get_et(self: *ElemStreamTimer, timeout_ns: u64, allocator: std.mem.Allocator) !ElemTree {
        return self._get_et(
            parse_et,
            timeout_ns,
            allocator,
        );
    }
};

pub fn parse_binary_string_et(reader: std.io.AnyReader, allocator: std.mem.Allocator) Parser.Error!ElemTree {
    var parser = Parser.init(reader, allocator);
    return try parser.parse_binary_string();
}

pub fn require_blob(reader: std.io.AnyReader, allocator: std.mem.Allocator) Parser.Error![]const u8 {
    var parser = Parser.init(reader, allocator);
    const et = try parser.parse_binary_string();
    const elem = et.root orelse error.MissingBlob;
    return elem.bin;
}

pub fn write(data: Elem, writer: anytype) !void {
    switch (data) {
        .bool, .double, .null => return Parser.Error._NotImplemented,
        .int => {
            try std.fmt.format(writer, ":{d}\r\n", .{data.int});
        },
        .arr => {
            try std.fmt.format(writer, "*{d}\r\n", .{data.arr.len});
            for (data.arr) |item| {
                try write(item, writer);
            }
        },
        .nbs => {
            try writer.writeAll("$-1\r\n");
        },
        .err => {
            try std.fmt.format(writer, "-{s}\r\n", .{data.err});
        },
        .bin => {
            try std.fmt.format(writer, "${d}\r\n", .{data.bin.len});
            try writer.writeAll(data.bin);
        },
        .str => {
            if (str_contains_any(data.str, "\r\n+")) {
                try std.fmt.format(writer, "${d}\r\n", .{data.str.len});
                try writer.writeAll(data.str);
                try writer.writeAll("\r\n");
            } else {
                try std.fmt.format(writer, "+{s}\r\n", .{data.str});
            }
        },
    }
}

const ParseTestCase = struct {
    buff: []const u8,
    result: anyerror!ElemTree,
    gpa: std.heap.GeneralPurposeAllocator(.{}),

    /// Initialize test case with source buffer *buff*
    fn init(buff: []const u8) ParseTestCase {
        var fbs = std.io.fixedBufferStream(buff);
        var gpa = std.heap.GeneralPurposeAllocator(.{}){};
        return ParseTestCase{
            .buff = buff,
            .gpa = gpa,
            .result = parse_et(fbs.reader().any(), gpa.allocator()),
        };
    }

    /// Initialize test case with source buffer *buff* and parse_binary_string()
    fn init_bin(buff: []const u8) ParseTestCase {
        var fbs = std.io.fixedBufferStream(buff);
        var gpa = std.heap.GeneralPurposeAllocator(.{}){};
        return ParseTestCase{
            .buff = buff,
            .gpa = gpa,
            .result = parse_binary_string_et(fbs.reader().any(), gpa.allocator()),
        };
    }

    fn _get_root(self: ParseTestCase) !Elem {
        const tree = try self.result;
        try std.testing.expectStringStartsWith(self.buff, tree.raw_bytes);
        const root_ptr = tree.root orelse return error.TestFailNoRoot;
        return root_ptr.*;
    }

    fn _get_arr_item(self: ParseTestCase, idx: usize) !Elem {
        const elem = try self._get_root();
        std.testing.expect(elem == ElemType.arr) catch return error.TestFailWrongTag;
        std.testing.expect(elem.arr.len >= (idx + 1)) catch return error.TestFailWrongSize;
        return elem.arr[idx];
    }

    /// Expect Parser.Error *oracle* thrown when parsing source buffer
    fn expect_error(self: ParseTestCase, oracle: Parser.Error) !void {
        try std.testing.expectEqual(oracle, self.result);
    }

    /// Expect null, ie. no elem at all found in the string (i.e. empty string was parsed)
    fn expect_no_elem(self: ParseTestCase) !void {
        const tree = try self.result;
        try std.testing.expectEqualStrings(self.buff, tree.raw_bytes);
        try std.testing.expectEqual(null, tree.root);
    }

    /// Expect ElemType.nbs
    fn expect_nbs(self: ParseTestCase) !void {
        const elem = try self._get_root();
        std.testing.expect(elem == ElemType.nbs) catch return error.TestFailWrongTag;
    }

    /// Expect ElemType.null
    fn expect_null(self: ParseTestCase) !void {
        const elem = try self._get_root();
        std.testing.expect(elem == ElemType.null) catch return error.TestFailWrongTag;
    }

    /// Expect ElemType.bool with bool value *oracle*
    fn expect_bool(self: ParseTestCase, oracle: bool) !void {
        const elem = try self._get_root();
        std.testing.expect(elem == ElemType.bool) catch return error.TestFailWrongTag;
        try std.testing.expectEqual(oracle, elem.bool);
    }

    /// Expect ElemType.double with f64 value *oracle*
    fn expect_double(self: ParseTestCase, oracle: f64) !void {
        const elem = try self._get_root();
        std.testing.expect(elem == ElemType.double) catch return error.TestFailWrongTag;
        try std.testing.expectEqual(oracle, elem.double);
    }

    /// Expect ElemType.int with i64 value *oracle*
    fn expect_int(self: ParseTestCase, oracle: i64) !void {
        const elem = try self._get_root();
        std.testing.expect(elem == ElemType.int) catch return error.TestFailWrongTag;
        try std.testing.expectEqual(oracle, elem.int);
    }

    /// Expect ElemType.err with string value *oracle*
    fn expect_err(self: ParseTestCase, oracle: []const u8) !void {
        const elem = try self._get_root();
        std.testing.expect(elem == ElemType.err) catch return error.TestFailWrongTag;
        try std.testing.expectEqualStrings(oracle, elem.err);
    }

    /// Expect ElemType.str with string value *oracle*
    fn expect_str(self: ParseTestCase, oracle: []const u8) !void {
        const elem = try self._get_root();
        std.testing.expect(elem == ElemType.str) catch return error.TestFailWrongTag;
        try std.testing.expectEqualStrings(oracle, elem.str);
    }

    /// Expect ElemType.bin with string value *oracle*
    fn expect_bin(self: ParseTestCase, oracle: []const u8) !void {
        const elem = try self._get_root();
        std.testing.expect(elem == ElemType.bin) catch return error.TestFailWrongTag;
        try std.testing.expectEqualStrings(oracle, elem.bin);
    }

    /// Expect ElemType.arr of length of *oracle*
    fn expect_arr_len(self: ParseTestCase, oracle: usize) !void {
        const elem = try self._get_root();
        std.testing.expect(elem == ElemType.arr) catch return error.TestFailWrongTag;
        try std.testing.expectEqual(oracle, elem.arr.len);
    }

    /// Expect ElemType.arr with ElemType.str at *idx*, with string value *oracle*
    fn expect_str_at(self: ParseTestCase, idx: usize, oracle: []const u8) !void {
        const elem = try self._get_arr_item(idx);
        std.testing.expect(elem == ElemType.str) catch return error.TestFailWrongTag;
        try std.testing.expectEqualStrings(oracle, elem.str);
    }

    /// Expect ElemType.arr with ElemType.bin at *idx*, with string value *oracle*
    fn expect_bin_at(self: ParseTestCase, idx: usize, oracle: []const u8) !void {
        const elem = try self._get_arr_item(idx);
        std.testing.expect(elem == ElemType.bin) catch return error.TestFailWrongTag;
        try std.testing.expectEqualStrings(oracle, elem.bin);
    }

    /// Expect ElemType.arr with ElemType.int at *idx*, with i64 value *oracle*
    fn expect_int_at(self: ParseTestCase, idx: usize, oracle: i64) !void {
        const elem = try self._get_arr_item(idx);
        std.testing.expect(elem == ElemType.int) catch return error.TestFailWrongTag;
        try std.testing.expectEqual(oracle, elem.int);
    }

    /// Expect ElemType.arr with ElemType.int at *idx*, with i64 value *oracle*
    fn expect_null_at(self: ParseTestCase, idx: usize) !void {
        const elem = try self._get_arr_item(idx);
        std.testing.expect(elem == ElemType.null) catch return error.TestFailWrongTag;
    }
};

test "parse()-NEW: nothing" {
    const mkcase = ParseTestCase.init;
    try mkcase("").expect_no_elem();
}

test "parse()-NEW: null" {
    const mkcase = ParseTestCase.init;
    try mkcase("_").expect_error(Parser.Error.MissingCRLF);
    try mkcase("_\n").expect_error(Parser.Error.MissingCRLF);
    try mkcase("_\r").expect_error(Parser.Error.MissingCRLF);
    try mkcase("_\n\r").expect_error(Parser.Error.MissingCRLF);
    try mkcase("_\r\n").expect_null();
}

test "parse()-NEW: nbs" {
    const mkcase = ParseTestCase.init;
    try mkcase("$").expect_error(Parser.Error.MissingCRLF);
    try mkcase("$-").expect_error(Parser.Error.MissingCRLF);
    try mkcase("$-1").expect_error(Parser.Error.MissingCRLF);
    try mkcase("$-1\n").expect_error(Parser.Error.MissingCRLF);
    try mkcase("$-1\r").expect_error(Parser.Error.MissingCRLF);
    try mkcase("$-1\r\n").expect_nbs();
    try mkcase("$-1\r\nX").expect_nbs();
}

test "parse()-NEW: simple error" {
    const mkcase = ParseTestCase.init;
    try mkcase("-").expect_error(Parser.Error.MissingCRLF);
    try mkcase("-\n").expect_error(Parser.Error.MissingCRLF);
    try mkcase("-\r").expect_error(Parser.Error.MissingCRLF);
    try mkcase("-\n\r").expect_error(Parser.Error.MissingCRLF);
    try mkcase("-\r\n").expect_err("");
    try mkcase("-A\r\n").expect_err("A");
    try mkcase("-AB\r\n").expect_err("AB");
    try mkcase("-+\r\n").expect_err("+");
}

test "parse()-NEW: simple string" {
    const mkcase = ParseTestCase.init;
    try mkcase("+").expect_error(Parser.Error.MissingCRLF);
    try mkcase("+\n").expect_error(Parser.Error.MissingCRLF);
    try mkcase("+\r").expect_error(Parser.Error.MissingCRLF);
    try mkcase("+\n\r").expect_error(Parser.Error.MissingCRLF);
    try mkcase("+\r\n").expect_str("");
    try mkcase("+A\r\n").expect_str("A");
    try mkcase("+AB\r\n").expect_str("AB");
    try mkcase("++\r\n").expect_str("+");
}

test "parse()-NEW: integer" {
    const mkcase = ParseTestCase.init;
    try mkcase(":").expect_error(Parser.Error.MissingCRLF);
    try mkcase(":\n").expect_error(Parser.Error.MissingCRLF);
    try mkcase(":\r").expect_error(Parser.Error.MissingCRLF);
    try mkcase(":\n\r").expect_error(Parser.Error.MissingCRLF);
    try mkcase(":\r\n").expect_error(Parser.Error.BadIntegerValue);
    try mkcase(":A\r\n").expect_error(Parser.Error.BadIntegerValue);
    try mkcase(":+A\r\n").expect_error(Parser.Error.BadIntegerValue);
    try mkcase("::1\r\n").expect_error(Parser.Error.BadIntegerValue);
    try mkcase(":1A\r\n").expect_error(Parser.Error.BadIntegerValue);
    try mkcase(":A1\r\n").expect_error(Parser.Error.BadIntegerValue);
    try mkcase(":0x01\r\n").expect_error(Parser.Error.BadIntegerValue);
    try mkcase(":0\r\n").expect_int(0);
    try mkcase(":1\r\n").expect_int(1);
    try mkcase(":123\r\n").expect_int(123);
    try mkcase(":-1\r\n").expect_int(-1);
    try mkcase(":+1\r\n").expect_int(1);
}

test "parse()-NEW: bulk string" {
    const mkcase = ParseTestCase.init;
    try mkcase("$").expect_error(Parser.Error.MissingCRLF);
    try mkcase("$1").expect_error(Parser.Error.MissingCRLF);
    try mkcase("$X").expect_error(Parser.Error.MissingCRLF);
    try mkcase("$XY").expect_error(Parser.Error.MissingCRLF);
    try mkcase("$\r\n").expect_error(Parser.Error.BadSize);
    try mkcase("$X\r\n").expect_error(Parser.Error.BadSize);
    try mkcase("$XY\r\n").expect_error(Parser.Error.BadSize);
    try mkcase("$1X\r\n").expect_error(Parser.Error.BadSize);
    try mkcase("$1_000\r\n").expect_error(Parser.Error.BadSize);
    try mkcase("$-2\r\n").expect_error(Parser.Error.BadSize); // -1 would be valid as NBS
    try mkcase("$0x01\r\n").expect_error(Parser.Error.BadSize);
    try mkcase("$1\r\n\r\n").expect_error(Parser.Error.MissingCRLF);
    try mkcase("$1\r\nA").expect_error(Parser.Error.MissingCRLF);
    try mkcase("$2\r\n\r\n").expect_error(Parser.Error.MissingCRLF);
    try mkcase("$2\r\nA").expect_error(Parser.Error.MissingData);
    try mkcase("$2\r\nAB").expect_error(Parser.Error.MissingCRLF);
    try mkcase("$0\r\nA").expect_error(Parser.Error.MissingCRLF);
    try mkcase("$0\r\nA\r\n").expect_error(Parser.Error.MissingCRLF);
    try mkcase("$0\r\n\r\nX").expect_str("");
    try mkcase("$0\r\n\r\n").expect_str("");
    try mkcase("$1\r\nA\r\n").expect_str("A");
    try mkcase("$1\r\n\r\r\n").expect_str("\r");
    try mkcase("$1\r\n\n\r\n").expect_str("\n");
    try mkcase("$2\r\nAB\r\n").expect_str("AB");
    try mkcase("$2\r\n\r\n\r\n").expect_str("\r\n");
    try mkcase("$9\r\n$3\r\nFOO\r\n\r\n").expect_str("$3\r\nFOO\r\n"); // including own code safely
}

test "parse()-NEW: binary string" {
    const mkcase = ParseTestCase.init_bin;
    try mkcase("$").expect_error(Parser.Error.MissingCRLF);
    try mkcase("$1").expect_error(Parser.Error.MissingCRLF);
    try mkcase("$X").expect_error(Parser.Error.MissingCRLF);
    try mkcase("$XY").expect_error(Parser.Error.MissingCRLF);
    try mkcase("$\r\n").expect_error(Parser.Error.BadSize);
    try mkcase("$X\r\n").expect_error(Parser.Error.BadSize);
    try mkcase("$XY\r\n").expect_error(Parser.Error.BadSize);
    try mkcase("$1X\r\n").expect_error(Parser.Error.BadSize);
    try mkcase("$1_000\r\n").expect_error(Parser.Error.BadSize);
    try mkcase("$-2\r\n").expect_error(Parser.Error.BadSize); // -1 would be valid as NBS
    try mkcase("$0x01\r\n").expect_error(Parser.Error.BadSize);
    try mkcase("$0\r\n").expect_bin("");
    try mkcase("$0\r\nA").expect_bin(""); // 'A' ignored
    try mkcase("$1\r\nA").expect_bin("A");
    try mkcase("$1\r\nAB").expect_bin("A"); // 'B' ignored
    try mkcase("$1\r\n").expect_error(Parser.Error.MissingData);
    try mkcase("$2\r\nAB").expect_bin("AB");
    try mkcase("$2\r\nABC").expect_bin("AB"); // 'C' ignored
    try mkcase("$2\r\nA").expect_error(Parser.Error.MissingData);
}

test "parse()-NEW: array: header" {
    const mkcase = ParseTestCase.init;
    try mkcase("*").expect_error(Parser.Error.MissingCRLF);
    try mkcase("*1").expect_error(Parser.Error.MissingCRLF);
    try mkcase("*12").expect_error(Parser.Error.MissingCRLF);
    try mkcase("*X").expect_error(Parser.Error.MissingCRLF);
    try mkcase("*XY").expect_error(Parser.Error.MissingCRLF);
    try mkcase("*\r\n").expect_error(Parser.Error.BadSize);
    try mkcase("*XY\r\n").expect_error(Parser.Error.BadSize);
    try mkcase("*1X\r\n").expect_error(Parser.Error.BadSize);
    try mkcase("*1_000\r\n").expect_error(Parser.Error.BadSize);
    try mkcase("*-1\r\n").expect_error(Parser.Error.BadSize);
    try mkcase("*0x01\r\n").expect_error(Parser.Error.BadSize);
}

test "parse()-NEW: array: empty" {
    const mkcase = ParseTestCase.init;
    try mkcase("*0\r").expect_error(Parser.Error.MissingCRLF);
    try mkcase("*0\n").expect_error(Parser.Error.MissingCRLF);
    try mkcase("*0\n\r").expect_error(Parser.Error.MissingCRLF);
    try mkcase("*0\r\nX").expect_arr_len(0);
    try mkcase("*0\r\nXY").expect_arr_len(0);
    try mkcase("*0\r\n").expect_arr_len(0);
}

test "parse()-NEW: array: item header error" {
    const mkcase = ParseTestCase.init;
    try mkcase("*1\r\n").expect_error(Parser.Error.MissingArrayElement);
    try mkcase("*1\r\nX").expect_error(Parser.Error.BadSigil);
    try mkcase("*1\r\n*").expect_error(Parser.Error.MissingCRLF);
}

test "parse()-NEW: array: bulk string" {
    const mkcase = ParseTestCase.init;
    var case: ParseTestCase = undefined;

    case = mkcase("*1\r\n$0\r\n\r\n");
    try case.expect_arr_len(1);
    try case.expect_str_at(0, "");
    case = mkcase("*1\r\n$1\r\nA\r\n");
    try case.expect_arr_len(1);
    try case.expect_str_at(0, "A");
    case = mkcase("*1\r\n$2\r\nAB\r\n");
    try case.expect_arr_len(1);
    try case.expect_str_at(0, "AB");
    case = mkcase("*1\r\n$2\r\n\r\n\r\n");
    try case.expect_arr_len(1);
    try case.expect_str_at(0, "\r\n");
}

test "parse()-NEW: array: simple string header errors" {
    const mkcase = ParseTestCase.init;
    try mkcase("*1\r\n+").expect_error(Parser.Error.MissingCRLF);
    try mkcase("*1\r\n+\n").expect_error(Parser.Error.MissingCRLF);
    try mkcase("*1\r\n+\r").expect_error(Parser.Error.MissingCRLF);
    try mkcase("*1\r\n+\n\r").expect_error(Parser.Error.MissingCRLF);
}

test "parse()-NEW: array: single, simple string" {
    const mkcase = ParseTestCase.init;
    var case: ParseTestCase = undefined;
    case = mkcase("*1\r\n+\r\n");
    try case.expect_arr_len(1);
    try case.expect_str_at(0, "");
    case = mkcase("*1\r\n+A\r\n");
    try case.expect_arr_len(1);
    try case.expect_str_at(0, "A");
    case = mkcase("*1\r\n+AB\r\n");
    try case.expect_arr_len(1);
    try case.expect_str_at(0, "AB");
}

test "parse()-NEW: array: multiple: simple string, simple string" {
    const mkcase = ParseTestCase.init;
    var case: ParseTestCase = undefined;
    case = mkcase("*2\r\n+\r\n+\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "");
    try case.expect_str_at(1, "");
    case = mkcase("*2\r\n+A\r\n+\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "A");
    try case.expect_str_at(1, "");
    case = mkcase("*2\r\n+\r\n+B\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "");
    try case.expect_str_at(1, "B");
    case = mkcase("*2\r\n+Foo bar\r\n+baz!\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "Foo bar");
    try case.expect_str_at(1, "baz!");
}

test "parse()-NEW: array: multiple: simple string, bulk string" {
    const mkcase = ParseTestCase.init;
    var case: ParseTestCase = undefined;
    case = mkcase("*2\r\n+\r\n$0\r\n\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "");
    try case.expect_str_at(1, "");
    case = mkcase("*2\r\n+A\r\n$0\r\n\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "A");
    try case.expect_str_at(1, "");
    case = mkcase("*2\r\n+\r\n$1\r\nB\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "");
    try case.expect_str_at(1, "B");
    case = mkcase("*2\r\n+Foo bar\r\n$4\r\nbaz!\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "Foo bar");
    try case.expect_str_at(1, "baz!");
}

test "parse()-NEW: array: multiple: bulk string, simple string" {
    const mkcase = ParseTestCase.init;
    var case: ParseTestCase = undefined;
    case = mkcase("*2\r\n$0\r\n\r\n+\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "");
    try case.expect_str_at(1, "");
    case = mkcase("*2\r\n+A\r\n$0\r\n\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "A");
    try case.expect_str_at(1, "");
    case = mkcase("*2\r\n$0\r\n\r\n+B\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "");
    try case.expect_str_at(1, "B");
    case = mkcase("*2\r\n$7\r\nFoo bar\r\n+baz!\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "Foo bar");
    try case.expect_str_at(1, "baz!");
}

test "parse()-NEW: array: multiple: bulk string, bulk string" {
    const mkcase = ParseTestCase.init;
    var case: ParseTestCase = undefined;
    case = mkcase("*2\r\n$0\r\n\r\n$0\r\n\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "");
    try case.expect_str_at(1, "");
    case = mkcase("*2\r\n$1\r\nA\r\n$0\r\n\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "A");
    try case.expect_str_at(1, "");
    case = mkcase("*2\r\n$0\r\n\r\n$1\r\nB\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "");
    try case.expect_str_at(1, "B");
    case = mkcase("*2\r\n$7\r\nFoo bar\r\n$4\r\nbaz!\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "Foo bar");
    try case.expect_str_at(1, "baz!");
}

test "parse()-NEW: array: multiple: simple string, null" {
    const mkcase = ParseTestCase.init;
    var case: ParseTestCase = undefined;
    case = mkcase("*2\r\n+\r\n_\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "");
    try case.expect_null_at(1);
    case = mkcase("*2\r\n+A\r\n_\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "A");
    try case.expect_null_at(1);
    case = mkcase("*2\r\n+Foo bar\r\n_\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "Foo bar");
    try case.expect_null_at(1);
}

test "parse()-NEW: array: multiple: null, simple string" {
    const mkcase = ParseTestCase.init;
    var case: ParseTestCase = undefined;
    case = mkcase("*2\r\n_\r\n+\r\n");
    try case.expect_arr_len(2);
    try case.expect_null_at(0);
    try case.expect_str_at(1, "");
    case = mkcase("*2\r\n_\r\n+A\r\n");
    try case.expect_arr_len(2);
    try case.expect_null_at(0);
    try case.expect_str_at(1, "A");
    case = mkcase("*2\r\n_\r\n+baz!\r\n");
    try case.expect_arr_len(2);
    try case.expect_null_at(0);
    try case.expect_str_at(1, "baz!");
}

test "parse()-NEW: array: multiple: bulk string, null" {
    const mkcase = ParseTestCase.init;
    var case: ParseTestCase = undefined;
    case = mkcase("*2\r\n$0\r\n\r\n_\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "");
    try case.expect_null_at(1);
    case = mkcase("*2\r\n+A\r\n_\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "A");
    try case.expect_null_at(1);
    case = mkcase("*2\r\n+Foo bar\r\n_\r\n");
    try case.expect_arr_len(2);
    try case.expect_str_at(0, "Foo bar");
    try case.expect_null_at(1);
}

test "parse()-NEW: array: multiple: null, bulk string" {
    const mkcase = ParseTestCase.init;
    var case: ParseTestCase = undefined;
    case = mkcase("*2\r\n_\r\n$0\r\n\r\n");
    try case.expect_arr_len(2);
    try case.expect_null_at(0);
    try case.expect_str_at(1, "");
    case = mkcase("*2\r\n_\r\n$1\r\nA\r\n");
    try case.expect_arr_len(2);
    try case.expect_null_at(0);
    try case.expect_str_at(1, "A");
    case = mkcase("*2\r\n_\r\n$4\r\nbaz!\r\n");
    try case.expect_arr_len(2);
    try case.expect_null_at(0);
    try case.expect_str_at(1, "baz!");
}
