const std = @import("std");

const util = @import("../util.zig");
const resp = @import("resp.zig");

const replication = @import("replication.zig");

const HandlerFn = @import("worker.zig").HandlerFn;
const HandlerResult = @import("worker.zig").HandlerResult;
const Logger = @import("logger.zig").Logger;
const Role = @import("task.zig").Role;
const Store = @import("store.zig").Store;
const Task = @import("task.zig").Task;
const WorkerPool = @import("worker.zig").WorkerPool;

pub const Config = struct {
    dir: ?[]const u8,
    dbfilename: []const u8,
    replication_host: ?[]const u8,
    replication_port: u16,
    max_dump_bytes: usize,
    thread_count: usize,
    replica_count: usize,
    replica_buff_size: usize,
    replica_timeout_ms: u64,
    command_timeout_ms: u64,
    port: u16,
    debug: bool,
    verbose: bool,
    logdir: ?[]const u8,
};

pub const Connection = struct {
    stream: std.net.Stream,
    address: std.net.Address,
};

pub const Server = struct {
    config: Config,
    store: Store,
    logger: Logger,
    allocator: std.mem.Allocator,
    replication_state: replication.State,
    connection_pool: ConnectionPool,
    last_task_id: usize = 0,
    main_thread_id: std.Thread.Id,
    cmd_worker_pool: CmdWorkerPool,
    replica_worker_pool: ReplicaWorkerPool,

    pub const CmdWorkerPool = WorkerPool("cmd", 0xDD);
    pub const ReplicaWorkerPool = WorkerPool("rep", 0xDD);
    pub const ConnectionPool = util.OpaquePool(Connection, 6, 48);

    pub fn create(
        config: Config,
        threads_buff: []u8,
        replicas_buff: []u8,
        allocator: std.mem.Allocator,
    ) !Server {
        const logger = Logger.init_root(.{
            .verbose = config.verbose,
            .debug = config.debug,
            .logdir = config.logdir,
        });

        const replication_state = try replication.State.init(config, allocator);

        // validate config
        {
            const ideal_cpus = switch (replication_state) {
                .master => 1 + config.thread_count + config.replica_count,
                .replica => 1 + config.thread_count + 1,
            };
            const have_cpus = try std.Thread.getCpuCount();
            switch (replication_state) {
                .master => {
                    if (config.thread_count == 0) {
                        logger.err("cannot work with zero threads: config.thread_count=0", .{});
                        return error.InvalidConfig;
                    }
                    if (have_cpus < ideal_cpus)
                        logger.warn(
                            "not enough CPUs; commands or replication might be blocked: have {d} but need {d} + {d} + 1 (commands, replicas, main thread)",
                            .{ have_cpus, config.thread_count, config.replica_count },
                        );
                },
                .replica => {
                    if (have_cpus < ideal_cpus) {
                        logger.warn(
                            "not enough CPUs; commands or replication might be blocked: have {d} but need {d} + 1 + 1 (commands, replication, main thread)",
                            .{ have_cpus, config.thread_count },
                        );
                    }
                },
            }
        }

        return Server{
            .config = config,
            .store = Store.init(allocator),
            .logger = logger,
            .allocator = allocator,
            .replication_state = replication_state,
            .connection_pool = try ConnectionPool.init(allocator),
            .cmd_worker_pool = CmdWorkerPool.init(
                threads_buff,
                config.thread_count,
                allocator,
            ),
            .replica_worker_pool = ReplicaWorkerPool.init(
                replicas_buff,
                config.replica_count,
                allocator,
            ),
            .main_thread_id = std.Thread.getCurrentId(),
        };
    }

    pub fn destroy(self: *Server) void {
        self.cmd_worker_pool.destroy();
        self.replica_worker_pool.destroy();
    }

    pub fn maybe_load_dump(self: *Server, allocator: std.mem.Allocator) !void {
        if (self.config.dir == null) return;
        const dir = self.config.dir.?;
        const dbfilename = self.config.dbfilename;
        self.store.load_rdb_dump(
            dir,
            dbfilename,
            self.config.max_dump_bytes,
            self.logger,
            allocator,
        ) catch |e| switch (e) {
            error.FileNotFound => {
                self.logger.warn("? ignoring non-existent dump file: {s}/{s}", .{ dir, dbfilename });
                return;
            },
            else => {
                self.logger.err("! error while loading dump file: {any} with {s}/{s}", .{ e, dir, dbfilename });
                return e;
            },
        };
    }

    fn queue_task(self: *Server, role: Role, handler_fn: HandlerFn, connection_h: ConnectionPool.Handle) !void {
        const task_args = Task.Args{
            .role = role,
            .connection_h = connection_h,
            .timeout_ms = switch (role) {
                .m2r, .r2m => self.config.replica_timeout_ms,
                .cmd => self.config.command_timeout_ms,
            },
        };
        switch (role) {
            .m2r, .r2m => try self.replica_worker_pool.queue(handler_fn, task_args),
            .cmd => try self.cmd_worker_pool.queue(handler_fn, task_args),
        }
    }

    pub fn create_master_connection(self: *Server) !ConnectionPool.Handle {
        const list = try std.net.getAddressList(
            self.allocator,
            self.replication_state.replica.master_host,
            self.replication_state.replica.master_port,
        );
        defer list.deinit();
        if (list.addrs.len == 0) return error.UnknownHostName;
        var stream: std.net.Stream = undefined;
        for (list.addrs) |addr| {
            stream = std.net.tcpConnectToAddress(addr) catch |err| switch (err) {
                error.ConnectionRefused => {
                    continue;
                },
                else => return err,
            };
            return try self.connection_pool.put(Connection{
                .stream = stream,
                .address = addr,
            });
        }
        return std.posix.ConnectError.ConnectionRefused;
    }

    pub fn serve(
        self: *Server,
        listener: *std.net.Server,
    ) !void {
        try self.cmd_worker_pool.start(self);
        try self.replica_worker_pool.start(self);

        if (self.replication_state == .replica) {
            const handle = try self.create_master_connection();
            try self.queue_task(
                .r2m,
                replication.handle_replica2master,
                handle,
            );
        }

        var conn: std.net.Server.Connection = undefined;
        while (true) {
            conn = try listener.accept();
            const handle = try self.connection_pool.put(Connection{
                .address = conn.address,
                .stream = conn.stream,
            });
            try self.queue_task(
                .cmd,
                handle_cmd,
                handle,
            );
        }
    }
};

/// Handle incoming RESP commands.
pub fn handle_cmd(task: *Task, server: *Server) !HandlerResult {
    task.logger.info("* command handler started; awaiting commands from stream", .{});
    while (true) {
        const et = try task.rcv_et();
        if (et.root == null) break; // connection closed by peeer
        const cmd = try task.parse_cmd_or_reject(et.root.?) orelse break; // we rejected
        const cmd_result = cmd.run(task, server);
        try task.send_elem(cmd_result.elem);

        // propagate?
        if (cmd.should(.propagate) and cmd_result.ok) {
            try task.maybe_propagate(et);
        } else {
            // only deinit if no propagation is happening
            defer et.deinit();
        }

        // keep stream as new replica?
        if (cmd.should(.start_replica)) {
            if (!cmd_result.ok) {
                task.logger.warn("? replica propagation setup failed", .{});
                break;
            }
            task.logger.info("* re-using connection for replica propagation", .{});
            try server.queue_task(
                .m2r,
                replication.handle_master2replica,
                task.connection_h,
            );
            return .{ .connection_policy = .keep };
        }
    }
    return .{};
}
