const std = @import("std");

const core = @import("core.zig");
const rdb = @import("rdb.zig");

const Elem = core.Elem;
const Logger = @import("logger.zig").Logger;

pub const PxUpdate = union(enum) {
    unset,
    keep,
    new: i64,
};

pub const Store = struct {
    pxats_db: std.StringHashMap(i64),
    elems_db: std.StringHashMap(*Elem),
    m: std.Thread.Mutex,
    allocator: std.mem.Allocator,

    const Error = error{
        NoPreviousTtl,
    };

    pub fn init(allocator: std.mem.Allocator) Store {
        return Store{
            .m = std.Thread.Mutex{},
            .pxats_db = std.StringHashMap(i64).init(allocator),
            .elems_db = std.StringHashMap(*Elem).init(allocator),
            .allocator = allocator,
        };
    }

    fn _adopt_key(self: Store, key: []const u8) ![]const u8 {
        if (self.elems_db.getEntry(key)) |old_entry| {
            return old_entry.key_ptr.*;
        } else {
            return try self.allocator.dupe(u8, key);
        }
    }

    fn _disown_key(self: *Store, key: []const u8) void {
        const old_entry = self.elems_db.getEntry(key);
        if (old_entry == null) return;
        self.allocator.free(old_entry.?.key_ptr.*);
    }

    fn _remove_if_expired(self: *Store, key: []const u8) !void {
        const pxat = self.pxats_db.get(key) orelse return;
        const now = std.time.milliTimestamp();
        if (now < pxat) return;
        try self.delete_elem(key);
    }

    pub fn delete_elem(self: *Store, key: []const u8) !void {
        const elem = self.elems_db.get(key) orelse return;
        try core.deep_free_elem(elem, self.allocator, core.ELEM_MAX_DEPTH);
        _ = self.pxats_db.remove(key);
        _ = self.elems_db.remove(key);
        self._disown_key(key);
    }

    pub fn get_elem(self: *Store, key: []const u8) !?Elem {
        try self._remove_if_expired(key);
        const elem_ptr = self.elems_db.get(key) orelse return null;
        return elem_ptr.*;
    }

    pub fn put_elem(self: *Store, key: []const u8, value: Elem, px_update: PxUpdate) !void {
        const our_pxat = switch (px_update) {
            .unset => null,
            .new => px_update.new,
            .keep => self.pxats_db.get(key) orelse return Error.NoPreviousTtl,
        };
        const our_elem = try core.deep_copy_elem(&value, self.allocator, core.ELEM_MAX_DEPTH);
        const our_key = try self._adopt_key(key);
        try self.elems_db.put(our_key, our_elem);
        try self.pxats_db.put(our_key, our_pxat orelse return);
    }

    pub fn describe(self: Store, logger: Logger) void {
        var n: usize = 0;
        var key_iter = self.elems_db.keyIterator();
        while (key_iter.next()) |key| {
            n += 1;
            logger.debug("Store.describe():key={s}", .{key.*});
        }
        logger.debug("Store.describe():n={d}", .{n});
    }

    pub fn load_rdb_dump(
        self: *Store,
        dirpath: []const u8,
        dbfilename: []const u8,
        max_bytes: usize,
        logger: Logger,
        allocator: std.mem.Allocator,
    ) !void {
        const dir = try std.fs.openDirAbsolute(dirpath, .{});
        const fh = try dir.openFile(dbfilename, .{ .mode = .read_only });
        const content = try fh.readToEndAlloc(allocator, max_bytes);
        var parser = rdb.Parser.init(content, allocator);
        const rdb_file = try parser.parse();
        for (rdb_file.database) |sub| {
            var iter = sub.values.iterator();
            while (iter.next()) |rdb_entry_ptr| {
                const key = rdb_entry_ptr.key_ptr.*;
                const elem = rdb_entry_ptr.value_ptr.*;
                const px_update = if (sub.pxats.get(key)) |pxat|
                    PxUpdate{ .new = pxat }
                else
                    PxUpdate{ .unset = {} };
                logger.info("* loading elem: {s} | {any} ({any})", .{ key, elem, px_update });
                try self.put_elem(key, elem, px_update);
            }
        }
    }
};
