const std = @import("std");

const resp = @import("resp.zig");
const util = @import("../util.zig");

const Command = @import("command.zig").Command;
const Elem = @import("core.zig").Elem;
const ElemTree = @import("core.zig").ElemTree;
const Logger = @import("logger.zig").Logger;
const ConnectionHandle = @import("server.zig").Server.ConnectionPool.Handle;
const ReplicationState = @import("replication.zig").State;
const ReplicaState = @import("replication.zig").ReplicaState;
const ReplicaRequest = @import("replication.zig").Request;
const ReplicationBatch = @import("replication.zig").ReplicaPool.BatchQueue.Batch;
const MasterState = @import("replication.zig").MasterState;
const WorkerPool = @import("worker.zig").WorkerPool;
const EMPTY_RDB = @import("rdb.zig").EMPTY_RDB;

pub const Role = enum { cmd, r2m, m2r };

pub const Task = struct {
    connection_h: ConnectionHandle,
    timeout_ms: u64,
    stream: std.net.Stream,
    logger: Logger,
    replication_state: *ReplicationState,
    allocator: std.mem.Allocator,

    pub const Args = struct {
        role: Role,
        connection_h: ConnectionHandle,
        timeout_ms: u64,
    };

    pub fn deinit(self: *Task) void {
        self.logger.debug("* done", .{});
        self.logger.deinit();
    }

    //
    // replication state access helpers
    //

    pub fn get_master_state(self: Task) ?*MasterState {
        if (self.replication_state.* == .replica) return null;
        return &self.replication_state.master;
    }

    pub fn get_replica_state(self: Task) ?*ReplicaState {
        if (self.replication_state.* == .master) return null;
        return &self.replication_state.replica;
    }

    pub fn require_master_state(self: Task) !*MasterState {
        if (self.replication_state.* == .replica) return error.NotAMaster;
        return &self.replication_state.master;
    }

    pub fn require_replica_state(self: Task) !*ReplicaState {
        if (self.replication_state.* == .master) return error.NotAReplica;
        return &self.replication_state.replica;
    }

    //
    // basic send/receive
    //

    pub fn require_bin_et(self: Task) !ElemTree {
        var timer = resp.ElemStreamTimer.init(self.stream, self.allocator);
        defer timer.deinit();
        const et = timer.get_bin_et(self.timeout_ms * std.time.ns_per_ms, self.allocator) catch |e| {
            self.logger.warn("? did not receive full RESP element: {!} (timeout_ms={d})", .{ e, self.timeout_ms });
            return e;
        };
        errdefer et.deinit();
        if (et.root == null) {
            self.logger.info("* connection closed by peer", .{});
            return error.ConnectionClosed;
        }
        return et;
    }

    pub fn require_et(self: Task) !ElemTree {
        return self.require_et_within(self.timeout_ms);
    }

    pub fn require_et_within(self: Task, timeout_ms: u64) !ElemTree {
        var timer = resp.ElemStreamTimer.init(self.stream, self.allocator);
        defer timer.deinit();
        const et = timer.get_et(timeout_ms * std.time.ns_per_ms, self.allocator) catch |e| {
            self.logger.warn("? did not receive full RESP element: {!} within {d} ms", .{ e, timeout_ms });
            return e;
        };
        errdefer et.deinit();
        if (et.root == null) {
            self.logger.info("* connection closed by peer", .{});
            return error.ConnectionClosed;
        }
        return et;
    }

    pub fn rcv_et(self: Task) !ElemTree {
        const et = resp.parse_et(self.stream.reader().any(), self.allocator) catch |e| {
            self.logger.err("! error parsing RESP element: {!}", .{e});
            return e;
        };
        if (et.root == null) {
            std.debug.assert(et.raw_bytes.len == 0);
            self.logger.info("* connection closed by peer", .{});
        } else {
            self.logger.info("< {any}", .{et});
        }
        return et;
    }

    pub fn send_bin_elem(self: Task, bin: []const u8) !void {
        self.logger.info(">> {s}", .{util.qb(bin)});
        const w = self.stream.writer().any();
        w.print("${d}\r\n{s}", .{ bin.len, bin }) catch |e| {
            self.logger.err("! error sending RESP .bin: {!}", .{e});
            return e;
        };
    }

    pub fn send_blob(self: Task, blob: []const u8) !void {
        self.logger.info(">>> {s}", .{util.qb(blob)});
        const w = self.stream.writer().any();
        w.writeAll(blob) catch |e| {
            self.logger.err("! error sending raw blob: {!}", .{e});
            return e;
        };
    }

    pub fn send_elem(self: Task, cmd_elem: Elem) !void {
        self.logger.info("> {any}", .{cmd_elem});
        resp.write(cmd_elem, self.stream.writer().any()) catch |e| {
            self.logger.err("! error sending RESP element: {!}", .{e});
            return e;
        };
    }

    //
    // simple operations
    //

    pub fn parse_cmd_or_reject(self: Task, cmd_elem: *Elem) !?Command {
        return Command.parse(cmd_elem.*, self.logger, self.allocator) catch |e| {
            try self.send_elem(Elem{ .err = @errorName(e) });
            self.logger.err("! error parsing REDIS command: {!}", .{e});
            return null;
        };
    }

    pub fn sleep(self: Task, ms: usize) void {
        self.logger.info("* sleeping for {d}ms", .{ms});
        std.debug.assert(ms < 1_000_000);
        const ns = std.math.mul(usize, ms, 1_000_000) catch unreachable;
        std.time.sleep(ns);
    }

    //
    // replication operations
    //

    pub fn handshake(
        self: *Task,
    ) !void {
        const rs = try self.require_replica_state();
        var response_et: ElemTree = undefined;
        defer response_et.deinit();

        // PING
        try self.send_elem(
            Elem{ .arr = &.{
                Elem{ .str = "PING" },
            } },
        );
        response_et = try self.rcv_et();
        try response_et.assert_cond(.{ .str_eq_i = "PONG" });

        // REPLCONF listening-port <OUR_PORT>
        const port_str = try std.fmt.allocPrint(self.allocator, "{d}", .{rs.master_port});
        defer self.allocator.free(port_str);
        try self.send_elem(
            Elem{ .arr = &.{
                Elem{ .str = "REPLCONF" },
                Elem{ .str = "listening-port" },
                Elem{ .str = port_str },
            } },
        );
        response_et = try self.rcv_et();
        try response_et.assert_cond(.{ .str_eq_i = "OK" });

        // REPLCONF capa
        try self.send_elem(
            Elem{ .arr = &.{
                Elem{ .str = "REPLCONF" },
                Elem{ .str = "capa" },
                Elem{ .str = "psync2" },
            } },
        );
        response_et = try self.rcv_et();
        try response_et.assert_cond(.{ .str_eq_i = "OK" });

        // PSYNC <REPLID> <OFFSET>
        //  .. but since it's handshake, REPLID="?" and OFFSET=-1
        try self.send_elem(
            Elem{ .arr = &.{
                Elem{ .str = "PSYNC" },
                Elem{ .str = "?" },
                Elem{ .str = "-1" },
            } },
        );
        response_et = try self.rcv_et();
        try response_et.assert_cond(.{ .str_startswith_i = "FULLRESYNC" });

        // blob with master's RDB
        response_et = try self.require_bin_et();
        try response_et.assert_cond(.{ .isa_bin = true, .len_ge = EMPTY_RDB.len });
    }

    fn _bump_master_offset(self: Task, bytes_n: u64) !void {
        const ms = try self.require_master_state();
        const old_offset = ms.repl_offset;
        ms.repl_offset = try std.math.add(u64, old_offset, bytes_n);
        self.logger.debug(
            ": bumped master offset: {d} + {d} => {d}",
            .{ old_offset, bytes_n, ms.repl_offset },
        );
    }

    pub fn _broadcast(self: Task, ms: *MasterState, req: ReplicaRequest, timeout_ms: u64) !*ReplicationBatch {
        self.logger.info("* broadcasting request: {any} to {d} replicas", .{ req, ms.replicas.count() });
        const batch_ptr = try ms.replicas.bq.queue_batch(req, try std.math.mul(u64, timeout_ms, std.time.ns_per_ms));
        ms.replicas.bq._dump("_broadcast()");
        return batch_ptr;
    }

    pub fn maybe_propagate(self: Task, et: ElemTree) !void {
        defer et.deinit();
        const ms = self.get_master_state() orelse return;
        try self._bump_master_offset(et.raw_bytes.len);
        const batch_ptr = try self._broadcast(ms, .{ .propagate = et }, 0);
        batch_ptr.disown();
        // while (batch.next()) |msg| {
        //     self.logger.debug(": thread responded: {any}", .{msg});
        //     switch (msg.response) {
        //         .sent => {
        //             self.logger.debug(
        //                 ": replica acked, propagation sent: replica_id={d} duration_ms={d}",
        //                 .{ msg.ctx_id, msg.duration_ms },
        //             );
        //         },
        //         .err => {
        //             self.logger.err(
        //                 "! propagation was not sent, dropping replica: replica_id={d} duration_ms={d} err={!}",
        //                 .{ msg.ctx_id, msg.duration_ms, msg.response.err },
        //             );
        //             ms.replicas.remove(msg.ctx_id);
        //         },
        //         else => {
        //             unreachable; // unexpected response to .propagation request
        //         },
        //     }
        //     ms.replicas.bq._dump(".maybe_propagate(), received propagation");
        // }
        // for (batch.jobs) |j| self.logger.debug(": j={any}", .{j});
    }

    pub fn bump_replica_offset(self: Task, bytes_n: u64) !void {
        const rs = try self.require_replica_state();
        const old_offset = rs.master_last_offset orelse 0;
        rs.master_last_offset = try std.math.add(u64, old_offset, bytes_n);
        self.logger.debug(
            ": bumped replica offset: {d} + {d} => {d}",
            .{ old_offset, bytes_n, rs.master_last_offset.? },
        );
    }

    pub fn check_replicas(self: Task, numreplicas: usize, timeout_ms: u64) !usize {
        if (numreplicas == 0) return 0;
        const ms = try self.require_master_state();
        const getack_elem = Elem{ .arr = &.{
            Elem{ .str = "REPLCONF" },
            Elem{ .str = "GETACK" },
            Elem{ .str = "*" },
        } };
        const getack_len = try resp.predict_write(getack_elem, self.allocator);
        var acked_count: usize = 0;
        var acked_togo = numreplicas;
        var batch = try self._broadcast(
            ms,
            .{ .getack = .{ .elem = &getack_elem, .timeout_ms = timeout_ms } },
            timeout_ms,
        );
        defer batch.deinit();
        while (batch.next()) |msg| {
            self.logger.debug(": thread responded: {any}", .{msg});
            switch (msg.response) {
                .err => {
                    self.logger.err(
                        "! ACK failed, dropping replica: replica_id={d} duration_ms={d} err={!}",
                        .{ msg.ctx_id, msg.duration_ms, msg.response.err },
                    );
                    ms.replicas.remove(msg.ctx_id);
                },
                .ack_offset => |offset| {
                    if (offset == ms.repl_offset) {
                        acked_count += 1;
                        acked_togo -|= 1;
                    } else {
                        self.logger.debug(
                            ": ACK not accepted: replica_id={d} duration_ms={d} .ack_offset={d} ms.repl_offset={d}",
                            .{ msg.ctx_id, msg.duration_ms, offset, ms.repl_offset },
                        );
                    }
                },
                else => {
                    unreachable; // unexpected response to .propagation request
                },
            }
            if (acked_togo == 0) break; // enough to satisfy *numreplicas*
        }
        for (batch.jobs) |j| self.logger.debug(": j={any}", .{j});
        for (batch.jobs) |j| {
            if (j == .timeouted) self.logger.debug("* replica late for timeout: replica_id={d}", .{j.timeouted});
        }
        try self._bump_master_offset(getack_len);
        return acked_count;
    }

    pub fn getack_one(self: Task, elem: *const Elem) !i64 {
        if (self.replication_state.* == .replica) unreachable; // replica must not call this
        self.logger.debug(": sending GETACK elem..", .{});
        try self.send_elem(elem.*);
        self.logger.debug(": ..sent, receiving response..", .{});
        const et = try self.require_et();
        self.logger.debug(": ..received: {any}", .{et});
        defer et.deinit();
        try et.assert_cond(.{ .len_eq = 3 });
        try et.assert_cond(.{ .idx = 0, .str_eq_i = "REPLCONF" });
        try et.assert_cond(.{ .idx = 1, .str_eq_i = "ACK" });
        return try et.parse_int_at(2);
    }
};
