const std = @import("std");
const builtin = @import("builtin");

const Connection = @import("server.zig").Connection;
const Logger = @import("logger.zig").Logger;
const Server = @import("server.zig").Server;
const Task = @import("task.zig").Task;
const Role = @import("task.zig").Role;

pub const HandlerResult = struct {
    connection_policy: enum { close, keep } = .close,
};

pub const HandlerFn = *const fn (*Task, *Server) anyerror!HandlerResult;

pub fn WorkerPool(comptime pfx: []const u8, comptime buff_mask: ?u8) type {
    return struct {
        threads_buff: []u8,
        arena: std.heap.ArenaAllocator,
        m: std.Thread.Mutex,
        c: std.Thread.Condition,
        job_queue: JobList,
        thread_count: usize,
        state: enum { new, running, stopped, disabled },
        next_task_id: usize = 0,
        server: ?*Server = null,

        const TaskIdPfx = pfx;
        const BuffMask = buff_mask;

        const JobList = std.SinglyLinkedList(Job);
        const WorkerPoolT = @This();

        pub const Workspace = struct {
            buff: []u8,
            fba: std.heap.FixedBufferAllocator,
        };

        pub const Job = struct {
            handler_fn: HandlerFn,
            task_args: Task.Args,
            task_id: usize,
        };

        pub fn init(
            threads_buff: []u8,
            thread_count: usize,
            allocator: std.mem.Allocator,
        ) WorkerPoolT {
            return WorkerPoolT{
                .thread_count = thread_count,
                .threads_buff = threads_buff,
                .arena = std.heap.ArenaAllocator.init(allocator),
                .m = std.Thread.Mutex{},
                .c = std.Thread.Condition{},
                .job_queue = JobList{},
                .state = .new,
            };
        }

        pub fn start(self: *WorkerPoolT, server: *Server) !void {
            self.server = server;
            if (self.thread_count == 0) {
                self.state = .disabled;
                return;
            }
            const thread_count = self.thread_count;
            if (@mod(self.threads_buff.len, thread_count) > 0) return error.BuffTooLarge;
            const buff_len = @divExact(self.threads_buff.len, thread_count);
            self.state = .running;
            for (0..thread_count) |idx| {
                const ws_ptr = try self.arena.allocator().create(Workspace);
                const bgn = idx * buff_len;
                const ws_buff = self.threads_buff[bgn..(bgn + buff_len)];
                ws_ptr.* = Workspace{
                    .buff = ws_buff,
                    .fba = std.heap.FixedBufferAllocator.init(ws_buff),
                };
                _ = try std.Thread.spawn(.{}, WorkerPoolT._worker_fn, .{ ws_ptr, self });
            }
        }

        pub fn destroy(self: *WorkerPoolT) void {
            self.state = .stopped;
            self.arena.deinit();
        }

        fn _worker_fn(
            ws: *Workspace,
            pool: *WorkerPoolT,
        ) void {
            const thread_id = std.Thread.getCurrentId();
            const server = pool.server.?; // server pointer must be set by now
            while (pool.state == .running) {
                var maybe_job: ?Job = null;
                {
                    pool.m.lock();
                    defer pool.m.unlock();
                    while (pool.job_queue.first == null) {
                        pool.c.wait(&pool.m);
                    }
                    const jnode = pool.job_queue.popFirst().?;
                    maybe_job = jnode.*.data;
                    pool.arena.allocator().destroy(jnode);
                }
                var job = maybe_job.?;

                // reset and wipe ws memory
                ws.fba.reset();
                if (WorkerPoolT.BuffMask) |m| {
                    for (ws.buff) |*b| b.* = m;
                }

                // create new task
                const connection: Connection = server.connection_pool.get(job.task_args.connection_h) orelse {
                    unreachable; // spawning task for a nonexistent connection!?
                };
                const logger_ctx = Logger.TaskCtx{
                    .task_id = job.task_id,
                    .task_id_pfx = WorkerPoolT.TaskIdPfx,
                    .thread_id = if (thread_id == server.main_thread_id) null else thread_id,
                    .handle = job.task_args.connection_h,
                    .address = connection.address,
                    .stream = connection.stream,
                    .role = job.task_args.role,
                };
                var task = Task{
                    .connection_h = job.task_args.connection_h,
                    .timeout_ms = job.task_args.timeout_ms,
                    .replication_state = &server.replication_state,
                    .stream = connection.stream,
                    .logger = Logger.init_for_task(server.logger.options, logger_ctx, ws.fba.allocator()),
                    .allocator = ws.fba.allocator(),
                };
                defer task.deinit();

                // run the handler
                const result = job.handler_fn(
                    &task,
                    server,
                ) catch |e| {
                    _ = server.connection_pool.delete(task.connection_h);
                    task.stream.close();
                    task.logger.err("handler returned error: {!}", .{e});
                    server.logger.err(
                        "connection handler returned error:a connection_h={any} address={any} e={!}",
                        .{ job.task_args.connection_h, connection.address, e },
                    );
                    return;
                };

                switch (result.connection_policy) {
                    .close => {
                        task.logger.info("* task done; closing connection", .{});
                        task.stream.close();
                        _ = server.connection_pool.delete(task.connection_h);
                    },
                    .keep => {
                        task.logger.info("* re-using connection for next task", .{});
                    },
                }
            }
        }

        pub fn queue(
            self: *WorkerPoolT,
            handler_fn: HandlerFn,
            task_args: Task.Args,
        ) !void {
            if (self.state == .disabled) return error.WorkerPoolDisabled;
            if (self.state == .stopped) return error.WorkerPoolStopped;
            if (self.state == .new) return error.WorkerPoolNotStarted;
            {
                self.m.lock();
                defer self.m.unlock();
                const jnode_ptr = try self.arena.allocator().create(JobList.Node);
                jnode_ptr.* = JobList.Node{ .data = Job{
                    .handler_fn = handler_fn,
                    .task_args = task_args,
                    .task_id = self.next_task_id,
                } };
                self.next_task_id += 1;
                self.job_queue.prepend(jnode_ptr);
            }
            self.c.signal();
        }
    };
}
