const std = @import("std");

pub fn clone(comptime T: type, value: anytype, allocator: std.mem.Allocator) !*T {
    const ptr = try allocator.create(T);
    ptr.* = value;
    return ptr;
}

pub fn eql_address(a: std.net.Address, b: std.net.Address) bool {
    return naive_client_id(a) == naive_client_id(b);
}

pub fn naive_client_id(address: std.net.Address) u48 {
    if (address.any.family != std.posix.AF.INET) unreachable;
    const addr_bytes = @as(*const [4]u8, @ptrCast(&address.in.sa.addr));
    const b1 = addr_bytes[0];
    const b2 = addr_bytes[0];
    const b3 = addr_bytes[0];
    const b4 = addr_bytes[0];
    const port = address.getPort();
    var id: u48 = b1;
    id <<= 8;
    id += b2;
    id <<= 8;
    id += b3;
    id <<= 8;
    id += b4;
    id <<= 8;
    id += port;
    return id;
}

const QMAX: usize = 64;

pub fn qb(blob: []const u8) []const u8 {
    return _qb(blob, QMAX);
}

fn _qb(blob: []const u8, max: ?usize) []const u8 {
    var al = std.ArrayList(u8).init(std.heap.page_allocator);
    _qblob(blob, max orelse QMAX, al.writer().any()) catch unreachable;
    return al.toOwnedSlice() catch unreachable;
}

pub fn qbs(blobs: []const []const u8) []const u8 {
    return _qbs(blobs, QMAX);
}

fn _qbs(blobs: []const []const u8, max: ?usize) []const u8 {
    var al = std.ArrayList(u8).init(std.heap.page_allocator);
    al.writer().writeAll("&.{") catch unreachable;
    for (blobs, 0..) |blob, idx| {
        if (idx > 0) al.writer().writeByte(',') catch unreachable;
        _qblob(blob, max orelse QMAX, al.writer().any()) catch unreachable;
    }
    al.writer().writeByte('}') catch unreachable;
    return al.toOwnedSlice() catch unreachable;
}

pub fn qs(str: []const u8) []const u8 {
    return _qs(str, QMAX);
}

fn _qs(str: []const u8, max: ?usize) []const u8 {
    var al = std.ArrayList(u8).init(std.heap.page_allocator);
    _qstr(str, max orelse QMAX, al.writer().any()) catch unreachable;
    return al.toOwnedSlice() catch unreachable;
}

pub fn qss(strs: []const []const u8) []const u8 {
    return _qss(strs, QMAX);
}

fn _qss(strs: []const []const u8, max: ?usize) []const u8 {
    var al = std.ArrayList(u8).init(std.heap.page_allocator);
    al.writer().writeAll("&.{") catch unreachable;
    for (strs, 0..) |str, idx| {
        if (idx > 0) al.writer().writeByte(',') catch unreachable;
        _qstr(str, max orelse QMAX, al.writer().any()) catch unreachable;
    }
    al.writer().writeByte('}') catch unreachable;
    return al.toOwnedSlice() catch unreachable;
}

pub fn qblob(blob: []const u8, writer: std.io.AnyWriter) !void {
    try _qblob(blob, QMAX, writer);
}

fn _qblob(blob: []const u8, max: usize, writer: std.io.AnyWriter) !void {
    const last_idx = @min(blob.len, max);
    try writer.print("[{d}]|", .{blob.len});
    try escape_nonascii(blob[0..last_idx], writer);
    if (blob.len > max) _ = try writer.write("..");
    try writer.writeByte('|');
}

pub fn qstr(str: []const u8, writer: std.io.AnyWriter) !void {
    return _qstr(str, QMAX, writer);
}

fn _qstr(str: []const u8, max: usize, writer: std.io.AnyWriter) !void {
    const last_idx = @min(str.len, max);
    try writer.writeByte('"');
    try escape(str[0..last_idx], writer);
    if (str.len > max) try writer.print("..({d} more)", .{str.len - max});
    try writer.writeByte('"');
}

pub fn escape(str: []const u8, writer: std.io.AnyWriter) !void {
    for (str) |c| _ = switch (c) {
        '\n' => try writer.write("\\n"),
        '\r' => try writer.write("\\r"),
        '\t' => try writer.write("\\t"),
        '\\' => try writer.write("\\\\"),
        '\'' => try writer.write("\\'"),
        '"' => try writer.write("\\\""),
        else => try writer.writeByte(c),
    };
}

pub fn escape_nonascii(str: []const u8, writer: std.io.AnyWriter) !void {
    for (str) |c| _ = switch (c) {
        '\n' => try writer.write("\\n"),
        '\r' => try writer.write("\\r"),
        '\t' => try writer.write("\\t"),
        '\\' => try writer.write("\\\\"),
        '\'' => try writer.write("\\'"),
        '"' => try writer.write("\\\""),
        else => if (std.ascii.isASCII(c)) try writer.writeByte(c) else try writer.print("\\x{x}", .{c}),
    };
}

test "escape" {
    const F = struct {
        fn assert_str(text: []const u8, oracle: []const u8) !void {
            var al = std.ArrayList(u8).init(std.testing.allocator);
            defer al.deinit();
            try escape(text, al.writer().any());
            try std.testing.expectEqualStrings(oracle, al.items);
        }
    };
    try F.assert_str("", "");
    try F.assert_str("A", "A");
    try F.assert_str("AB", "AB");
    try F.assert_str("\n", "\\n");
    try F.assert_str("\r", "\\r");
    try F.assert_str("\t", "\\t");
    try F.assert_str("\\", "\\\\");
    try F.assert_str("'", "\\'");
    try F.assert_str("\"", "\\\"");
    try F.assert_str("\r\n", "\\r\\n");
    try F.assert_str("foo\nbar", "foo\\nbar");
    try F.assert_str("foo\nbar\n", "foo\\nbar\\n");
    try F.assert_str("foo\r\nbar", "foo\\r\\nbar");
    try F.assert_str("foo\r\nbar\r\n", "foo\\r\\nbar\\r\\n");
    try F.assert_str("\r\nfoo\r\nbar\r\n", "\\r\\nfoo\\r\\nbar\\r\\n");
    try F.assert_str("foo\tbar\n", "foo\\tbar\\n");
}

test "qblob" {
    const F = struct {
        fn assert_str(text: []const u8, oracle: []const u8) !void {
            var al = std.ArrayList(u8).init(std.testing.allocator);
            defer al.deinit();
            try _qblob(text, 1024, al.writer().any());
            try std.testing.expectEqualStrings(oracle, al.items);
        }
        fn assert_limited_str(text: []const u8, max: usize, oracle: []const u8) !void {
            var al = std.ArrayList(u8).init(std.testing.allocator);
            defer al.deinit();
            try _qblob(text, max, al.writer().any());
            try std.testing.expectEqualStrings(oracle, al.items);
        }
    };
    try F.assert_str("", "[0]||");
    try F.assert_str("A", "[1]|A|");
    try F.assert_str("AB", "[2]|AB|");
    try F.assert_str("\n", "[1]|\\n|");
    try F.assert_str("\r", "[1]|\\r|");
    try F.assert_str("\t", "[1]|\\t|");
    try F.assert_str("\\", "[1]|\\\\|");
    try F.assert_str("'", "[1]|\\'|");
    try F.assert_str("\"", "[1]|\\\"|");
    try F.assert_str("\r\n", "[2]|\\r\\n|");
    try F.assert_str("foo\nbar", "[7]|foo\\nbar|");
    try F.assert_str("foo\nbar\n", "[8]|foo\\nbar\\n|");
    try F.assert_str("foo\r\nbar", "[8]|foo\\r\\nbar|");
    try F.assert_str("foo\r\nbar\r\n", "[10]|foo\\r\\nbar\\r\\n|");
    try F.assert_str("\r\nfoo\r\nbar\r\n", "[12]|\\r\\nfoo\\r\\nbar\\r\\n|");
    try F.assert_str("foo\tbar\n", "[8]|foo\\tbar\\n|");
    try F.assert_limited_str("", 0, "[0]||");
    try F.assert_limited_str("", 1, "[0]||");
    try F.assert_limited_str("", 2, "[0]||");
    try F.assert_limited_str("A", 0, "[1]|..|");
    try F.assert_limited_str("A", 1, "[1]|A|");
    try F.assert_limited_str("A", 2, "[1]|A|");
    try F.assert_limited_str("AB", 0, "[2]|..|");
    try F.assert_limited_str("AB", 1, "[2]|A..|");
    try F.assert_limited_str("AB", 2, "[2]|AB|");
}

test "qstr" {
    const F = struct {
        fn assert_str(text: []const u8, oracle: []const u8) !void {
            var al = std.ArrayList(u8).init(std.testing.allocator);
            defer al.deinit();
            try _qstr(text, 1024, al.writer().any());
            try std.testing.expectEqualStrings(oracle, al.items);
        }
        fn assert_limited_str(text: []const u8, max: usize, oracle: []const u8) !void {
            var al = std.ArrayList(u8).init(std.testing.allocator);
            defer al.deinit();
            try _qstr(text, max, al.writer().any());
            try std.testing.expectEqualStrings(oracle, al.items);
        }
    };
    try F.assert_str("", "\"\"");
    try F.assert_str("A", "\"A\"");
    try F.assert_str("AB", "\"AB\"");
    try F.assert_str("\n", "\"\\n\"");
    try F.assert_str("\r", "\"\\r\"");
    try F.assert_str("\t", "\"\\t\"");
    try F.assert_str("\\", "\"\\\\\"");
    try F.assert_str("'", "\"\\'\"");
    try F.assert_str("\"", "\"\\\"\"");
    try F.assert_str("\r\n", "\"\\r\\n\"");
    try F.assert_str("foo\nbar", "\"foo\\nbar\"");
    try F.assert_str("foo\nbar\n", "\"foo\\nbar\\n\"");
    try F.assert_str("foo\r\nbar", "\"foo\\r\\nbar\"");
    try F.assert_str("foo\r\nbar\r\n", "\"foo\\r\\nbar\\r\\n\"");
    try F.assert_str("\r\nfoo\r\nbar\r\n", "\"\\r\\nfoo\\r\\nbar\\r\\n\"");
    try F.assert_str("foo\tbar\n", "\"foo\\tbar\\n\"");
    try F.assert_limited_str("", 0, "\"\"");
    try F.assert_limited_str("", 1, "\"\"");
    try F.assert_limited_str("", 2, "\"\"");
    try F.assert_limited_str("A", 0, "\"..(1 more)\"");
    try F.assert_limited_str("A", 1, "\"A\"");
    try F.assert_limited_str("A", 2, "\"A\"");
    try F.assert_limited_str("AB", 0, "\"..(2 more)\"");
    try F.assert_limited_str("AB", 1, "\"A..(1 more)\"");
    try F.assert_limited_str("AB", 2, "\"AB\"");
}

// True if ASCII strings *a* and *b* are equal with case-insensitive comparison.
pub fn eql_i(a: []const u8, b: []const u8) bool {
    if (a.len != b.len) return false;
    for (a, b) |ac, bc| {
        std.debug.assert(std.ascii.isASCII(ac));
        std.debug.assert(std.ascii.isASCII(bc));
        if (ac == bc) continue;
        if (std.ascii.eqlIgnoreCase(a, b)) continue;
        return false;
    }
    return true;
}

// True if ASCII string *haystack* starts with *needle* with case-insensitive comparison.
pub fn startswith_i(needle: []const u8, haystack: []const u8) bool {
    if (needle.len > haystack.len) return false;
    return eql_i(needle, haystack[0..needle.len]);
}

pub fn upto(buff: []const u8, len: usize) []const u8 {
    if (buff.len <= len) return buff;
    return buff[0..len];
}

pub const LoggerOptions = struct {
    debug: bool = false,
    verbose: bool = false,
    prefix: ?[]const u8 = null,
};

pub const Logger = struct {
    options: LoggerOptions,

    pub fn init(options: LoggerOptions) Logger {
        return Logger{
            .options = options,
        };
    }

    pub fn debug(self: Logger, comptime fmt: []const u8, args: anytype) void {
        if (!self.options.debug) return;
        var arr: [1000]u8 = undefined;
        const msg = std.fmt.bufPrint(@as([]u8, &arr), fmt, args) catch return;
        std.log.debug("{s}{s}", .{ self.options.prefix orelse "", msg });
    }

    pub fn err(self: Logger, comptime fmt: []const u8, args: anytype) void {
        var arr: [1000]u8 = undefined;
        const msg = std.fmt.bufPrint(@as([]u8, &arr), fmt, args) catch return;
        std.log.err("{s}{s}", .{ self.options.prefix orelse "", msg });
    }

    pub fn info(self: Logger, comptime fmt: []const u8, args: anytype) void {
        if (!self.options.verbose) return;
        var arr: [1000]u8 = undefined;
        const msg = std.fmt.bufPrint(@as([]u8, &arr), fmt, args) catch return;
        std.log.info("{s}{s}", .{ self.options.prefix orelse "", msg });
    }

    pub fn warn(self: Logger, comptime fmt: []const u8, args: anytype) void {
        var arr: [1000]u8 = undefined;
        const msg = std.fmt.bufPrint(@as([]u8, &arr), fmt, args) catch return;
        std.log.warn("{s}{s}", .{ self.options.prefix orelse "", msg });
    }
};

pub const FBAPool = struct {
    buff: []u8,
    slots: []Slot,
    allocator: std.mem.Allocator,
    m: std.Thread.Mutex,
    const Slot = struct {
        buff: []u8,
        fba: std.heap.FixedBufferAllocator,
        free: bool,
        fn take(self: *FBAPool) void {
            self.fba.reset();
            for (0..self.buff.len) |i| self.buff[i] = 0xEE;
            self.free = false;
        }
        fn reset(self: *FBAPool) void {
            self.fba.reset();
            // for (0..self.buff.len) |i| self.buff[i] = 0xEE;
            self.free = true;
        }
    };
    const Lot = struct {
        allocator: std.mem.Allocator,
        idx: usize,
    };
    pub fn init(chunk_count: usize, chunk_size: usize, allocator: std.mem.Allocator) !FBAPool {
        const buff = try allocator.alloc(u8, try std.math.mul(usize, chunk_count, chunk_size));
        const slots = try allocator.alloc(Slot, chunk_size);
        // for (buff) |*c| c.* = 0xEE;
        for (0..chunk_size) |i| {
            const mem_bgn = try std.math.mul(usize, i, chunk_size);
            const mem_end = try std.math.add(usize, mem_bgn, chunk_size);
            const chunk_buff = buff[mem_bgn..mem_end];
            slots[i] = Slot{
                .buff = chunk_buff,
                .fba = std.heap.FixedBufferAllocator.init(chunk_buff),
                .free = false,
            };
        }
        return .{
            .buff = buff,
            .slots = slots,
            .allocator = allocator,
            .m = std.Thread.Mutex{},
        };
    }
    pub fn deinit(self: FBAPool) void {
        self.allocator.free(self.slots);
        self.allocator.free(self.buff);
    }
    fn _find_free(self: FBAPool) ?usize {
        for (0..self.slots.len) |i| if (self.slots[i].free) return i;
        return null;
    }
    pub fn get(self: *FBAPool) !Lot {
        self.m.lock();
        defer self.m.unlock();
        const free_idx = self._find_free() orelse return error.PoolFull;
        self.slots[free_idx].free = false;
        self.slots[free_idx].fba.reset();
        return Lot{
            .idx = free_idx,
            .allocator = self.slots[free_idx].fba.allocator(),
        };
    }
    pub fn claim(self: *FBAPool) !Lot {
        self.m.lock();
        defer self.m.unlock();
        const free_idx = self._find_free() orelse return error.PoolFull;
        self.slots[free_idx].take();
        return Lot{
            .idx = free_idx,
            .allocator = self.slots[free_idx].fba.allocator(),
        };
    }
    pub fn free(self: *FBAPool, lot: Lot) void {
        self.m.lock();
        defer self.m.unlock();
        self.slots[lot.idx].reset();
    }
};

const RangeRel = enum { inside, around, same, left, right, other };

pub const Range = struct {
    bgn: usize,
    end: usize,
};

pub fn range_rel(needle: Range, haystack: Range) !RangeRel {
    if (needle.bgn > needle.end) return error.ReverseRange;
    if (haystack.bgn > haystack.end) return error.ReverseRange;
    if (needle.bgn == haystack.bgn and needle.end == haystack.end) return .same;
    if (needle.bgn >= haystack.bgn and needle.end <= haystack.end) return .inside;
    if (haystack.bgn >= needle.bgn and haystack.end <= needle.end) return .around;
    // std.log.warn("needle.bgn={d} needle.end={d}", .{ needle.bgn, needle.end });
    // std.log.warn("haystack.bgn={d} haystack.end={d}", .{ haystack.bgn, haystack.end });
    if (needle.end < haystack.bgn) return .left;
    if (needle.bgn > haystack.end) return .right;
    return .other;
}

pub fn range_from_ptr(
    comptime T: type,
    ptr: *const T,
) Range {
    return Range{
        .bgn = @intFromPtr(ptr),
        .end = (@intFromPtr(ptr) + @sizeOf(T)) - 1,
    };
}

pub fn range_from_sli(
    comptime T: type,
    sli: []const T,
) ?Range {
    if (sli.len == 0) return null;
    return Range{
        .bgn = @intFromPtr(&sli[0]),
        .end = (@intFromPtr(&sli[sli.len - 1]) + @sizeOf(T)) - 1,
    };
}

fn test_range_rel(
    comptime map: []const u8,
    needle_addr: usize,
    needle_size: usize,
    haystack_addr: usize,
    haystack_size: usize,
    oracle: RangeRel,
) !void {
    _ = map;
    if (haystack_size == 0) return error.BadSize;
    if (needle_size == 0) return error.BadSize;
    try std.testing.expectEqual(oracle, try range_rel(
        Range{ .bgn = needle_addr, .end = needle_addr + needle_size - 1 },
        Range{ .bgn = haystack_addr, .end = haystack_addr + haystack_size - 1 },
    ));
}

test "range_rel" {
    try std.testing.expectEqual(error.ReverseRange, range_rel(
        Range{ .bgn = 0, .end = 1 },
        Range{ .bgn = 1, .end = 0 },
    ));
    try std.testing.expectEqual(error.ReverseRange, range_rel(
        Range{ .bgn = 1, .end = 0 },
        Range{ .bgn = 0, .end = 1 },
    ));
    try std.testing.expectEqual(error.ReverseRange, range_rel(
        Range{ .bgn = 1, .end = 0 },
        Range{ .bgn = 1, .end = 0 },
    ));

    // memory:         "|01234|"
    // haystack:       "|_    |"
    try test_range_rel("|-    |", 0, 1, 0, 1, .same);
    try test_range_rel("|--   |", 0, 2, 0, 1, .around);
    try test_range_rel("|---  |", 0, 3, 0, 1, .around);
    try test_range_rel("| -   |", 1, 1, 0, 1, .right);
    try test_range_rel("| --  |", 1, 2, 0, 1, .right);
    try test_range_rel("| --- |", 1, 3, 0, 1, .right);
    try test_range_rel("|  -  |", 2, 1, 0, 1, .right);
    try test_range_rel("|  -- |", 2, 2, 0, 1, .right);
    try test_range_rel("|  ---|", 2, 3, 0, 1, .right);
    try test_range_rel("|   - |", 3, 1, 0, 1, .right);
    try test_range_rel("|   --|", 3, 2, 0, 1, .right);
    try test_range_rel("|    -|", 4, 1, 0, 1, .right);

    // memory:         "|01234|"
    // haystack:       "|__   |"
    try test_range_rel("|-    |", 0, 1, 0, 2, .inside);
    try test_range_rel("|--   |", 0, 2, 0, 2, .same);
    try test_range_rel("|---  |", 0, 3, 0, 2, .around);
    try test_range_rel("| -   |", 1, 1, 0, 2, .inside);
    try test_range_rel("| --  |", 1, 2, 0, 2, .other);
    try test_range_rel("| --- |", 1, 3, 0, 2, .other);
    try test_range_rel("|  -  |", 2, 1, 0, 2, .right);
    try test_range_rel("|  -- |", 2, 2, 0, 2, .right);
    try test_range_rel("|  ---|", 2, 3, 0, 2, .right);
    try test_range_rel("|   - |", 3, 1, 0, 2, .right);
    try test_range_rel("|   --|", 3, 2, 0, 2, .right);
    try test_range_rel("|    -|", 4, 1, 0, 2, .right);

    // memory:         "|01234|"
    // haystack:       "|___  |"
    try test_range_rel("|-    |", 0, 1, 0, 3, .inside);
    try test_range_rel("|--   |", 0, 2, 0, 3, .inside);
    try test_range_rel("|---  |", 0, 3, 0, 3, .same);
    try test_range_rel("| -   |", 1, 1, 0, 3, .inside);
    try test_range_rel("| --  |", 1, 2, 0, 3, .inside);
    try test_range_rel("| --- |", 1, 3, 0, 3, .other);
    try test_range_rel("|  -  |", 2, 1, 0, 3, .inside);
    try test_range_rel("|  -- |", 2, 2, 0, 3, .other);
    try test_range_rel("|  ---|", 2, 3, 0, 3, .other);
    try test_range_rel("|   - |", 3, 1, 0, 3, .right);
    try test_range_rel("|   --|", 3, 2, 0, 3, .right);
    try test_range_rel("|    -|", 4, 1, 0, 3, .right);

    // memory:         "|01234|"
    // haystack:       "| _   |"
    try test_range_rel("|-    |", 0, 1, 1, 1, .left);
    try test_range_rel("|--   |", 0, 2, 1, 1, .around);
    try test_range_rel("|---  |", 0, 3, 1, 1, .around);
    try test_range_rel("| -   |", 1, 1, 1, 1, .same);
    try test_range_rel("| --  |", 1, 2, 1, 1, .around);
    try test_range_rel("| --- |", 1, 3, 1, 1, .around);
    try test_range_rel("|  -  |", 2, 1, 1, 1, .right);
    try test_range_rel("|  -- |", 2, 2, 1, 1, .right);
    try test_range_rel("|  ---|", 2, 3, 1, 1, .right);
    try test_range_rel("|   - |", 3, 1, 1, 1, .right);
    try test_range_rel("|   --|", 3, 2, 1, 1, .right);
    try test_range_rel("|    -|", 4, 1, 1, 1, .right);

    // memory:         "|01234|"
    // haystack:       "| __  |"
    try test_range_rel("|-    |", 0, 1, 1, 2, .left);
    try test_range_rel("|--   |", 0, 2, 1, 2, .other);
    try test_range_rel("|---  |", 0, 3, 1, 2, .around);
    try test_range_rel("| -   |", 1, 1, 1, 2, .inside);
    try test_range_rel("| --  |", 1, 2, 1, 2, .same);
    try test_range_rel("| --- |", 1, 3, 1, 2, .around);
    try test_range_rel("|  -  |", 2, 1, 1, 2, .inside);
    try test_range_rel("|  -- |", 2, 2, 1, 2, .other);
    try test_range_rel("|  ---|", 2, 3, 1, 2, .other);
    try test_range_rel("|   - |", 3, 1, 1, 2, .right);
    try test_range_rel("|   --|", 3, 2, 1, 2, .right);
    try test_range_rel("|    -|", 4, 1, 1, 2, .right);

    // memory:         "|01234|"
    // haystack:       "| ___ |"
    try test_range_rel("|-    |", 0, 1, 1, 3, .left);
    try test_range_rel("|--   |", 0, 2, 1, 3, .other);
    try test_range_rel("|---  |", 0, 3, 1, 3, .other);
    try test_range_rel("| -   |", 1, 1, 1, 3, .inside);
    try test_range_rel("| --  |", 1, 2, 1, 3, .inside);
    try test_range_rel("| --- |", 1, 3, 1, 3, .same);
    try test_range_rel("|  -  |", 2, 1, 1, 3, .inside);
    try test_range_rel("|  -- |", 2, 2, 1, 3, .inside);
    try test_range_rel("|  ---|", 2, 3, 1, 3, .other);
    try test_range_rel("|   - |", 3, 1, 1, 3, .inside);
    try test_range_rel("|   --|", 3, 2, 1, 3, .other);
    try test_range_rel("|    -|", 4, 1, 1, 3, .right);

    // memory:         "|01234|"
    // haystack:       "|  _  |"
    try test_range_rel("|-    |", 0, 1, 2, 1, .left);
    try test_range_rel("|--   |", 0, 2, 2, 1, .left);
    try test_range_rel("|---  |", 0, 3, 2, 1, .around);
    try test_range_rel("| -   |", 1, 1, 2, 1, .left);
    try test_range_rel("| --  |", 1, 2, 2, 1, .around);
    try test_range_rel("| --- |", 1, 3, 2, 1, .around);
    try test_range_rel("|  -  |", 2, 1, 2, 1, .same);
    try test_range_rel("|  -- |", 2, 2, 2, 1, .around);
    try test_range_rel("|  ---|", 2, 3, 2, 1, .around);
    try test_range_rel("|   - |", 3, 1, 2, 1, .right);
    try test_range_rel("|   --|", 3, 2, 2, 1, .right);
    try test_range_rel("|    -|", 4, 1, 2, 1, .right);

    // memory:         "|01234|"
    // haystack:       "|  __ |"
    try test_range_rel("|-    |", 0, 1, 2, 2, .left);
    try test_range_rel("|--   |", 0, 2, 2, 2, .left);
    try test_range_rel("|---  |", 0, 3, 2, 2, .other);
    try test_range_rel("| -   |", 1, 1, 2, 2, .left);
    try test_range_rel("| --  |", 1, 2, 2, 2, .other);
    try test_range_rel("| --- |", 1, 3, 2, 2, .around);
    try test_range_rel("|  -  |", 2, 1, 2, 2, .inside);
    try test_range_rel("|  -- |", 2, 2, 2, 2, .same);
    try test_range_rel("|  ---|", 2, 3, 2, 2, .around);
    try test_range_rel("|   - |", 3, 1, 2, 2, .inside);
    try test_range_rel("|   --|", 3, 2, 2, 2, .other);
    try test_range_rel("|    -|", 4, 1, 2, 2, .right);

    // memory:         "|01234|"
    // haystack:       "|  ___|"
    try test_range_rel("|-    |", 0, 1, 2, 3, .left);
    try test_range_rel("|--   |", 0, 2, 2, 3, .left);
    try test_range_rel("|---  |", 0, 3, 2, 3, .other);
    try test_range_rel("| -   |", 1, 1, 2, 3, .left);
    try test_range_rel("| --  |", 1, 2, 2, 3, .other);
    try test_range_rel("| --- |", 1, 3, 2, 3, .other);
    try test_range_rel("|  -  |", 2, 1, 2, 3, .inside);
    try test_range_rel("|  -- |", 2, 2, 2, 3, .inside);
    try test_range_rel("|  ---|", 2, 3, 2, 3, .same);
    try test_range_rel("|   - |", 3, 1, 2, 3, .inside);
    try test_range_rel("|   --|", 3, 2, 2, 3, .inside);
    try test_range_rel("|    -|", 4, 1, 2, 3, .inside);

    // memory:         "|01234|"
    // haystack:       "|   _ |"
    try test_range_rel("|-    |", 0, 1, 3, 1, .left);
    try test_range_rel("|--   |", 0, 2, 3, 1, .left);
    try test_range_rel("|---  |", 0, 3, 3, 1, .left);
    try test_range_rel("| -   |", 1, 1, 3, 1, .left);
    try test_range_rel("| --  |", 1, 2, 3, 1, .left);
    try test_range_rel("| --- |", 1, 3, 3, 1, .around);
    try test_range_rel("|  -  |", 2, 1, 3, 1, .left);
    try test_range_rel("|  -- |", 2, 2, 3, 1, .around);
    try test_range_rel("|  ---|", 2, 3, 3, 1, .around);
    try test_range_rel("|   - |", 3, 1, 3, 1, .same);
    try test_range_rel("|   --|", 3, 2, 3, 1, .around);
    try test_range_rel("|    -|", 4, 1, 3, 1, .right);

    // memory:         "|01234|"
    // haystack:       "|   __|"
    try test_range_rel("|-    |", 0, 1, 3, 2, .left);
    try test_range_rel("|--   |", 0, 2, 3, 2, .left);
    try test_range_rel("|---  |", 0, 3, 3, 2, .left);
    try test_range_rel("| -   |", 1, 1, 3, 2, .left);
    try test_range_rel("| --  |", 1, 2, 3, 2, .left);
    try test_range_rel("| --- |", 1, 3, 3, 2, .other);
    try test_range_rel("|  -  |", 2, 1, 3, 2, .left);
    try test_range_rel("|  -- |", 2, 2, 3, 2, .other);
    try test_range_rel("|  ---|", 2, 3, 3, 2, .around);
    try test_range_rel("|   - |", 3, 1, 3, 2, .inside);
    try test_range_rel("|   --|", 3, 2, 3, 2, .same);
    try test_range_rel("|    -|", 4, 1, 3, 2, .inside);

    // memory:         "|01234|"
    // haystack:       "|    _|"
    try test_range_rel("|-    |", 0, 1, 4, 1, .left);
    try test_range_rel("|--   |", 0, 2, 4, 1, .left);
    try test_range_rel("|---  |", 0, 3, 4, 1, .left);
    try test_range_rel("| -   |", 1, 1, 4, 1, .left);
    try test_range_rel("| --  |", 1, 2, 4, 1, .left);
    try test_range_rel("| --- |", 1, 3, 4, 1, .left);
    try test_range_rel("|  -  |", 2, 1, 4, 1, .left);
    try test_range_rel("|  -- |", 2, 2, 4, 1, .left);
    try test_range_rel("|  ---|", 2, 3, 4, 1, .around);
    try test_range_rel("|   - |", 3, 1, 4, 1, .left);
    try test_range_rel("|   --|", 3, 2, 4, 1, .around);
    try test_range_rel("|    -|", 4, 1, 4, 1, .same);

    //
}

/// Pool of items manipulated via handlers.
///
/// Storing an item returns a handle and item can be retrieved back
/// only using handle.  Supports handle iterator, item iterator
/// and handle selection via custom matching function.
///
/// *idx_bits* specify maximum amount of items that the pool can store
/// at the same time.  Once this capacity is reached, calling .put()
/// returns error.PoolFull until an item is deleted by calling .delete().
///
/// After deleting an item, storing a new one is guarranteed to produce
/// new handle until a grand total of *2^(idx_bits + gen_bits)*.  Once
/// this amount is stored, pool is exhausted and any attempts to store
/// new items return Error.PoolExhausted.
///
/// All memory is allocated on .init().  The memory footprint is O(n),
/// where `n` would scale with power of 2 to *idx_bits*.  That is, for each
/// item op to two to the *idx_bits*, the pool will allocate space for
/// the item itself plus some overhead.
///
/// It's not guarranteed that consecutive items will be located next
/// to each other in the memory.
pub fn OpaquePool(
    comptime T: type,
    comptime idx_bits: usize,
    comptime gen_bits: usize,
) type {
    const HandleIdxT = @Type(std.builtin.Type{ .Int = .{ .bits = idx_bits, .signedness = .unsigned } });
    const HandleT = @Type(std.builtin.Type{ .Int = .{ .bits = gen_bits + idx_bits, .signedness = .unsigned } });
    const CapaT = @Type(std.builtin.Type{ .Int = .{ .bits = gen_bits + idx_bits + 1, .signedness = .unsigned } });
    const CAPA_: CapaT = std.math.maxInt(HandleIdxT) + 1;
    const MAX_HANDLE = std.math.maxInt(HandleT);

    return struct {
        allocator: std.mem.Allocator,
        items: []T,
        used_handles: []?Handle,
        last_handle: ?Handle,
        stored_items: CapaT,

        const CAPA = CAPA_;
        pub const Handle = HandleT;
        const Item = T;
        const Error = error{ PoolFull, PoolExhausted, NoSuchItem };
        const MatchP = *const fn (*T, anytype) bool;
        const _Pool = @This();

        const ItemIterator = struct {
            cur: CapaT = 0,
            pool: _Pool,
            pub fn next(self: *ItemIterator) ?Item {
                while (self.cur < self.pool.items.len) {
                    if (self.pool.used_handles[self.cur]) |_| {
                        const out = self.pool.items[self.cur];
                        self.cur += 1;
                        return out;
                    }
                    self.cur += 1;
                }
                return null;
            }
        };

        const HandleIterator = struct {
            cur: CapaT = 0,
            pool: _Pool,
            pub fn next(self: *HandleIterator) ?Handle {
                while (self.cur < self.pool.items.len) {
                    if (self.pool.used_handles[self.cur]) |handle| {
                        self.cur += 1;
                        return handle;
                    }
                    self.cur += 1;
                }
                return null;
            }
        };

        const DebugView = struct {
            pool: *_Pool,
            fn init(p: *_Pool) DebugView {
                return .{ .pool = p };
            }
            pub fn format(
                self: DebugView,
                comptime fmt: []const u8,
                options: std.fmt.FormatOptions,
                writer: anytype,
            ) !void {
                _ = fmt;
                _ = options;
                try writer.writeAll("[ ");
                for (self.pool.used_handles, 0..) |maybe_handle, idx| {
                    if (idx > 0) try writer.writeAll(", ");
                    if (maybe_handle) |handle| {
                        try writer.print("{d}: H:{d}=>({any})", .{
                            idx,
                            handle,
                            self.pool.items[idx],
                        });
                    } else {
                        try writer.print("{d}: EMPTY", .{
                            idx,
                        });
                    }
                }
                try writer.print(" ] | last_handle={?d}", .{self.pool.last_handle});
            }
        };

        pub fn init(allocator: std.mem.Allocator) !_Pool {
            const items = try allocator.alloc(T, CAPA);
            const used_handles = try allocator.alloc(?Handle, CAPA);
            for (used_handles) |*maybe_handle| maybe_handle.* = null;
            return .{
                .allocator = allocator,
                .items = items,
                .used_handles = used_handles,
                .last_handle = null,
                .stored_items = 0,
            };
        }

        pub fn deinit(self: _Pool) void {
            self.allocator.free(self.items);
            self.allocator.free(self.used_handles);
        }

        fn _find_idx(self: _Pool, handle: Handle) ?CapaT {
            for (self.used_handles, 0..) |maybe_handle, idx| {
                if (maybe_handle == handle) return @intCast(idx);
            }
            return null;
        }

        fn _next_handle(self: *_Pool) ?Handle {
            if (self.last_handle == MAX_HANDLE) return null;
            self.last_handle = if (self.last_handle) |handle| handle + 1 else 0;
            return self.last_handle;
        }

        pub fn has(self: _Pool, handle: Handle) bool {
            return self._find_idx(handle) != null;
        }

        pub fn stored(self: _Pool) usize {
            return self.stored_items;
        }

        pub fn available(self: _Pool) usize {
            return CAPA - self.stored_items;
        }

        pub fn get(self: *_Pool, handle: Handle) ?T {
            if (self._find_idx(handle)) |idx| return self.items[idx];
            return null;
        }

        pub fn get_ptr(self: *_Pool, handle: Handle) ?*T {
            if (self._find_idx(handle)) |idx| return &self.items[idx];
            return null;
        }

        pub fn put(self: *_Pool, item: T) !Handle {
            for (0..self.used_handles.len) |idx| {
                if (self.used_handles[idx] != null) continue;
                if (self._next_handle()) |new_handle| {
                    self.items[idx] = item;
                    self.stored_items += 1;
                    self.used_handles[idx] = new_handle;
                    return new_handle;
                } else {
                    return Error.PoolExhausted;
                }
            }
            return Error.PoolFull;
        }

        pub fn require(self: *_Pool, handle: Handle) !T {
            // std.log.err("OpaquePool(...).require():self={*}", .{self});
            // std.log.err("OpaquePool(...).require():handle={any}", .{handle});
            // std.log.err("OpaquePool(...).require():  ... {any}", .{
            //     DebugView.init(self),
            // });
            return self.get(handle) orelse Error.NoSuchItem;
        }

        pub fn require_ptr(self: *_Pool, handle: Handle) !*T {
            return self.get_ptr(handle) orelse Error.NoSuchItem;
        }

        pub fn item_iterator(self: _Pool) ItemIterator {
            return ItemIterator{ .cur = 0, .pool = self };
        }

        pub fn handle_iterator(self: _Pool) HandleIterator {
            return HandleIterator{ .cur = 0, .pool = self };
        }

        pub fn select(self: *_Pool, match_fn: MatchP, match_ctx: anytype) ?Handle {
            for (0..self.used_handles.len) |idx| {
                if (self.used_handles[idx] == null) continue;
                if (match_fn(&self.items[idx], match_ctx)) return self.used_handles[idx];
            }
            return null;
        }

        pub fn update(self: *_Pool, handle: Handle, item: T) !void {
            if (self._find_idx(handle)) |idx| {
                self.items[idx] = item;
            } else {
                return Error.NoSuchItem;
            }
        }

        pub fn delete(self: *_Pool, handle: Handle) bool {
            // std.log.err("OpaquePool(...).delete():self={*}", .{self});
            // std.log.err("OpaquePool(...).delete():handle={any}", .{handle});
            if (self._find_idx(handle)) |idx| {
                self.used_handles[idx] = null;
                self.stored_items -= 1;
                return true;
            }
            return false;
        }
    };
}

fn OpaquePoolTestCase(
    comptime T: type,
    comptime idx_bits: usize,
    comptime gen_bits: usize,
    an_item: T,
) type {
    const Pool = OpaquePool(T, idx_bits, gen_bits);
    const DEBUG = false;

    return struct {
        pool: Pool,
        allocator: std.mem.Allocator,

        const OpaquePoolTestCaseT = @This();

        fn init(allocator: std.mem.Allocator) !OpaquePoolTestCaseT {
            return .{
                .pool = try Pool.init(allocator),
                .allocator = allocator,
            };
        }

        fn _debug(comptime fmt: []const u8, args: anytype) void {
            if (!DEBUG) return;
            std.log.err(fmt, args);
        }

        fn _show_pool(self: *OpaquePoolTestCaseT, comptime prefix: []const u8) !void {
            _debug("{s:.>20}: {any}", .{ prefix, Pool.DebugView.init(&self.pool) });
        }

        fn deinit(self: OpaquePoolTestCaseT) void {
            self.pool.deinit();
        }

        fn try_put(self: *OpaquePoolTestCaseT, oracle: Pool.Handle, item: Pool.Item) !void {
            _debug("OpaquePoolTestCase(..).try_put():oracle={d},item={any}", .{ oracle, item });
            try std.testing.expectEqual(oracle, try self.pool.put(item));
            try self._show_pool("after .put()");
        }

        fn try_update(self: *OpaquePoolTestCaseT, handle: Pool.Handle, item: Pool.Item) !void {
            _debug("OpaquePoolTestCase(..).try_update():handle={d},item={any}", .{ handle, item });
            try self.pool.update(handle, item);
            try std.testing.expectEqual(item, self.pool.get(handle));
            try self._show_pool("after .update()");
        }

        fn assert_full(self: *OpaquePoolTestCaseT) !void {
            try std.testing.expectError(Pool.Error.PoolFull, self.pool.put(an_item));
        }

        fn assert_exhausted(self: *OpaquePoolTestCaseT) !void {
            try std.testing.expectError(Pool.Error.PoolExhausted, self.pool.put(an_item));
        }

        const ItemOracle = struct {
            handle: Pool.Handle,
            item: T,
        };

        fn _predicted_item(oracles: []const ItemOracle, handle: Pool.Handle) ?T {
            for (oracles) |oracle| {
                if (oracle.handle == handle) return oracle.item;
            }
            return null;
        }

        /// Assert that precisely these items are stored under these handles.
        /// Items must be provided in the same order as would be returned by
        /// .item_iterator and .handle_iterator (which is an implementation detail,
        /// so be careful to update accordingly).
        /// Following assertions are made:
        ///   *  For every handle pair in *oracles*, .get returns the corresponding item.
        ///   *  For every handle pair in *oracles*, .get_ptr returns pointer to the corresponding item.
        ///   *  For every handle pair in *oracles*, .has returns true.
        ///   *  For every posible handle NOT in *oracles*, .get() returns null.
        ///   *  For every posible handle NOT in *oracles*, .get_ptr() returns null.
        ///   *  For every posible handle NOT in *oracles*, .require() returns error.
        ///   *  For every posible handle NOT in *oracles*, .require_ptr() returns error.
        ///   *  For every posible handle NOT in *oracles*, .delete() returns false.
        ///   *  For every posible handle NOT in *oracles*, .has() returns false.
        ///   *  For iterator retrieved from .item_iterator(), .next() will provide all
        ///      items as listed in *oracles*, then null on two subsequent calls.
        ///   *  For iterator retrieved from .handle(), .next() will provide all
        ///      items as listed in *oracles*, then null on two subsequent calls.
        fn assert_items(self: *OpaquePoolTestCaseT, oracles: []const ItemOracle) !void {
            _debug("OpaquePoolTestCase(..).assert_items():oracles={any}", .{oracles});
            for (0..std.math.maxInt(Pool.Handle)) |possible_handle| {
                const handle: Pool.Handle = @intCast(possible_handle);
                if (_predicted_item(oracles, handle)) |item| {
                    try std.testing.expectEqual(item, self.pool.get(handle));
                    try std.testing.expectEqual(item, self.pool.get_ptr(handle).?.*);
                    try std.testing.expect(self.pool.has(handle));
                } else {
                    try std.testing.expectEqual(null, self.pool.get(handle));
                    try std.testing.expectEqual(null, self.pool.get_ptr(handle));
                    try std.testing.expectEqual(Pool.Error.NoSuchItem, self.pool.require(handle));
                    try std.testing.expectEqual(Pool.Error.NoSuchItem, self.pool.require_ptr(handle));
                    try std.testing.expect(!self.pool.has(handle));
                    try std.testing.expect(!self.pool.delete(handle));
                    try std.testing.expectError(Pool.Error.NoSuchItem, self.pool.update(handle, an_item));
                }
            }
            var item_itr = self.pool.item_iterator();
            var handle_itr = self.pool.handle_iterator();
            for (oracles) |oracle| {
                try std.testing.expectEqual(oracle.item, item_itr.next() orelse return error.TestFailIteratorEmpty);
                try std.testing.expectEqual(oracle.handle, handle_itr.next() orelse return error.TestFailIteratorEmpty);
            }
            try std.testing.expectEqual(null, handle_itr.next());
            try std.testing.expectEqual(null, item_itr.next());
            try std.testing.expectEqual(null, handle_itr.next());
            try std.testing.expectEqual(null, item_itr.next());
        }

        const MatchCond = enum { eq_100, gt_101 };

        fn match(item: *T, ctx: anytype) bool {
            switch (ctx) {
                .eq_100 => return item.* == 100,
                .gt_101 => return item.* > 101,
            }
        }

        fn assert_selects(self: *OpaquePoolTestCaseT, oracle: Pool.Handle, match_cond: MatchCond) !void {
            _debug("OpaquePoolTestCase(..).assert_selects():oracle={any} match_cond={}", .{ oracle, @tagName(match_cond) });
            try std.testing.expectEqual(oracle, self.pool.select(match, match_cond));
        }

        fn assert_selects_nothing(self: *OpaquePoolTestCaseT, match_cond: MatchCond) !void {
            _debug("OpaquePoolTestCase(..).assert_selects():match_cond={}", .{@tagName(match_cond)});
            try std.testing.expectEqual(null, self.pool.select(match, match_cond));
        }

        /// assert that deletion of item with *handle* succeeds
        fn try_delete(self: *OpaquePoolTestCaseT, handle: Pool.Handle) !void {
            _debug("OpaquePoolTestCase(..).try_delete():handle={d}", .{handle});
            try std.testing.expect(self.pool.delete(handle));
            try self._show_pool("after .delete()");
        }

        fn assert_used(self: OpaquePoolTestCaseT, oracle: usize) !void {
            _debug("OpaquePoolTestCase(..).assert_used():oracle={d}", .{oracle});
            try std.testing.expectEqual(oracle, self.pool.stored());
            try std.testing.expectEqual(Pool.CAPA - oracle, self.pool.available());
        }
    };
}

test "OpaquePool: 1b + 1b" {
    // 1b + 1b
    var case = try OpaquePoolTestCase(u8, 1, 1, 255).init(std.testing.allocator);
    defer case.deinit();
    errdefer case._show_pool("FAILING") catch {};

    try case._show_pool("after .init()");
    try case.assert_used(0);

    try case.try_put(0, 100);
    try case.assert_items(&.{
        .{ .handle = 0, .item = 100 },
    });
    try case.assert_used(1);

    try case.try_put(1, 101);
    try case.assert_items(&.{
        .{ .handle = 0, .item = 100 },
        .{ .handle = 1, .item = 101 },
    });
    try case.assert_selects(0, .eq_100);
    try case.assert_selects_nothing(.gt_101);
    try case.assert_used(2);
    try case.assert_full();

    try case.try_delete(1);
    try case.assert_items(&.{
        .{ .handle = 0, .item = 100 },
    });
    try case.assert_used(1);

    try case.try_put(2, 102);
    try case.assert_items(&.{
        .{ .handle = 0, .item = 100 },
        .{ .handle = 2, .item = 102 },
    });
    try case.assert_used(2);
    try case.assert_full();

    try case.try_update(2, 122);
    try case.assert_items(&.{
        .{ .handle = 0, .item = 100 },
        .{ .handle = 2, .item = 122 },
    });
    try case.assert_selects(0, .eq_100);
    try case.assert_selects(2, .gt_101);
    try case.assert_used(2);
    try case.assert_full();

    try case.try_delete(0);
    try case.assert_items(&.{
        .{ .handle = 2, .item = 122 },
    });
    try case.assert_selects_nothing(.eq_100);
    try case.assert_selects(2, .gt_101);
    try case.assert_used(1);

    try case.try_put(3, 103);
    try case.assert_items(&.{
        .{ .handle = 3, .item = 103 },
        .{ .handle = 2, .item = 122 },
    });
    try case.assert_selects_nothing(.eq_100);
    try case.assert_selects(3, .gt_101);
    try case.assert_used(2);
    try case.assert_full();

    try case.try_delete(3);
    try case.assert_items(&.{
        .{ .handle = 2, .item = 122 },
    });
    try case.assert_selects_nothing(.eq_100);
    try case.assert_selects(2, .gt_101);
    try case.assert_used(1);

    try case.try_delete(2);
    try case.assert_items(&.{});
    try case.assert_selects_nothing(.eq_100);
    try case.assert_selects_nothing(.gt_101);
    try case.assert_used(0);

    try case.assert_exhausted();
}

test "OpaquePool: 2b + 2b" {
    // 2b + 2b
    var case = try OpaquePoolTestCase(u8, 4, 4, 255).init(std.testing.allocator);
    defer case.deinit();
    errdefer case._show_pool("FAILING") catch {};

    try case.try_put(0, 100);
    try case.try_put(1, 101);
    try case.try_put(2, 102);
    try case.try_put(3, 103);
    try case.try_put(4, 104);
    try case.try_put(5, 105);
    try case.try_put(6, 106);
    try case.try_put(7, 107);
    try case.try_put(8, 108);
    try case.try_put(9, 109);
    try case.try_put(10, 110);
    try case.try_put(11, 111);
    try case.try_put(12, 112);
    try case.try_put(13, 113);
    try case.try_put(14, 114);
    try case.try_put(15, 115);
    try case.assert_items(&.{
        .{ .handle = 0, .item = 100 },
        .{ .handle = 1, .item = 101 },
        .{ .handle = 2, .item = 102 },
        .{ .handle = 3, .item = 103 },
        .{ .handle = 4, .item = 104 },
        .{ .handle = 5, .item = 105 },
        .{ .handle = 6, .item = 106 },
        .{ .handle = 7, .item = 107 },
        .{ .handle = 8, .item = 108 },
        .{ .handle = 9, .item = 109 },
        .{ .handle = 10, .item = 110 },
        .{ .handle = 11, .item = 111 },
        .{ .handle = 12, .item = 112 },
        .{ .handle = 13, .item = 113 },
        .{ .handle = 14, .item = 114 },
        .{ .handle = 15, .item = 115 },
    });
    try case.assert_used(16);
    try case.assert_full();

    try case.try_delete(13);
    try case.assert_items(&.{
        .{ .handle = 0, .item = 100 },
        .{ .handle = 1, .item = 101 },
        .{ .handle = 2, .item = 102 },
        .{ .handle = 3, .item = 103 },
        .{ .handle = 4, .item = 104 },
        .{ .handle = 5, .item = 105 },
        .{ .handle = 6, .item = 106 },
        .{ .handle = 7, .item = 107 },
        .{ .handle = 8, .item = 108 },
        .{ .handle = 9, .item = 109 },
        .{ .handle = 10, .item = 110 },
        .{ .handle = 11, .item = 111 },
        .{ .handle = 12, .item = 112 },
        .{ .handle = 14, .item = 114 },
        .{ .handle = 15, .item = 115 },
    });
    try case.assert_used(15);

    try case.try_put(16, 116);
    try case.assert_items(&.{
        .{ .handle = 0, .item = 100 },
        .{ .handle = 1, .item = 101 },
        .{ .handle = 2, .item = 102 },
        .{ .handle = 3, .item = 103 },
        .{ .handle = 4, .item = 104 },
        .{ .handle = 5, .item = 105 },
        .{ .handle = 6, .item = 106 },
        .{ .handle = 7, .item = 107 },
        .{ .handle = 8, .item = 108 },
        .{ .handle = 9, .item = 109 },
        .{ .handle = 10, .item = 110 },
        .{ .handle = 11, .item = 111 },
        .{ .handle = 12, .item = 112 },
        .{ .handle = 16, .item = 116 },
        .{ .handle = 14, .item = 114 },
        .{ .handle = 15, .item = 115 },
    });

    try case.try_delete(0);
    try case.try_delete(1);
    try case.try_delete(2);
    try case.try_delete(3);
    try case.try_delete(4);
    try case.try_delete(5);
    try case.try_delete(6);
    try case.try_delete(7);
    try case.try_delete(8);
    try case.try_delete(9);
    try case.try_delete(10);
    try case.try_delete(11);
    try case.try_delete(12);
    try case.try_delete(14);
    try case.try_delete(15);
    try case.try_delete(16);
    try case.assert_items(&.{});
    try case.assert_used(0);

    for (17..256) |handle| {
        try case.try_put(@intCast(handle), 234);
        try case.try_delete(@intCast(handle));
        try case.assert_items(&.{});
        try case.assert_used(0);
    }
    try case.assert_used(0);
    try case.assert_exhausted();
}

test "OpaquePool: 8b + 8b" {
    const Pool = OpaquePool(u8, 8, 8);
    var pool = try Pool.init(std.testing.allocator);
    defer pool.deinit();
    try std.testing.expectEqual(0, pool.stored());
    try std.testing.expectEqual(256, pool.available());
}

test "OpaquePool: 16b + 16b" {
    const Pool = OpaquePool(u8, 16, 16);
    var pool = try Pool.init(std.testing.allocator);
    defer pool.deinit();
    try std.testing.expectEqual(0, pool.stored());
    try std.testing.expectEqual(65536, pool.available());
}

fn Stack(
    comptime T: type,
    comptime capa: usize,
) type {
    return struct {
        buff: [capa]T,
        len: usize = 0,
        capa_: usize = capa,
        const Error = error{ NotEnoughSpace, NotEnoughItems };
        const Self = @This();
        fn append(self: *Self, item: T) !void {
            const new_len = try std.math.add(usize, self.end, 1);
            if (new_len > self.buff.len) return Error.NotEnoughSpace;
            const old_len = self.len;
            self.len = new_len;
            self.buff[old_len] = item;
        }
        fn append_slice(self: *Self, slice: []T) !void {
            const new_len = try std.math.add(usize, self.end, slice.len);
            if (new_len > self.buff.len) return Error.NotEnoughSpace;
            const old_len = self.len;
            self.len = new_len;
            @memcpy(self.buff[old_len..new_len], slice);
        }
        fn drop(self: *Self, n: usize) void {
            if (n < self.len) self.len -= n else self.len == 0;
        }
        fn drop_atleast(self: *Self, n: usize) !void {
            if (self.len < n) return Error.NotEnoughItems;
        }
        fn items(self: Self) []T {
            return self.buff[0..self.len];
        }
        fn pop_item(self: *Self) Error!T {
            if (self.len == 0) return Error.NotEnoughItems;
            const out = self.buff[self.len - 1];
            self.len -= 1;
            return out;
        }
        fn pop_slice(self: *Self, n: usize, allocator: std.mem.Allocator) ![]T {
            if (self.len < n) return Error.NotEnoughItems;
            const new_len = try std.math.sub(usize, self.end, n);
            const out = try allocator.alloc(n, T);
            @memcpy(out, self.buff[new_len..]);
            self.len = new_len;
            return out;
        }
        fn pop_slice_atmost(self: *Self, n: usize, allocator: std.mem.Allocator) ![]T {
            const actual_n = @min(self.len, n);
            return self.pop_slice(actual_n, allocator);
        }
        fn reset(self: *Self) !void {
            self.len = 0;
        }
    };
}

/// Return index of *haystack* item which is identical to *needle*
/// (i.e. occupies same memory).
///
/// Unlike std.mem.IndexOf*() ignores values and takes only memory
/// locations into account.
///
/// Return null if there's no such item.
fn index_of_overlap(comptime T: type, needle: *T, haystack: []T) ?usize {
    const needle_addr = @intFromPtr(needle);
    const haystack_bgn_addr = @intFromPtr(haystack.ptr);
    const haystack_end_addr = @intFromPtr(haystack[haystack.len - 1]);
    if (haystack.len == 0) return null;
    if (needle_addr < haystack_bgn_addr) return null;
    if (needle_addr > haystack_end_addr) return null;
    const offset = haystack_bgn_addr - needle_addr;
    const idx = offset / @sizeOf(T);
    return idx;
}

const IdxRange = struct { bgn: usize, end: usize };

/// Return indices of *haystack* where some items of range are identical
/// items in *src* (i.e. occupy the same memory).
///
/// Return null if there's no overlap.
///
/// Ignores values and takes only memory locations into account.
///
/// Useful to check before potentially copying overlapping slices.
fn overlapping_range(comptime T: type, src: []T, dst: []T) ?IdxRange {
    if (src.len == 0 or dst.len == 0) return null;
    const overlap_bgn = index_of_overlap(src[0], dst) orelse return null;
    const overlap_end = index_of_overlap(src[src.len - 1], dst) orelse return null;
    return IdxRange{ .bgn = overlap_bgn, .end = overlap_end };
}

pub fn TSQueue(comptime T: type, comptime idx_bits: usize) type {
    const IdxT = @Type(std.builtin.Type{ .Int = .{ .bits = idx_bits, .signedness = .unsigned } });
    const CapaT = @Type(std.builtin.Type{ .Int = .{ .bits = idx_bits + 1, .signedness = .unsigned } });
    const CAPA_: CapaT = @intCast(std.math.maxInt(IdxT) + 1);
    return struct {
        buff: [Capa]Item = undefined,
        head: Idx = 0,
        items: CapaT = 0,
        c: std.Thread.Condition,
        m: std.Thread.Mutex,

        const _Queue = @This();
        const Item = T;
        const Idx = IdxT;
        const Capa = CAPA_;
        const Error = error{QueueFull};

        const DebugView = struct {
            queue: *_Queue,
            fn init(q: *_Queue) DebugView {
                return .{ .queue = q };
            }
            pub fn format(
                self: DebugView,
                comptime fmt: []const u8,
                options: std.fmt.FormatOptions,
                writer: anytype,
            ) !void {
                _ = fmt;
                _ = options;
                try writer.writeAll("[ ");

                // map out which slots are occupied
                var map: [_Queue.Capa]bool = undefined;
                for (0.._Queue.Capa) |i| map[i] = false;
                var map_idx = self.queue.head;
                var map_togo = self.queue.items;
                while (map_togo > 0) {
                    map[map_idx] = true;
                    map_idx +%= 1;
                    map_togo -= 1;
                }

                // show queue contents
                for (self.queue.buff, 0..) |slot, idx| {
                    if (idx > 0) try writer.writeByte(' ');
                    if (map[idx]) {
                        try writer.print("{d}: ({any})", .{ idx, slot });
                    } else {
                        try writer.print("{d}: EMPTY", .{idx});
                    }
                }

                try writer.print(" ] | head={d} items={d}", .{ self.queue.head, self.queue.items });
            }
        };

        pub fn init() _Queue {
            return .{
                .c = std.Thread.Condition{},
                .m = std.Thread.Mutex{},
            };
        }

        pub fn append(self: *_Queue, item: Item) !void {
            if (self.items == Capa) return Error.QueueFull;
            {
                self.m.lock();
                defer self.m.unlock();
                const virt_idx: usize = self.head + self.items;
                self.buff[@mod(virt_idx, _Queue.Capa)] = item;
                self.items += 1;
            }
            self.c.signal();
        }

        pub fn pop(self: *_Queue) Item {
            var out: Item = undefined;
            {
                self.m.lock();
                defer self.m.unlock();
                while (self.items == 0) {
                    self.c.wait(&self.m);
                }
                out = self.buff[self.head];
                self.head +%= 1;
                self.items -= 1;
            }
            return out;
        }
    };
}

pub fn Queue(comptime T: type, comptime idx_bits: usize) type {
    const IdxT = @Type(std.builtin.Type{ .Int = .{ .bits = idx_bits, .signedness = .unsigned } });
    const CapaT = @Type(std.builtin.Type{ .Int = .{ .bits = idx_bits + 1, .signedness = .unsigned } });
    const CAPA_: CapaT = @intCast(std.math.maxInt(IdxT) + 1);
    return struct {
        buff: [Capa]Item = undefined,
        head: Idx = 0,
        items: CapaT = 0,

        const _Queue = @This();
        const Item = T;
        const Idx = IdxT;
        const Capa = CAPA_;
        const Error = error{QueueFull};

        const DebugView = struct {
            queue: *_Queue,
            fn init(q: *_Queue) DebugView {
                return .{ .queue = q };
            }
            pub fn format(
                self: DebugView,
                comptime fmt: []const u8,
                options: std.fmt.FormatOptions,
                writer: anytype,
            ) !void {
                _ = fmt;
                _ = options;
                try writer.writeAll("[ ");

                // map out which slots are occupied
                var map: [_Queue.Capa]bool = undefined;
                for (0.._Queue.Capa) |i| map[i] = false;
                var map_idx = self.queue.head;
                var map_togo = self.queue.items;
                while (map_togo > 0) {
                    map[map_idx] = true;
                    map_idx +%= 1;
                    map_togo -= 1;
                }

                // show queue contents
                for (self.queue.buff, 0..) |slot, idx| {
                    if (idx > 0) try writer.writeByte(' ');
                    if (map[idx]) {
                        try writer.print("{d}: ({any})", .{ idx, slot });
                    } else {
                        try writer.print("{d}: EMPTY", .{idx});
                    }
                }

                try writer.print(" ] | head={d} items={d}", .{ self.queue.head, self.queue.items });
            }
        };

        pub fn init() _Queue {
            return .{};
        }

        const ItemPtrIterator = struct {
            queue: *_Queue,
            offset: usize = 0,

            pub fn next(self: *ItemPtrIterator) ?*Item {
                if (self.offset >= self.queue.items) return null;
                const virt_idx: usize = self.queue.head + self.offset;
                self.offset += 1;
                return &self.queue.buff[@mod(virt_idx, _Queue.Capa)];
            }
        };

        pub fn append(self: *_Queue, item: Item) !void {
            if (self.items == Capa) return Error.QueueFull;
            const virt_idx: usize = self.head + self.items;
            self.buff[@mod(virt_idx, _Queue.Capa)] = item;
            self.items += 1;
        }

        pub fn append_ptr(self: *_Queue, item: Item) !*Item {
            if (self.items == Capa) return Error.QueueFull;
            const virt_idx: usize = self.head + self.items;
            self.buff[@mod(virt_idx, _Queue.Capa)] = item;
            self.items += 1;
            return &self.buff[@mod(virt_idx, _Queue.Capa)];
        }

        pub fn peek_ptr(self: *_Queue) ?*Item {
            if (self.items == 0) return null;
            return &self.buff[self.head];
        }

        pub fn ptr_iter(self: *_Queue) ItemPtrIterator {
            return .{ .queue = self };
        }

        pub fn pop(self: *_Queue) Item {
            var out: Item = undefined;
            out = self.buff[self.head];
            self.head +%= 1;
            self.items -= 1;
            return out;
        }
    };
}

fn QueueTestCase(
    comptime T: type,
    comptime idx_bits: usize,
    an_item: T,
) type {
    const QueueT = Queue(T, idx_bits);
    const DEBUG = false;

    return struct {
        queue: QueueT,

        const Self = @This();

        fn init() Self {
            return .{
                .queue = QueueT.init(),
            };
        }

        fn _debug(comptime fmt: []const u8, args: anytype) void {
            if (!DEBUG) return;
            std.log.err(fmt, args);
        }

        fn _show_queue(self: *Self, comptime prefix: []const u8) !void {
            _debug("{s:.>20}: {any}", .{ prefix, QueueT.DebugView.init(&self.queue) });
        }

        fn try_append(self: *Self, item: QueueT.Item) !void {
            _debug("QueueTestCase(..).try_append():item={any}", .{item});
            try self.queue.append(item);
            try self._show_queue("after .append()");
        }

        fn assert_full(self: *Self) !void {
            try std.testing.expectError(QueueT.Error.QueueFull, self.queue.append(an_item));
        }

        const ItemOracle = struct {
            idx: QueueT.Idx,
            item: T,
        };

        fn _predicted_item(oracles: []const ItemOracle, idx: QueueT.Idx) ?T {
            for (oracles) |oracle| {
                if (oracle.idx == idx) return oracle.item;
            }
            return null;
        }

        /// Assert that precisely these items are stored in queue buffer.
        fn assert_items(self: *Self, oracles: []const ItemOracle) !void {
            _debug("QueueTestCase(..).assert_items():oracles={any}", .{oracles});
            for (0..self.queue.buff.len) |i| {
                const idx: QueueT.Idx = @intCast(i);
                if (_predicted_item(oracles, idx)) |item| {
                    try std.testing.expectEqual(item, self.queue.buff[idx]);
                }
            }
        }

        /// assert that popping gets us item *oracle*
        fn try_pop(self: *Self, oracle: QueueT.Item) !void {
            _debug("QueueTestCase(..).try_pop():oracle={d}", .{oracle});
            try std.testing.expectEqual(oracle, self.queue.pop());
            try self._show_queue("after .pop()");
        }

        fn assert_used(self: Self, oracle: usize) !void {
            _debug("QueueTestCase(..).assert_used():oracle={d}", .{oracle});
            try std.testing.expectEqual(oracle, self.queue.items);
        }
    };
}

test "Queue: 1b" {
    // 1b
    var case = QueueTestCase(u8, 1, 255).init();
    errdefer case._show_queue("FAILING") catch {};

    try case._show_queue("after .init()");
    try case.assert_used(0);

    try case.try_append(100);
    try case.assert_items(&.{
        .{ .idx = 0, .item = 100 },
    });
    try case.assert_used(1);

    try case.try_append(101);
    try case.assert_items(&.{
        .{ .idx = 0, .item = 100 },
        .{ .idx = 1, .item = 101 },
    });
    try case.assert_used(2);
    try case.assert_full();

    try case.try_pop(100);
    try case.assert_items(&.{
        .{ .idx = 1, .item = 101 },
    });
    try case.assert_used(1);

    try case.try_append(102);
    try case.assert_items(&.{
        .{ .idx = 0, .item = 102 },
        .{ .idx = 1, .item = 101 },
    });
    try case.assert_used(2);
    try case.assert_full();

    try case.try_pop(101);
    try case.assert_items(&.{
        .{ .idx = 0, .item = 102 },
    });
    try case.assert_used(1);

    try case.try_pop(102);
    try case.assert_items(&.{});
    try case.assert_used(0);
}

test "Queue: 3b" {
    // 1b
    var case = QueueTestCase(u8, 3, 255).init();
    errdefer case._show_queue("FAILING") catch {};

    try case._show_queue("after .init()");
    try case.assert_used(0);

    try case.try_append(100);
    try case.assert_items(&.{
        .{ .idx = 0, .item = 100 },
    });
    try case.assert_used(1);

    try case.try_append(101);
    try case.try_append(102);
    try case.try_append(103);
    try case.assert_items(&.{
        .{ .idx = 0, .item = 100 },
        .{ .idx = 1, .item = 101 },
        .{ .idx = 2, .item = 102 },
        .{ .idx = 3, .item = 103 },
    });
    try case.assert_used(4);

    try case.try_pop(100);
    try case.try_pop(101);
    try case.assert_items(&.{
        .{ .idx = 2, .item = 102 },
        .{ .idx = 3, .item = 103 },
    });
    try case.assert_used(2);

    try case.try_append(104);
    try case.try_append(105);
    try case.try_append(106);
    try case.try_append(107);
    try case.try_append(108);
    try case.try_append(109);
    try case.assert_items(&.{
        .{ .idx = 0, .item = 108 },
        .{ .idx = 1, .item = 109 },
        .{ .idx = 2, .item = 102 },
        .{ .idx = 3, .item = 103 },
        .{ .idx = 4, .item = 104 },
        .{ .idx = 5, .item = 105 },
        .{ .idx = 6, .item = 106 },
        .{ .idx = 7, .item = 107 },
    });
    try case.assert_used(8);
    try case.assert_full();

    try case.try_pop(102);
    try case.try_pop(103);
    try case.try_pop(104);
    try case.try_pop(105);
    try case.assert_items(&.{
        .{ .idx = 0, .item = 108 },
        .{ .idx = 1, .item = 109 },
        .{ .idx = 6, .item = 106 },
        .{ .idx = 7, .item = 107 },
    });
    try case.assert_used(4);

    try case.try_append(110);
    try case.try_append(111);
    try case.try_append(112);
    try case.try_append(113);
    try case.assert_items(&.{
        .{ .idx = 0, .item = 108 },
        .{ .idx = 1, .item = 109 },
        .{ .idx = 2, .item = 110 },
        .{ .idx = 3, .item = 111 },
        .{ .idx = 4, .item = 112 },
        .{ .idx = 5, .item = 113 },
        .{ .idx = 6, .item = 106 },
        .{ .idx = 7, .item = 107 },
    });
    try case.assert_used(8);
    try case.assert_full();

    try case.try_pop(106);
    try case.try_pop(107);
    try case.try_pop(108);
    try case.try_pop(109);
    try case.try_pop(110);
    try case.try_pop(111);
    try case.try_pop(112);
    try case.assert_items(&.{
        .{ .idx = 5, .item = 113 },
    });
    try case.assert_used(1);

    try case.try_pop(113);
    try case.assert_items(&.{});
    try case.assert_used(0);
}

/// Distributes single request to a group of worker threads.  Caller should
/// assign each thread worker one of pipes, where it waits for a request
/// using Pipe.wait_request() and after processing it, sends response back
/// using Pipe.send_response().
///
/// The parent thread will send requests using .queue_batch(), which will
/// accept request and a timeout, and return an iterator, which in turn
/// can be used to go through all requests which were returned within the
/// specified timeout.
///
/// Request and response (RequestT and ResponseT, respectively) can be any
/// types.  Furhermore, there is CtxIdT, which is used to identify responses
/// coming from the pipe group.
///
/// See indivudual functions for details.
///
/// Example:
///
///     // --------------
///     // in main thread
///     // --------------
///     const Pipes = BatchQueue(Foo, Bar, Task.Id);
///     var pipes = try Pipes.init(10, allocator);
///     defer pipes.deinit();
///
///     var task = Task.create();
///     try pool.spawn(my_handler, task);
///
///     // --------------------------
///     // within each thread handler
///     // --------------------------
///     const my_pipe_ptr = try pipes.enable(task.id);
///     const req: Foo = my_pipe_ptr.wait_request();
///             // .. process it...
///     my_pipe_ptr.send_response(Bar{...});
///
///     // -----------------------
///     // back in the main thread
///     // -----------------------
///     var itr = pipes.queue_batch(Foo{...});
///     defer itr.deinit();
///     while (itr.next()) |msg| {
///         std.log.info(
///             "got response for task: id={any}, reponse={any} (took {d} ms}",
///             .{msg.ctx_id, msg.response, msg.work_duration_ms},
///         );
///     }
///
///
pub fn BatchQueue(RequestT: type, ResponseT: type, CtxIdT: type) type {
    return struct {
        ctx_ids: []?CtxIdT,
        batches: InnerBatchQueue,
        allocator: std.mem.Allocator,
        m: std.Thread.Mutex,
        c: std.Thread.Condition,
        next_batch_id: usize = 0,

        pub fn init(size: usize, allocator: std.mem.Allocator) !BatchQueueT {
            const m = std.Thread.Mutex{};
            const c = std.Thread.Condition{};
            const ctx_ids = try allocator.alloc(?CtxIdT, size);
            for (0..size) |i| ctx_ids[i] = null;
            return .{
                .batches = InnerBatchQueue.init(),
                .ctx_ids = ctx_ids,
                .allocator = allocator,
                .m = m,
                .c = c,
            };
        }

        pub fn _dump(self: *BatchQueueT, comptime context: []const u8) void {
            std.log.debug("--- BatchQueue BEGIN " ++ context ++ " ---", .{});
            std.log.debug(" .. next_batch_id: {d}", .{self.next_batch_id});
            for (self.ctx_ids) |maybe_ctx_id| {
                std.log.debug(" .. ctx_id: {?d}", .{maybe_ctx_id});
            }
            var iter = self.batches.ptr_iter();
            while (iter.next()) |batch_ptr| {
                std.log.debug(" .. B: {any}", .{Batch.DebugView.init(batch_ptr)});
            }
            std.log.debug("--- BatchQueue END " ++ context ++ " ---", .{});
        }

        const BatchQueueT = @This();
        const InnerBatchQueue = Queue(Batch, 8);
        const Result = struct {
            response: ResponseT,
            work_duration_ms: u64,
            pub fn format(self: Result, comptime fmt: []const u8, options: std.fmt.FormatOptions, writer: anytype) !void {
                _ = fmt;
                _ = options;
                try writer.print(
                    "Result{{.response={any},.work_duration_ms={d}}}",
                    .{ self.response, self.work_duration_ms },
                );
            }
        };
        const JobState = enum { new, working, result_pending, result_retrieved, timeouted };
        const Job = union(JobState) {
            new,
            working: i64,
            result_pending: Result,
            result_retrieved,
            timeouted: CtxIdT,
        };
        const JobContext = struct {
            batch: *Batch,
            ctx_id: CtxIdT,
            request: RequestT,
            pub fn respond(self: JobContext, response: ResponseT) !void {
                try self.batch.send_response(self.ctx_id, response);
            }
        };

        const ResponseMessage = struct {
            ctx_id: CtxIdT,
            response: ResponseT,
            duration_ms: u64,
        };

        pub const Batch = struct {
            bq: *BatchQueueT,
            batch_id: usize,
            ctx_ids: []CtxIdT,
            request: RequestT,
            remaining_ns: u64,
            jobs: []Job,
            disowned: bool = false,

            fn init(
                bq: *BatchQueueT,
                batch_id: usize,
                request: RequestT,
                timeout_ns: u64,
            ) !Batch {
                var ctx_num: usize = 0; // how many live contexts are there
                for (0..bq.ctx_ids.len) |i| ctx_num += (if (bq.ctx_ids[i] == null) 0 else 1);
                var jobs_al = try std.ArrayList(Job).initCapacity(bq.allocator, ctx_num);
                var context_ids_al = try std.ArrayList(CtxIdT).initCapacity(bq.allocator, ctx_num);
                for (0..bq.ctx_ids.len) |idx| {
                    const ctx_id = bq.ctx_ids[idx] orelse continue;
                    try context_ids_al.append(ctx_id);
                    try jobs_al.append(.new);
                }
                return Batch{
                    .bq = bq,
                    .batch_id = batch_id,
                    .ctx_ids = try context_ids_al.toOwnedSlice(),
                    .request = request,
                    .remaining_ns = timeout_ns,
                    .jobs = try jobs_al.toOwnedSlice(),
                };
            }

            const DebugView = struct {
                batch: *Batch,
                fn init(b: *Batch) DebugView {
                    return .{ .batch = b };
                }
                pub fn format(
                    self: DebugView,
                    comptime fmt: []const u8,
                    options: std.fmt.FormatOptions,
                    writer: anytype,
                ) !void {
                    _ = fmt;
                    _ = options;
                    try writer.print("Batch{{{s}batch_id={d},request={any},remaining_ns={d},jobs=[", .{
                        if (self.batch.disowned) "DISOWNED," else "",
                        self.batch.batch_id,
                        self.batch.request,
                        self.batch.remaining_ns,
                    });
                    var comma: bool = false;
                    for (0..self.batch.jobs.len) |idx| {
                        if (comma) try writer.writeByte(',');
                        if (!comma) comma = true;
                        const job = self.batch.jobs[idx];
                        try writer.print("@{d}=", .{self.batch.ctx_ids[idx]});
                        switch (job) {
                            .new => try writer.writeAll(".new"),
                            .working => |ts| try writer.print(".working={d}", .{ts}),
                            .result_pending => |r| try writer.print(".result_pending=.{{{any}}}", .{r}),
                            .result_retrieved => try writer.writeAll(".result_retrieved"),
                            .timeouted => |c| try writer.print(".timeouted={d}", .{c}),
                        }
                    }
                    try writer.writeAll(" ]");
                }
            };

            /// a disowned batch will automatically deinit (ie. pop from queue) once the
            /// last request has been processed.
            pub fn disown(self: *Batch) void {
                self.disowned = true;
            }

            /// Unqueue and free memory.
            ///
            /// Asserts that this batch was at the top of the queue (it's not possible
            /// to dequeue items later than head of the queue).
            pub fn deinit(self: *Batch) void {
                const popped = self.bq.batches.pop();
                if (popped.batch_id != self.batch_id) unreachable; // deleting batches in wrong order
                self.bq.allocator.free(self.ctx_ids);
                self.bq.allocator.free(self.jobs);
            }

            fn _job_ptr_by_ctx_id(self: *Batch, ctx_id: CtxIdT) ?*Job {
                const idx = self._idx_by_ctx_id(ctx_id) orelse return null;
                return &self.jobs[idx];
            }

            fn _idx_by_ctx_id(self: *Batch, ctx_id: CtxIdT) ?usize {
                for (0..self.ctx_ids.len) |idx| {
                    if (self.ctx_ids[idx] != ctx_id) continue;
                    return idx;
                }
                return null;
            }

            fn assign_new_job(self: *Batch, ctx_id: CtxIdT) !?JobContext {
                const job_ptr = self._job_ptr_by_ctx_id(ctx_id) orelse return error.InvalidCtxId;
                if (job_ptr.* != .new) return null;
                job_ptr.* = Job{ .working = std.time.milliTimestamp() };
                return JobContext{
                    .batch = self,
                    .request = self.request,
                    .ctx_id = ctx_id,
                };
            }

            // true if there's nothing more to process, ever.
            fn _processed(self: *Batch) bool {
                for (self.jobs) |j| {
                    switch (j) {
                        .new, .working => return false,
                        else => {},
                    }
                }
                return true;
            }

            // true if there's nothing more to retrieve, ever.
            fn _exhausted(self: *Batch) bool {
                for (self.jobs) |j| {
                    switch (j) {
                        .new, .working, .result_pending => return false,
                        else => {},
                    }
                }
                return true;
            }

            /// find a job with result, mark it as .result_retrieved and return ResponseMessage;
            fn _pop_response(self: *Batch) ?ResponseMessage {
                for (0..self.jobs.len) |idx| {
                    if (self.jobs[idx] != .result_pending) continue;
                    const out = ResponseMessage{
                        .ctx_id = self.ctx_ids[idx],
                        .response = self.jobs[idx].result_pending.response,
                        .duration_ms = self.jobs[idx].result_pending.work_duration_ms,
                    };
                    self.jobs[idx] = .result_retrieved;
                    return out;
                }
                return null; // either all done or none pending now
            }

            fn _show_self(self: *Batch) void {
                std.log.err(": {any}", .{DebugView.init(self)});
                for (self.jobs) |j| std.log.err(": j={any}", .{j});
            }

            /// Return either ResponseMessage, or null if all contexts have
            /// responded or timeout ran out.
            ///
            /// ResponseMessage contains the actual response as submitted by
            /// child,uration in milliseconds (based on wall clock, not time
            /// spent waiting), and context ID.
            ///
            /// The .watch_states slice tracks current status of individual
            /// pipes being watched by the iterator.  Each item in this slice
            /// contains context id of the pipe and current status:
            ///   *  .todo - waiting for response (should never appear on
            ///       exhausted iterator),
            ///   *  .result_pending - response has been returned,
            ///   *  .supersedes - response lost due to arrival of newer request,
            ///   *  .timeout - response lost due to timeout.
            ///
            /// This is particularly useful after the iterator has finished
            /// as it allows main thread to act specifically on contexts where
            /// the thread might be stuck.
            pub fn next(self: *Batch) ?ResponseMessage {
                self.bq.m.lock();
                defer self.bq.m.unlock();
                while (true) { // break => timeout
                    if (self.disowned) unreachable; // cannot get responses from disowned batch
                    if (self.remaining_ns == 0) break;
                    if (self._exhausted()) return null;
                    if (self._pop_response()) |msg| return msg;
                    const start_nst = std.time.nanoTimestamp();
                    self.bq.c.timedWait(&self.bq.m, self.remaining_ns) catch break;
                    const duration_ns = std.time.nanoTimestamp() - start_nst;
                    self.remaining_ns -|= @intCast(duration_ns);
                }
                for (0..self.jobs.len) |i| {
                    if (self.jobs[i] == .result_retrieved) continue;
                    self.jobs[i] = .{ .timeouted = self.ctx_ids[i] };
                }
                return null;
            }

            pub fn send_response(self: *Batch, ctx_id: CtxIdT, response: ResponseT) !void {
                {
                    self.bq.m.lock();
                    defer self.bq.m.unlock();
                    const job_ptr = self._job_ptr_by_ctx_id(ctx_id) orelse return error.InvalidCtxId;
                    switch (job_ptr.*) {
                        .new => unreachable, // cannot respond to new job
                        .working => {},
                        .result_pending => unreachable, // response already present
                        .result_retrieved => unreachable, // response already taken
                        .timeouted => unreachable, // job already timeouted
                    }
                    const work_duration_ms: u64 = @intCast(std.time.milliTimestamp() - job_ptr.working);
                    job_ptr.* = Job{ .result_pending = Result{
                        .response = response,
                        .work_duration_ms = work_duration_ms,
                    } };
                    if (self.disowned and self._processed()) self.deinit();
                }
                self.bq.c.signal();
            }
        };

        /// Returns memory allocated on .init().
        pub fn deinit(self: BatchQueueT) void {
            self.allocator.free(self.ctx_ids);
        }

        /// Returns count of currently held contexts.
        pub fn _count(self: *BatchQueueT) usize {
            var num: usize = 0;
            for (0..self.ctx_ids.len) |idx| {
                if (self.ctx_ids[idx] == null) continue;
                num += 1;
            }
            return num;
        }

        fn _idx_by_ctx_id(self: *BatchQueueT, ctx_id: CtxIdT) ?usize {
            for (0..self.ctx_ids.len) |idx| {
                if (self.ctx_ids[idx] != ctx_id) continue;
                return idx;
            }
            return null;
        }

        fn _idx_by_empty(self: *BatchQueueT) ?usize {
            for (0..self.ctx_ids.len) |idx| {
                if (self.ctx_ids[idx] != null) continue;
                return idx;
            }
            return null;
        }

        /// Use in parent thread to send *request* with all contexts previously enabled
        /// using .enable().
        ///
        /// Returns iterator which will provide responses, either until all
        /// of the active pipes have responded, or the cumulative time of waiting
        /// has reached *timeout_ns*.
        ///
        /// Allocates memory from self.allocator in order to create the iterator.
        /// Memory used by the iterator must be de-allocated by calling its
        /// .deinit() method.  (Calling self.deinit() is not enough.)
        ///
        /// Responses from iterator are embedded in struct with some meta-data;
        /// see Batch.next().
        pub fn queue_batch(self: *BatchQueueT, request: RequestT, timeout_ns: u64) !*Batch {
            const request_id = self.next_batch_id;
            self.next_batch_id +%= 1;
            const batch_ptr = b: {
                self.m.lock();
                defer self.m.unlock();
                break :b try self.batches.append_ptr(try Batch.init(
                    self,
                    request_id,
                    request,
                    timeout_ns,
                ));
            };
            self.c.signal();
            return batch_ptr;
        }

        /// Deletes pipe associated with the corresponding *ctx_id*.
        ///
        /// Invalidates pipe pointer as well as any iterators previously
        /// obtained using .queue_batch().
        ///
        /// Does nothing if no pipe with *ctx_id* exists.
        pub fn delete_ctx(self: *BatchQueueT, ctx_id: CtxIdT) void {
            self.m.lock();
            defer self.m.unlock();
            const idx = self._idx_by_ctx_id(ctx_id) orelse return;
            self.ctx_ids[idx] = null;
        }

        /// Use in parent thread to add pipe and mark it with *ctx_id*.
        ///
        /// *ctx_id* must be unique for each child thread and will be
        /// included with the response to enable parent interpret
        /// responses in correct context.
        pub fn add_ctx(self: *BatchQueueT, ctx_id: CtxIdT) !void {
            self.m.lock();
            defer self.m.unlock();
            const free_idx = self._idx_by_empty() orelse return error.PipeGroupFull;
            self.ctx_ids[free_idx] = ctx_id;
        }

        /// Use in child thread to block until request is received.
        ///
        /// *ctx_id* must be the id passed by .add() when creating
        /// parent for this child thread.
        ///
        /// Returns context object;  chuld must simply call .respond() on the
        /// returned object, passing the created response.
        pub fn wait_request(self: *BatchQueueT, ctx_id: CtxIdT) !JobContext {
            const job_ctx = b: {
                self.m.lock();
                defer self.m.unlock();
                while (true) : (self.c.wait(&self.m)) {
                    const batch_ptr = self.batches.peek_ptr() orelse continue;
                    if (batch_ptr.disowned) continue;
                    if (try batch_ptr.assign_new_job(ctx_id)) |jc| break :b jc;
                }
            };
            self.c.signal();
            return job_ctx;
        }
    };
}
